#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include <d3d8.h>

#include <System/Tools.h>

#include <Engine/Mesh.h>

#include <Math/Vector.h>

#include <DirectXRenderer/DirectXCache.h>
#include <DirectXRenderer/DirectXRenderer.h>
#include <DirectXRenderer/DirectXHelpers.h>

namespace Ge
{

//------------------------------------------------------------------------------
DirectXCache::DirectXCache(const Mesh& mesh, DirectXRenderer& renderer) :
//------------------------------------------------------------------------------
mMesh(mesh),
mRenderer(renderer),
mVertexFormat(0),
mVertexBuffer(0),
mIndexBuffer(0),
mLastVertexCount(0),
mLastIndexCount(0)
{
}


//------------------------------------------------------------------------------
DirectXCache::~DirectXCache()
//------------------------------------------------------------------------------
{
    DX_RELEASE(mVertexBuffer);
    DX_RELEASE(mIndexBuffer);
}


//------------------------------------------------------------------------------
bool DirectXCache::update()
//------------------------------------------------------------------------------
{   
    GE_ASSERT_MSG(mMesh.getVertexCount() > 0, "A mesh with no vertices?");
    // If number of vertices in the mesh has changed reallocate to a new vertex
    // buffer that fits the new size.
    bool ok(reallocateVertexBuffer());
    // Fill the vertex buffer with the data
    if (ok && mMesh.getVertexCount() > 0)
    {
        ok = createVertexBuffer();
    }

    // If the number of indices in the mesh has changed reallocate to a fitting
    // new index buffer.
    if (ok)
    {
        ok = reallocateIndexBuffer();
    }

    // Fill the index buffer with the data
    if (ok && mMesh.getIndexCount() > 0)
    {
        ok = createIndexBuffer();
    }

    GE_ASSERT(ok);
    return ok;
}

//------------------------------------------------------------------------------
bool DirectXCache::reallocateVertexBuffer()
//------------------------------------------------------------------------------
{
    // Determine Direct3d flexible vertex format
    UInt vertexFormat = D3DFVF_XYZ;

    if (mMesh.getColors(0) != 0)
        vertexFormat |= D3DFVF_DIFFUSE;

    if (mMesh.getColors(1) != 0)
        vertexFormat |= D3DFVF_SPECULAR;

    if (mMesh.getActiveFrame().getNormals() != 0)
        vertexFormat |= D3DFVF_NORMAL;

    if (mMesh.getUvType(0) == Mesh::UV_ARRAY)
        vertexFormat |= D3DFVF_TEX1;

    if (mMesh.getUvType(1) == Mesh::UV_ARRAY)
    {
        vertexFormat &= ~D3DFVF_TEX1;
//        vertexFormat |= D3DFVF_TEX2;
        vertexFormat |= (D3DFVF_TEX2|D3DFVF_TEXCOORDSIZE2(0)|D3DFVF_TEXCOORDSIZE2(1));
    }

    if (mLastVertexCount != mMesh.getVertexCount() ||
        vertexFormat != mVertexFormat)
    {
        mVertexFormat = vertexFormat;
        mLastVertexCount = mMesh.getVertexCount();

        // Release old buffer first
        DX_RELEASE(mVertexBuffer);

        DWORD usage = D3DUSAGE_WRITEONLY;
        if (mMesh.isDynamic())
        {
            usage |= D3DUSAGE_DYNAMIC;
        }

        // Allocate new vertex buffer
        return DX_CHECK_MSG(mRenderer.getDirect3dDevice().CreateVertexBuffer(
            mMesh.getVertexCount()*Dx::getSizeFromVertexFormat(mVertexFormat),
            usage,
            mVertexFormat,
            mMesh.isDynamic() ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED,
            &mVertexBuffer), "Could not create vertex buffer");
    }
    else
        return true;
}

//------------------------------------------------------------------------------
bool DirectXCache::reallocateIndexBuffer()
//------------------------------------------------------------------------------
{
    if (mMesh.getIndexCount() == 0)
    {
        DX_RELEASE(mIndexBuffer);
        mLastIndexCount = 0;
        return true;
    }
    else if (mLastIndexCount != mMesh.getIndexCount())
    {
        // Release old buffer first
        DX_RELEASE(mIndexBuffer);
        mLastIndexCount = mMesh.getIndexCount();

        DWORD usage = D3DUSAGE_WRITEONLY;
        if (mMesh.isDynamic())
        {
            usage |= D3DUSAGE_DYNAMIC;
        }

        // Allocate index buffer
        return DX_CHECK_MSG(mRenderer.getDirect3dDevice().CreateIndexBuffer(
            mMesh.getIndexCount()*sizeof(UInt16),
            usage,
            D3DFMT_INDEX16,
            mMesh.isDynamic() ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED,
            &mIndexBuffer), "Could not create index buffer");
    }
    else
        return true;
}

//------------------------------------------------------------------------------
bool DirectXCache::createVertexBuffer()
//------------------------------------------------------------------------------
{
    bool ret(true);

    // Fill vertex buffer
    BYTE* vbData;
    DWORD flags = mMesh.isDynamic() ? D3DLOCK_DISCARD : 0;
    ret = DX_CHECK_MSG(mVertexBuffer->Lock(0, 0, (BYTE**)&vbData, flags),
        "Could not lock vertex buffer");
    if (ret)
    {
        UInt stride = Dx::getSizeFromVertexFormat(getVertexFormat());
        UInt vc = mMesh.getVertexCount();

        // Copy vertices
        copyDestinationStride(
            reinterpret_cast<char*>(vbData),
            reinterpret_cast<const char*>(mMesh.getActiveFrame().getVertices()),
            sizeof(Vector3), vc, stride);                     
        vbData += sizeof(Vector3);

        // Copy normals
        if (mMesh.getActiveFrame().getNormals() != 0)
        {
            copyDestinationStride(
                reinterpret_cast<char*>(vbData),
                reinterpret_cast<const char*>(mMesh.getActiveFrame().getNormals()),
                sizeof(Vector3), vc, stride);
            vbData += sizeof(Vector3);
        }
    
        // Copy colors
        for (UInt c = 0; c<2; ++c)
        {
            if (mMesh.getColors(c) != 0)
            {
                UInt8* color = reinterpret_cast<UInt8*>(vbData);
                const Color* colors = mMesh.getColors(c);
                for (UInt i = 0; i<mMesh.getVertexCount(); ++i)
                {
                    Dx::colorToD3dColor(
                        colors[i], *(reinterpret_cast<D3DCOLOR*>(color))
                        );
                    color += stride;
                }
                vbData += sizeof(D3DCOLOR);
            }
        }

        // Copy uv coordinates
        for (UInt t = 0; t<2; ++t)
        {
            if (mMesh.getUvType(t) == Mesh::UV_ARRAY)
            {
                copyDestinationStride(
                    reinterpret_cast<char*>(vbData),
                    reinterpret_cast<const char*>(mMesh.getUvs(t)),
                    sizeof(Vector2), vc, stride);
                vbData += sizeof(Vector2);
            }
        }

        ret = DX_CHECK_MSG(mVertexBuffer->Unlock(), "Could not unlock vertex buffer");
    }

    return ret;
}
        

//------------------------------------------------------------------------------
bool DirectXCache::createIndexBuffer()
//------------------------------------------------------------------------------
{
    bool ret(true);

    // Copy indices
    UInt16* ibData;
    DWORD flags = mMesh.isDynamic() ? D3DLOCK_DISCARD : 0;
    ret = DX_CHECK_MSG(mIndexBuffer->Lock(
        0,
        mMesh.getIndexCount()*sizeof(UInt16),
        (BYTE**)&ibData,
        flags), "Could not lock index buffer");

    if (ret)
    {
        for (Int i = 0; i<mMesh.getIndexCount(); i++)
            ibData[i] = mMesh.getIndices()[i];

        ret = DX_CHECK_MSG(mIndexBuffer->Unlock(),
            "Could not unlock index buffer");
    }
    
    return ret;
}



//------------------------------------------------------------------------------
void DirectXCache::render() const
//------------------------------------------------------------------------------
{
    mRenderer.render(*this);
}



//------------------------------------------------------------------------------
UInt DirectXCache::getVertexFormat() const
//------------------------------------------------------------------------------
{
    return mVertexFormat;
}


//------------------------------------------------------------------------------
IDirect3DVertexBuffer8& DirectXCache::getVertexBuffer() const
//------------------------------------------------------------------------------
{
    GE_ASSERT(mVertexBuffer != 0);
    return *mVertexBuffer;
}


//------------------------------------------------------------------------------
IDirect3DIndexBuffer8& DirectXCache::getIndexBuffer() const
//------------------------------------------------------------------------------
{
    GE_ASSERT(mIndexBuffer != 0);
    return *mIndexBuffer;
}

//------------------------------------------------------------------------------
const Mesh& DirectXCache::getMesh() const
//------------------------------------------------------------------------------
{
    return mMesh;
}


} // namespace Ge