#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

// Comment this in to enable the shader playground
// #define SHADER

#include <System/Assert.h>
#include <System/Log.h>

#include "DirectXRenderer.h"
#include "DirectXCache.h"
#include "DirectXTexture.h"
#include "ShaderPlayground.h"

#include <Math/Matrix.h>

#include <Engine/Material.h>
#include <Engine/Light.h>
#include <Engine/Mesh.h>
#include <Engine/DisplayStatistics.h>
#include <Engine/Emitter.h>
#include <Engine/Particle.h>
#include <Engine/Camera.h>
#include <Engine/VideoCard.h>

#include <DirectXRenderer/DirectXHelpers.h>
#include <d3d8.h>
#include <d3dx8.h>
#include <sstream>

#ifdef SHADER
#   include <Engine/Camera.h>
#   pragma message("Custom Shader")
#else
#   pragma message("Fixed Shader")
#endif

namespace
{

using namespace Ge;

inline
//------------------------------------------------------------------------------
HWND toHWND(Ge::Handle handle)
//------------------------------------------------------------------------------
{
    return reinterpret_cast<HWND>(handle);
}

//------------------------------------------------------------------------------
// Shadow stuff
//------------------------------------------------------------------------------

IDirect3DVertexBuffer8* shadowBuffer = 0;
DWORD shadowFvf = D3DFVF_XYZ;
DWORD shadowQuadFvf = D3DFVF_XYZRHW|D3DFVF_DIFFUSE;
const UInt SHADOW_BUFFER_SIZE = 1800;

//------------------------------------------------------------------------------
bool createShadowBuffer(DirectXRenderer* that)
//------------------------------------------------------------------------------
{
    return DX_CHECK_MSG(that->getDirect3dDevice().CreateVertexBuffer(
            SHADOW_BUFFER_SIZE*Dx::getSizeFromVertexFormat(shadowFvf),
            D3DUSAGE_WRITEONLY|D3DUSAGE_DYNAMIC,
            shadowFvf,
            D3DPOOL_DEFAULT,
            &shadowBuffer), "Could not create shadow buffer");
}

//------------------------------------------------------------------------------
void releaseShadowBuffer()
//------------------------------------------------------------------------------
{
    if (shadowBuffer != 0)
    {
        shadowBuffer->Release();
    }
}


//------------------------------------------------------------------------------
// Particle stuff
//------------------------------------------------------------------------------

IDirect3DVertexBuffer8* particleBuffer = 0;
DWORD particleFvf = D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1;
// One particle quad has 4 vertices. Buffer can hold 300 particles.
const UInt PARTICLE_BUFFER_SIZE = 4*300;
struct ParticleVertex
{
    Vector3 mPosition;
    DWORD mColor;
    Vector2 mUv;
};

//------------------------------------------------------------------------------
bool createParticleBuffer(DirectXRenderer* that)
//------------------------------------------------------------------------------
{
    return DX_CHECK_MSG(that->getDirect3dDevice().CreateVertexBuffer(
            sizeof(ParticleVertex)*PARTICLE_BUFFER_SIZE,
            D3DUSAGE_WRITEONLY|D3DUSAGE_DYNAMIC,
            particleFvf,
            D3DPOOL_DEFAULT,
            &particleBuffer), "Could not create particle buffer");
}

//------------------------------------------------------------------------------
void releaseParticleBuffer()
//------------------------------------------------------------------------------
{
    if (particleBuffer != 0)
    {
        particleBuffer->Release();
    }
}


} // namespace






namespace Ge
{

//------------------------------------------------------------------------------
DirectXRenderer::DirectXRenderer(Handle window,
                                 const Color& background) :
//------------------------------------------------------------------------------
Renderer(background),
mWindow(window),
mLightsInScene(0),
mPrintFont(0),
mDirect3d(0),
mDirect3dDevice(0),
mMaterial(0)
{
    bool check(true);
    
    GE_LOG_INFO("Initializing Direct3D ..."<<std::endl);

    check = createDirect3d();
    GE_ASSERT(check);
    if (check)
    {
        GE_LOG_INFO("Obtained Direct3D interface."<<std::endl);
        check = createDirect3dDevice();
        GE_ASSERT(check);
    }
    
    if (check)
    {
        GE_LOG_INFO("Created Direct3D device."<<std::endl);
        check = createPrintFont();
        GE_ASSERT(check);
    }

    if (check)
    {
        GE_LOG_INFO("Created print font."<<std::endl);
        GE_LOG_INFO("Direct3D initialization complete."<<std::endl);
    }

    
    // Set default RenderStates
    mStates = StateBlock::createDelta();
    set(StateBlock());

//    // Set lighting to default mode
//    switchLighting(mLighting);

    createShadowBuffer(this);

    createParticleBuffer(this);

#ifdef SHADER
    initShader();
#endif
}


//------------------------------------------------------------------------------
DirectXRenderer::~DirectXRenderer()
//------------------------------------------------------------------------------
{
#ifdef NDEBUG
    GE_LOG_INFO("Shutting down Direct3D ..."<<std::endl);
#endif
    
    if (mDirect3dDevice)
    {
        releaseParticleBuffer();
        releaseShadowBuffer();

        getDirect3dDevice().SetTexture(0, 0);
        getDirect3dDevice().SetTexture(1, 0);
        getDirect3dDevice().SetTexture(2, 0);
        getDirect3dDevice().SetTexture(3, 0);
        mDirect3dDevice->Release();
    }

    if (mDirect3d)           
        mDirect3d->Release();

    if (mPrintFont)
        mPrintFont->Release();

#ifdef NDEBUG
    GE_LOG_INFO("Direct3D shut down complete."<<std::endl);
#endif
    
    mDirect3dDevice = 0;
    mDirect3d = 0;
}



//------------------------------------------------------------------------------
bool DirectXRenderer::createDirect3d()
//------------------------------------------------------------------------------
{
    // Create the D3D object.
    mDirect3d = Direct3DCreate8(D3D_SDK_VERSION);
    GE_ASSERT(mDirect3d != 0);
    
    return mDirect3d == 0 ? false : true;
}


//------------------------------------------------------------------------------
bool DirectXRenderer::createDirect3dDevice()
//------------------------------------------------------------------------------
{
    bool ret(true);

    if (theVideoCard().getTextureUnitCount() < 2)
    {
        GE_ASSERT_MSG(false, "I'm assuming at least 2 TUs.");
        GE_LOG_ERROR("Card has less than 2 textures units."<<std::endl);
        DirectXRenderer::~DirectXRenderer();
        exit(EXIT_FAILURE);
    }

    if (!theVideoCard().supportsWindowedRendering())
    {
        GE_ASSERT_MSG(false, "Card does not support windowed rendering.");
        GE_LOG_ERROR("Card does not support windowed rendering."<<std::endl);
        DirectXRenderer::~DirectXRenderer();
        exit(EXIT_FAILURE);
    }

    // Get the current desktop display mode, so we can set up a back
    // buffer of the same format
    D3DDISPLAYMODE displayMode;
    ret = DX_CHECK_MSG(mDirect3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &displayMode),
        "Could not get display mode");
    
    if (ret)
    {
        // Set up the present parameters structure used to create the D3DDevice
        // (We use windowed mode and D3DSWAPEFFECT_DISCARD
        D3DPRESENT_PARAMETERS presentParameters;
        memset(&presentParameters, 0, sizeof(presentParameters));
        presentParameters.Windowed = TRUE;
        presentParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
        presentParameters.BackBufferFormat = displayMode.Format;
        presentParameters.BackBufferCount = 1;
        presentParameters.hDeviceWindow = HWND(mWindow);
        
        // Depth stencil parameters
        presentParameters.EnableAutoDepthStencil = TRUE;
        if (theVideoCard().supports(VideoCard::DS_24_8))
        {
            presentParameters.AutoDepthStencilFormat = D3DFMT_D24S8;
            GE_LOG_INFO("Use 24/8 depth/stencil."<<std::endl);
        }
        else if (theVideoCard().supports(VideoCard::DS_24_4))
        {
            presentParameters.AutoDepthStencilFormat = D3DFMT_D24X4S4;
            GE_LOG_INFO("Use 24/4 depth/stencil."<<std::endl);
        }
        else if (theVideoCard().supports(VideoCard::DS_32))
        {
            presentParameters.AutoDepthStencilFormat = D3DFMT_D32;
            GE_LOG_INFO("Use 32/0 depth/stencil."<<std::endl);
        }
        else
        {
            bool ok = theVideoCard().supports(VideoCard::DS_16);
            GE_ASSERT(ok);
            if (!ok)
            {
                GE_LOG_ERROR("No appropriate dept/stencil format found."<<std::endl);
            }
            else
            {
                GE_LOG_INFO("Use 16/0 depth/stencil."<<std::endl);
                presentParameters.AutoDepthStencilFormat = D3DFMT_D16;
            }
        }

        HWND hw = toHWND(mWindow);

        DWORD behaviour;
        if (theVideoCard().supportsHwTl())
        {
            GE_LOG_INFO("Card has hardware T&L."<<std::endl);
            behaviour = D3DCREATE_HARDWARE_VERTEXPROCESSING;
        }
        else
        {
            GE_LOG_INFO("Card doesn't support hardware T&L."<<std::endl);
            behaviour = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        }
        

        // Create the D3DDevice.
        ret = DX_CHECK_MSG(mDirect3d->CreateDevice(D3DADAPTER_DEFAULT,
                                           D3DDEVTYPE_HAL,
                                           hw,
                                           behaviour,
                                           &presentParameters,
                                           &mDirect3dDevice),
            "Could not create Direct3dDevice");
   
        if (ret)
        {
            // If not done, then scaling results in wrong lighting
            mDirect3dDevice->SetRenderState(D3DRS_NORMALIZENORMALS, TRUE);
        }
    }

    return ret;
}


//------------------------------------------------------------------------------
void DirectXRenderer::set(const StateBlock& sb)
//------------------------------------------------------------------------------
{
    StateBlock delta = StateBlock::createDelta(mStates, sb);

    // ToDo: I should replace this by a template member later,
    // but it works fine now.
#define CHECK_STATE(stateEnum, stateType) \
    if (delta.get(RenderState::##stateType) != RenderState::NO_CHANGE) \
    { \
        RenderState::##stateEnum state = \
            RenderState::##stateEnum(delta.get(RenderState::##stateType)); \
        set(state); \
    }

    CHECK_STATE(Blend, BLEND);
    CHECK_STATE(Fill, FILL);
    CHECK_STATE(Shade, SHADE);
    CHECK_STATE(Specular, SPECULAR);
    CHECK_STATE(Depth, DEPTH);
    CHECK_STATE(DepthCompare, DEPTH_COMPARE);
    CHECK_STATE(Stencil, STENCIL);
    CHECK_STATE(ColorWrite, COLOR_WRITE);
    CHECK_STATE(Culling, CULLING);
    CHECK_STATE(Lighting, LIGHTING);
}

//------------------------------------------------------------------------------
void DirectXRenderer::set(const Material& m)
//------------------------------------------------------------------------------
{
    mMaterial = &m;
}

//------------------------------------------------------------------------------
bool DirectXRenderer::createPrintFont()
//------------------------------------------------------------------------------
{
    bool ret(true);
    // Create font
    HFONT hFont = CreateFont(15, 0, 0, 0, FALSE, FALSE,
                          FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
                          CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
                          VARIABLE_PITCH, "Arial" );

    GE_ASSERT(hFont != NULL);
    if (hFont != NULL)
        // Create D3d-font from font
        ret = DX_CHECK_MSG(D3DXCreateFont(mDirect3dDevice, hFont, &mPrintFont),
            "Could not create D3d-Font");
    else
        ret = false;

    return ret;
}



//------------------------------------------------------------------------------
void DirectXRenderer::beginScene()
//------------------------------------------------------------------------------
{
    // Clear buffers
    D3DCOLOR back;
    Dx::colorToD3dColor(getBackgroundColor(), back);
    mDirect3dDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL,
                           back, 1.0f, 0);

    // ToDo: The whole light stuff must be reworked
    // Reset lights in scene
    mLightsInScene = 0;
    for (UInt i = 0; i<getPossibleNumberOfLights(); ++i)
    {
        mDirect3dDevice->LightEnable(i, 0);
    }

    // Begin the scene
    mDirect3dDevice->BeginScene();
}

//------------------------------------------------------------------------------
void DirectXRenderer::render(const Mesh& mesh)
//------------------------------------------------------------------------------
{
    if (mesh.cache() == 0)
    {
        mesh.cache() = GE_NEW DirectXCache(mesh, *this);
        mesh.cache()->update();
    }

    mesh.cache()->render();
}


//------------------------------------------------------------------------------
void DirectXRenderer::render(const Lines& lines)
//------------------------------------------------------------------------------
{
//    mDirect3dDevice->DrawIndexedPrimitiveUP(
//        D3DPT_LINELIST,
//        0,
//        lines.getVertexCount(),
//        lines.getLineCount(),
//        lines.getIndices(),
//        D3DFMT_INDEX16,
//        lines.getVertices(),
}

//------------------------------------------------------------------------------
void DirectXRenderer::render(const DirectXCache& cache)
//------------------------------------------------------------------------------
{
    // Set vertex stream
    mDirect3dDevice->SetVertexShader(cache.getVertexFormat());
    mDirect3dDevice->SetStreamSource(0, &cache.getVertexBuffer(),
        Dx::getSizeFromVertexFormat(cache.getVertexFormat()));
    mDirect3dDevice->SetIndices(&cache.getIndexBuffer(), 0);

    const Mesh& mesh = cache.getMesh();
    for (UInt i = 0; i<mMaterial->getPassCount(); ++i)
    {
        mMaterial->apply(i, mView);

        mDirect3dDevice->DrawIndexedPrimitive(
            D3DPT_TRIANGLELIST,
            0, // Starting vertex
            mesh.getVertexCount(),
            0,
            mesh.getPrimitiveCount());
    }

    // Update statistics
    theDisplayStatistics().render(mesh.getVertexCount(),
                                  mesh.getPrimitiveCount());
}




// ToDo: Later only render the shadow buffer when it's full and not for every
// shadow volume. This avoids unnecessary locking of vertex buffers
//------------------------------------------------------------------------------
void DirectXRenderer::render(const ShadowVolume& sv)
//------------------------------------------------------------------------------
{
    // Disable color & z-buffer write
    set(RenderState::COLOR_WRITE_OFF);
    set(RenderState::DEPTH_READ_ONLY);
    set(RenderState::DEPTH_COMPARE_LESS);
    set(RenderState::STENCIL_ON);
  
    // Current position in the shadow volumes vertex array. Since the size
    // of the shadow vertex buffer is limited to SHADOW_BUFFER_SIZE it is
    // possible that the buffer needs to be locked multiple times to draw
    // one shadow volume.
    UInt volumeVertexPos = 0;
    while (volumeVertexPos < sv.getVertexCount())
    {
        Vector3* vbData = 0;
        DX_CHECK_MSG(shadowBuffer->Lock(0, 0, (BYTE**)&vbData, D3DLOCK_DISCARD),
            "Could not lock vertex buffer");

        // SHADOW_BUFFER_SIZE must be divisible by 3!! So that always full triangles
        // are copied here.
        for (UInt v = 0;
             v<minimum(sv.getVertexCount()-volumeVertexPos, SHADOW_BUFFER_SIZE);
             ++v)
        {
            // Copy one triangle to the vertex buffer.
            memcpy(vbData, &(sv.getVertices()[v+volumeVertexPos]), sizeof(Vector3));
            vbData++;
        }

        shadowBuffer->Unlock();
        
        UInt triangleCount = v/3;
        volumeVertexPos += v;

        mDirect3dDevice->SetStreamSource(0, shadowBuffer, 3*sizeof(float));
        mDirect3dDevice->SetVertexShader(D3DFVF_XYZ);

        mDirect3dDevice->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_ALWAYS);
        mDirect3dDevice->SetRenderState(D3DRS_STENCILWRITEMASK, ~0);
        mDirect3dDevice->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);
        mDirect3dDevice->SetRenderState(D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP);    
    
        // Render back facing
        mDirect3dDevice->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_INCR);
        set(RenderState::CULLING_CLOCKWISE);
        mDirect3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, triangleCount);

        // Render front facing
        mDirect3dDevice->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_DECR);
        set(RenderState::CULLING_COUNTERCLOCKWISE);
        mDirect3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, triangleCount);
    }
}


// ToDo: Later only render the particle buffer when it's full and not for every
// emitter. This avoids unnecessary locking of vertex buffers.
// ToDo: Rework whole particle stuff. It's only hack the way it is now :(
//------------------------------------------------------------------------------
void DirectXRenderer::render(const Emitter& e)
//------------------------------------------------------------------------------
{
    // The particles have coordinates in world space.
    setWorldMatrix(Matrix3x4::IDENTITY());
    Matrix3x3 billboard = Matrix3x3::IDENTITY();
    mView.getSubMatrix(billboard);
    invertAffine(billboard);
    set(RenderState::DEPTH_READ_ONLY);
    set(RenderState::BLEND_SOURCE_ALPHA);
    set(RenderState::LIGHTING_OFF);
    set(RenderState::DEPTH_COMPARE_LESS_OR_EQUAL);

    e.getMaterial().apply(0, mView);

    // Copy particles to vertex buffer.
    ParticleVertex* vbData = 0;
    DX_CHECK_MSG(particleBuffer->Lock(0, 0, (BYTE**)&vbData, D3DLOCK_DISCARD),
        "Could not lock vertex buffer");
    
    UInt count = 0;
    for (Emitter::ConstIterator i = e.getBegin(); i != e.getEnd(); ++i)
    {
        ++count;

        DWORD color = i->getColor().get(Color::ARGB);
        Vector3 position =  i->getPosition();
        Float s2 = i->getSize()/2;
        Matrix3x3 transformation;
        makeZRotation(transformation, i->getAngle());
        transformation = billboard*transformation;

        vbData->mPosition = transformation*Vector3(-s2,s2,0) + position;
        vbData->mColor = color;
        ::Ge::set(vbData->mUv, 0.0f,0.0f);
        ++vbData;

        vbData->mPosition = transformation*Vector3(s2,s2,0) + position;
        vbData->mColor = color;
        ::Ge::set(vbData->mUv, 1.0f,0.0f);
        ++vbData;

        vbData->mPosition = transformation*Vector3(-s2,-s2,0) + position;
        vbData->mColor = color;
        ::Ge::set(vbData->mUv, 0.0f,1.0f);
        ++vbData;

        vbData->mPosition = transformation*Vector3(s2,-s2,0) + position;
        vbData->mColor = color;
        ::Ge::set(vbData->mUv, 1.0f,1.0f);
        ++vbData;
    }

    particleBuffer->Unlock();

    getDirect3dDevice().SetStreamSource(0, particleBuffer, sizeof(ParticleVertex));
    getDirect3dDevice().SetVertexShader(particleFvf);

    for (UInt k = 0; k<count; ++k)
    {
        getDirect3dDevice().DrawPrimitive(D3DPT_TRIANGLESTRIP, 4*k, 2);
    }
}

//------------------------------------------------------------------------------
void DirectXRenderer::endScene()
//------------------------------------------------------------------------------
{
    // End the scene.
    mDirect3dDevice->EndScene();

    // Flip
    mDirect3dDevice->Present(NULL, NULL, NULL, NULL);
}



//------------------------------------------------------------------------------
void DirectXRenderer::setWorldMatrix(const Matrix4x4& matrix)
//------------------------------------------------------------------------------
{
#ifdef SHADER
    world = matrix;
#endif

    // Store the current world matrix
    mWorld = matrix;

    D3DMATRIX d3dMatrix;
    Dx::matrixToD3dMatrix(matrix, d3dMatrix);
    mDirect3dDevice->SetTransform(D3DTS_WORLD, &d3dMatrix);
}

//------------------------------------------------------------------------------
void DirectXRenderer::setWorldMatrix(const Matrix3x4& matrix)
//------------------------------------------------------------------------------
{
    Matrix4x4 world;
    world = Matrix4x4::IDENTITY();
    world.setSubMatrix(matrix);
    setWorldMatrix(world);
}


//------------------------------------------------------------------------------
void DirectXRenderer::setViewMatrix(const Matrix4x4& matrix)
//------------------------------------------------------------------------------
{
#ifdef SHADER
    view = matrix;
#endif

    // Store the current view matrix
    mView = matrix;

    D3DMATRIX d3dMatrix;
    Dx::matrixToD3dMatrix(matrix, d3dMatrix);
    mDirect3dDevice->SetTransform(D3DTS_VIEW, &d3dMatrix);
}


//------------------------------------------------------------------------------
void DirectXRenderer::setProjectionMatrix(const Matrix4x4& matrix)
//------------------------------------------------------------------------------
{
#ifdef SHADER
    projection = matrix;
#endif

    D3DXMATRIX d3dMatrix, temp;

    Dx::matrixToD3dMatrix(matrix, d3dMatrix);
    mDirect3dDevice->SetTransform(D3DTS_PROJECTION, &d3dMatrix);
}


//------------------------------------------------------------------------------
void DirectXRenderer::add(const Light& light)
//------------------------------------------------------------------------------
{
    D3DLIGHT8 d3dLight;
    memset(&d3dLight, 0, sizeof(d3dLight));
    
    bool isDirectional = light.getPosition()[3] == 0;
    d3dLight.Type = isDirectional ? D3DLIGHT_DIRECTIONAL : D3DLIGHT_POINT;

    Dx::colorToD3dColor(Color::BLACK, d3dLight.Ambient);
    Dx::colorToD3dColor(light.getColor(), d3dLight.Diffuse);
    Dx::colorToD3dColor(light.getColor(), d3dLight.Specular);

    if (isDirectional)
    {
        d3dLight.Direction.x = -light.getPosition()[0];
        d3dLight.Direction.y = -light.getPosition()[1];
        d3dLight.Direction.z = -light.getPosition()[2];
    }
    else
    {
        d3dLight.Position.x = light.getPosition()[0];
        d3dLight.Position.y = light.getPosition()[1];
        d3dLight.Position.z = light.getPosition()[2];
        // Linear attenuation. Light has 10% intensity at the
        // range border.
        d3dLight.Attenuation1 = 10.0f/light.getRange();
        d3dLight.Range = light.getRange();
    }
    
    GE_ASSERT(mLightsInScene+1 <= getPossibleNumberOfLights());
    mDirect3dDevice->SetLight(mLightsInScene, &d3dLight);
    mDirect3dDevice->LightEnable(mLightsInScene, TRUE);
    mLightsInScene++;
}

//------------------------------------------------------------------------------
void DirectXRenderer::switchLights(bool b)
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<mLightsInScene; ++i)
    {
        mDirect3dDevice->LightEnable(i, b);
    }
}

////------------------------------------------------------------------------------
//void DirectXRenderer::switchLighting(bool on)
////------------------------------------------------------------------------------
//{
//    GE_ASSERT(mDirect3dDevice != 0);
//    
//    if (on)
//    {
//        set(RenderState::LIGHTING_ON);
//    }
//    else
//    {
//        set(RenderState::LIGHTING_OFF);
//    }
//
//    mLighting = on;
//}

//------------------------------------------------------------------------------
void DirectXRenderer::setAmbientLight(const Color& color)
//------------------------------------------------------------------------------
{
    GE_ASSERT(mDirect3dDevice != 0);
    mDirect3dDevice->SetRenderState(D3DRS_AMBIENT, color.get(Color::ARGB));
}

//------------------------------------------------------------------------------
UInt DirectXRenderer::getPossibleNumberOfLights() const
//------------------------------------------------------------------------------
{
    // ### Optimize later: read only once and buffer it
    D3DCAPS8 caps;
    GE_ASSERT(mDirect3dDevice != 0);
    mDirect3dDevice->GetDeviceCaps(&caps);
    return caps.MaxActiveLights;
}





////------------------------------------------------------------------------------
//void DirectXRenderer::set(VertexShader& shader) const
////------------------------------------------------------------------------------
//{
//    /*
//    // Can't use reference here, since reference can only be assigned once
//    const VertexShader::Constant* constant = 0;
//    for (UInt i = 0; i<shader.getConstantCount(); i++)
//    {
//        constant = &(shader.getConstant(i));
//        mDirect3dDevice->SetVertexShaderConstant(
//            constant->mRegister, &(constant->mValue), 1);
//    }
//
//    if (shader.getHandle() == 0)
//        assembleShader(shader);
//
//    mDirect3dDevice->SetVertexShader(reinterpret_cast<DWORD>(shader.getHandle()));
//    */
//}
//

//------------------------------------------------------------------------------
IDirect3DDevice8& DirectXRenderer::getDirect3dDevice() const
//------------------------------------------------------------------------------
{
    return *mDirect3dDevice;
}


//------------------------------------------------------------------------------
void DirectXRenderer::print(UInt x, UInt y, const char* text)
//------------------------------------------------------------------------------
{
    RECT r = { x, y, x, y };
    // DT_CALCRECT calculates the correct RECT size to fit the string
    mPrintFont->DrawTextA(text, -1, &r, DT_CALCRECT, 0xffffffff);
    // DT_LEFT means oriented to the left
    mPrintFont->DrawTextA(text, -1, &r, DT_LEFT, 0xffffffff);
}


//------------------------------------------------------------------------------
void DirectXRenderer::print(const DisplayStatistics& stats)
//------------------------------------------------------------------------------
{
    std::stringstream text;
    text<<"Fps: "<<stats.getFramesPerSecond()<<std::endl;
    text<<"Vertices: "<<stats.getVertices()<<std::endl;
    text<<"Triangles: "<<stats.getTriangles();
    print(5, 30, text.str().c_str());
}








































//------------------------------------------------------------------------------
// Shader playground
//------------------------------------------------------------------------------

////------------------------------------------------------------------------------
//void DirectXRenderer::assembleShader(VertexShader& shader) const
////------------------------------------------------------------------------------
//{
//    ID3DXBuffer* tokens; 
//    ID3DXBuffer* errors;
//    Handle handle(0);
//
//    if (shader.getCode() != 0)
//    {
//        // Assemble the shader   
//        DX_CHECK_MSG(D3DXAssembleShader(shader.getCode(), strlen(shader.getCode()),
//            0, NULL, &tokens, &errors), "Vertex shader assembly error");
//
//        // Create the vertex shader handle
//        DX_CHECK(getDirect3dDevice().CreateVertexShader(
//            shaderDeclaration, (DWORD*)tokens->GetBufferPointer(),
//            reinterpret_cast<DWORD*>(&handle), 0));
//    }
//    else if (shader.getTokens() != 0)
//    {
//        DX_CHECK(getDirect3dDevice().CreateVertexShader(
//            shader.getDeclaration(), shader.getTokens(),
//            reinterpret_cast<DWORD*>(&handle), 0));
//    }
//    else
//        GE_ASSERT_MSG(false, "Neither code nor tokens present for this shader");
//
//    GE_ASSERT(handle != 0);
//    shader.setHandle(handle);
//}

//------------------------------------------------------------------------------
void DirectXRenderer::initShader()
//------------------------------------------------------------------------------
{
    ID3DXBuffer* tokens; 
    ID3DXBuffer* errors;

    // Assemble the shader   
    DX_CHECK_MSG(D3DXAssembleShader(shaderCode, strlen(shaderCode),
        0, NULL, &tokens, &errors), "Vertex shader assembly error");

    // Create the vertex shader handle
    DX_CHECK(getDirect3dDevice().CreateVertexShader(
        shaderDeclaration, (DWORD*)tokens->GetBufferPointer(),
        &shader, 0));
}

} // namespace Ge

