#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------


#include <d3dx8.h>

#include <Math/Vector.h>

#include <Engine/Mesh.h>
#include <Engine/Cluster.h>
#include <Engine/Link.h>
#include <Engine/Locator.h>
#include <Engine/Texture.h>


#include <DirectXRenderer/XImporter.h>
#include <DirectXRenderer/DirectXRenderer.h>
#include <DirectXRenderer/DirectXTexture.h>
#include <DirectXRenderer/DirectXHelpers.h>

#include <System/FilePath.h>

namespace
{

// Copies the <vertices>, <normals>, ... from <mesh>, starting at vertex <start>
// in the <mesh>'s vertex buffer. <count> vertices will be copied.
//------------------------------------------------------------------------------
void copyVertexData(ID3DXMesh& mesh, Ge::Vector3*& vertices, Ge::Vector3*& normals,
                    Ge::Color*& colors, Ge::Vector2*& textureUvs,
                    Ge::UInt start, Ge::UInt count)
//------------------------------------------------------------------------------
{
    using namespace Ge;

    // Get vertices pointer
    BYTE* d3dVertices = 0;
    mesh.LockVertexBuffer(0, &d3dVertices);
    
    // Verify vertex format
    bool knownVertexFormat = (mesh.GetFVF() &
        ~(D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1|D3DFVF_DIFFUSE)) == 0;

    GE_ASSERT_MSG(knownVertexFormat, "Unknown vertex format");
    if (knownVertexFormat)
    {
        UInt increment(0);
        vertices = GE_NEW Vector3[count];
        increment += sizeof(Vector3);

        if (mesh.GetFVF() & D3DFVF_NORMAL)
        {
            normals = GE_NEW Vector3[count];
            increment += sizeof(Vector3);
        }
        
        if (mesh.GetFVF() & D3DFVF_DIFFUSE)
        {
            colors = GE_NEW Color[count];
            increment += sizeof(D3DCOLOR);
        }

        if (mesh.GetFVF() & D3DFVF_TEX1)
        {
            textureUvs = GE_NEW Vector2[count];
            increment += sizeof(Vector2);
        }

        d3dVertices += start*increment;
        for (UInt i = 0; i<count; i++)
        {
            // ATTENTION: Due to Microsofts flexible vertex format the
            // order of copying vertex data has to be position, normal,
            // texture coordinate!!!

            vertices[i] = *(reinterpret_cast<Vector3*>(d3dVertices));
            d3dVertices += sizeof(Vector3);

            // If the mesh has normals copy them
            if (mesh.GetFVF() & D3DFVF_NORMAL)
            {
                normals[i] = *(reinterpret_cast<Vector3*>(d3dVertices));
                d3dVertices += sizeof(Vector3);
            }

            // If the mesh has diffuse vertex colors copy them
            if (mesh.GetFVF() & D3DFVF_DIFFUSE)
            {
                Dx::d3dColorToColor(*(reinterpret_cast<D3DCOLOR*>(d3dVertices)),
                                    colors[i]);
                d3dVertices += sizeof(D3DCOLOR);
            }

            // If the mesh has textures copy the texture coordinates
            if (mesh.GetFVF() & D3DFVF_TEX1)
            {
                textureUvs[i] = *(reinterpret_cast<Vector2*>(d3dVertices));
                d3dVertices += sizeof(Vector2);
            }
        }
    }

    mesh.UnlockVertexBuffer();
}

// Copies the <count> indices from the <mesh>'s index buffer to <indices>,
// starting at index <start>. <offset> gets subtracted from each vertex
// that is indexed.
// This function is needed, since the splitting of the
// DirectX mesh into smaller meshes (each corresponding to a material/texture)
// is done by copying the vertices into separate Meshes. When this is done,
// the vertex indices in the original DirectX mesh are not correct anymore,
// since vertices in every submesh restart at 0. This is taken into account
// by subtracting <offset>
//------------------------------------------------------------------------------
void copyIndexData(ID3DXMesh& mesh, Ge::UInt16*& indices,
                   Ge::UInt start, Ge::UInt count, Ge::UInt offset)
//------------------------------------------------------------------------------
{
    using namespace Ge;

    UInt16* indexArray = 0;
    mesh.LockIndexBuffer(0, reinterpret_cast<BYTE**>(&indexArray));
    indices = GE_NEW UInt16[count];
    
    for (UInt i = 0; i<count; i++)
    {
        indices[i] = indexArray[start+i]-offset;
    }

    mesh.UnlockIndexBuffer();
}

} // namespace





namespace Ge
{

//------------------------------------------------------------------------------
XImporter::XImporter(const char* file, const DirectXRenderer& renderer) :
//------------------------------------------------------------------------------
mMesh(0),
mMaterials(0),
mTextures(0),
mMaterialCount(0),
mRenderer(renderer),
mWasImported(false)
{
    GE_ASSERT(file != 0);
    if (file != 0)
    {
        FilePath fp(file);

        // Load the mesh from file
        DWORD materialCount;
        HRESULT ok = D3DXLoadMeshFromX(
            const_cast<char*>(file),
            D3DXMESH_SYSTEMMEM,
            &(renderer.getDirect3dDevice()), 
            NULL,
            &mMaterialBuffer,
            &materialCount,
            &mMesh);

        // File was successfully loaded
        GE_ASSERT_MSG(!FAILED(ok), "Error loading .x file");
        if (!FAILED(ok))
        {
            D3DXMATERIAL* d3dxMaterials =
                static_cast<D3DXMATERIAL*>(mMaterialBuffer->GetBufferPointer());
            mMaterialCount = materialCount;
            mMaterials = GE_NEW D3DMATERIAL8[mMaterialCount];
            mTextures = GE_NEW LPDIRECT3DTEXTURE8[mMaterialCount];
            
            for (UInt i = 0; i<mMaterialCount; i++)
            {
                // Copy the materials
                mMaterials[i] = d3dxMaterials[i].MatD3D;

                // Set the ambient color for the materials (D3DX does not do this)
                mMaterials[i].Ambient = mMaterials[i].Diffuse;
     
                // Load textures
                if (FAILED(D3DXCreateTextureFromFile(
                    &(renderer.getDirect3dDevice()),
                    (fp.getDirectory() + d3dxMaterials[i].pTextureFilename).c_str(),
                    &mTextures[i])))
                {
                    mTextures[i] = 0;
                }
            }
        }

        // Optimize it
        mMesh->OptimizeInplace(D3DXMESHOPT_ATTRSORT | D3DXMESHOPT_COMPACT, 0, 0, 0, 0);
    }
}


//------------------------------------------------------------------------------
XImporter::~XImporter()
//------------------------------------------------------------------------------
{
    if (mMaterials)
        GE_DELETE_A mMaterials;

    if (mTextures)
    {
        // If the mesh wasn't imported the textures must be released. Otherwise
        // the memory will be leaked. If the mesh was imported, the texture will
        // be freed after use
        if (!mWasImported)
            for (UInt i = 0; i<mMaterialCount; i++)
                if (mTextures[i] != 0)
                    mTextures[i]->Release();

        GE_DELETE_A mTextures;
    }


    if (mMesh)
    {
        mMesh->Release();
    }

    if (mMaterialBuffer)
    {
        mMaterialBuffer->Release();
    }
}

//------------------------------------------------------------------------------
const D3DMATERIAL8* XImporter::getD3dMaterial(UInt index) const
//------------------------------------------------------------------------------
{
    if (mMaterials)
    {
        return &mMaterials[index];
    }
    else
    {
        GE_ASSERT(false);
        return 0;
    }
}


//------------------------------------------------------------------------------
UInt XImporter::getMaterialCount() const
//------------------------------------------------------------------------------
{
    return mMaterialCount;
}

//------------------------------------------------------------------------------
Mesh** XImporter::constructMeshes(bool recalculateNormals,
                                  bool flipUvs) const
//------------------------------------------------------------------------------
{
    typedef Mesh* MeshPtr;
    MeshPtr* ret = 0;

    // These pointers are temporary. Each for loop allocates new memory for
    // storing the vertices & indices before they are copied to the newly
    // generated Mesh
    Vector3* vertices = 0;
    Vector3* normals = 0;
    Vector2* textureUvs = 0;
    Color* colors = 0;
    UInt16* indices = 0;

    // A D3DXATTRIBUTERANGE stores properties for DirectX mesh subsets
    D3DXATTRIBUTERANGE* attributes = GE_NEW D3DXATTRIBUTERANGE[getMaterialCount()];
    DWORD size = getMaterialCount();
    if (DX_CHECK(mMesh->GetAttributeTable(attributes, &size)))
    {
        ret = GE_NEW MeshPtr[size];
        for (UInt i = 0; i<size; i++)
        {
            // Attributes could be not in order, but now I'm too lazy
            // to check that
            GE_ASSERT(attributes[i].AttribId == i);
            
            copyVertexData(*mMesh, vertices, normals, colors, textureUvs,
                           attributes[i].VertexStart, attributes[i].VertexCount);
            
            copyIndexData(*mMesh, indices, attributes[i].FaceStart*3,
                          attributes[i].FaceCount*3, attributes[i].VertexStart);

            ret[i] = GE_NEW Mesh(
                1,
                attributes[i].VertexCount,
                vertices,
                normals,
                textureUvs,
                colors,
                attributes[i].FaceCount*3,
                indices);

            if (recalculateNormals)
            {
                ret[i]->calculateNormals(true, true);
            }

            if (flipUvs)
            {
                ret[i]->flipUvs(0);
            }

            // Reset pointers for next Mesh.
            vertices = 0;
            normals = 0;
            textureUvs = 0;
            colors = 0;
            indices = 0;
        }
    }
    GE_DELETE_A attributes;

    return ret;
}

//------------------------------------------------------------------------------
Material** XImporter::constructMaterials() const
//------------------------------------------------------------------------------
{
    typedef Material* MaterialPtr;
    MaterialPtr* ret = GE_NEW MaterialPtr[getMaterialCount()];
    D3DXMATERIAL* d3dxMaterials =
        static_cast<D3DXMATERIAL*>(mMaterialBuffer->GetBufferPointer());

    for (UInt i = 0; i<getMaterialCount(); i++)
    {
        ret[i] = GE_NEW Material;
        Dx::d3dMaterialToMaterial(*(getD3dMaterial(i)), *(ret[i]));

        if (mTextures[i] != 0)
        {
            ret[i]->set(0, *GE_NEW DirectXTexture(*(mTextures[i]),
                                           d3dxMaterials[i].pTextureFilename,
                                           mRenderer));
        }

    }

    return ret;
}

//------------------------------------------------------------------------------
Cluster* XImporter::import(bool importMaterialAndTexture,
                           bool recalculateNormals, bool flipUvs) const
//------------------------------------------------------------------------------
{
    Cluster* ret = 0;
    GE_ASSERT(getMaterialCount() > 0);
    if (getMaterialCount() > 0)
    {
        // Construct the needed data
        Material** materials = constructMaterials();
        Mesh** meshes = constructMeshes(recalculateNormals, flipUvs);
        GE_ASSERT(materials != 0 && meshes != 0);
        // Group them into a Cluster
        ret = constructCluster(importMaterialAndTexture,
                               meshes, materials);
        
        if (materials)
        {
            GE_DELETE_A materials;
        }

        if (meshes)
        {
            GE_DELETE_A meshes;
        }
    }
    else
    {
        GE_ASSERT_MSG(false, "Meshes without materials aren't imported yet");
    }

    mWasImported = true;
    return ret;
}


//------------------------------------------------------------------------------
Cluster* XImporter::constructCluster(bool importMaterialAndTexture,
                                     Mesh* meshes[],
                                     Material* materials[]) const
//------------------------------------------------------------------------------
{
    Cluster* ret = 0;
    if (getMaterialCount() > 0)
    {
        ret = GE_NEW Cluster;
        Locator* locator = Locator::create(Matrix3x4::IDENTITY());
        ret->set(*locator);
        for (UInt i = 0; i<getMaterialCount(); ++i)
        {
            Link* link = GE_NEW Link(*locator, *meshes[i]);
            ret->add(*link);
            if (materials[i] != 0)
            {
                Material& mat = *materials[i];
                if (mat.getTexture(0) != 0 && !importMaterialAndTexture)
                {
                    mat.set(Color::WHITE);
                }

                link->set(mat);
//                GE_DELETE materials[i];
            }
            else
            {
                GE_ASSERT_MSG(false,
                "I thought there are no sub meshes without a valid material in an .x file");
            }
        }
    }

    return ret;
}


} // namespace Ge