//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef ENGINE_AUDIO_VISUAL_H
#define ENGINE_AUDIO_VISUAL_H

#include <System/Object.h>
#include <System/Pointer.h>
#include <System/Timer.h>

#include <Engine/LocalTransform.h>
#include <Engine/WorldTransform.h>
#include <Engine/TimeController.h>
#include "StateBlock.h"
#include "Material.h"


#include <Math/Sphere.h>
#include <Math/Intersection.h>

#include <vector>
#include <string>

namespace Ge
{

class Renderer;
class Camera;
class Sorter;
class HierarchyAnimation;
class KeyframeController;
class Node;

// Base class of everything in the scene tree. Controls transformations,
// render states, time controllers and cloning.
class AudioVisual : public Object
{
public:
    AudioVisual(const char* name = "NoName");

    //------------------------------------------------------------------------------
    // Renderer specific
    //------------------------------------------------------------------------------
    // This function cannot be const, since e.g. Billboards are changed during
    // rendering
    virtual
    void render(Renderer&) = 0;
    // During the accumulation pass, the AudioVisual can be added to the
    // sorter (using camera info for culling and billboard updates).
    // If a bounding volume is contained inside the frustum, then it is not
    // necessary to continue culling for it's children
    virtual
    void accumulate(Sorter&, const Camera&, bool continueCulling);
    
    //------------------------------------------------------------------------------
    // Geometry update
    //------------------------------------------------------------------------------
    void updateGeometry();

    LocalTransform& localTransform();
    const LocalTransform& localTransform() const;
    const WorldTransform& worldTransform() const;
    
    //------------------------------------------------------------------------------
    // Parent handling
    //------------------------------------------------------------------------------
    void setParent(AudioVisual* parent);
    const AudioVisual* getParent() const;

    //------------------------------------------------------------------------------
    // Animation
    //------------------------------------------------------------------------------
    // Attaching a HierarchyAnimation leads to a recursive call of
    // HierarchyAnimation::attachMatchingControllers()
    virtual
    void attach(HierarchyAnimation&);
    void attach(KeyframeController&);
    // Returns false if no TimeControllers are attached
    bool tick(const Timer::Unit&);

    //------------------------------------------------------------------------------
    // Misc
    //------------------------------------------------------------------------------
    // Clones the entire subtree starting at <*this>. Shared data like textures,
    // materials, meshes, ... are only referenced and not copied.
    virtual
    AudioVisual* clone() const = 0;
    virtual
    Node* asNode();
    virtual
    const Node* asNode() const;
    
    // This is a world space bound
    const Sphere& getBound() const;

    const std::string& getName() const;
    void setName(const char*);

    // This is needed for debugging (I did not implement ostream<<AudioVisual
    // because I might need other types of stream output for an AudioVisual and don't
    // want to bother with mutators).
    virtual
    void printSubTree(std::ostream&, UInt indentLevel = 0) const;

    const Material& getMaterial() const;

    void set(const Material&);

    StateBlock getStateBlock() const;

    void setStateBlock(const StateBlock&);

protected:
    friend class Node;

    // If there is a parent then multiply its world transform
    // and the AudioVisual's local transform to get the new world transform
    virtual
    void updateTransform();
    // Merge bounds of sisters to update parent bound (goes up recursively)
    virtual
    void updateBoundUp();
    // Transform bound according to world transform (goes recursively down
    // the scene tree)
    virtual
    void updateBoundDown() = 0;

    // Set the world space bound of this AudioVisual
    void setBound(const Sphere&);
    // Returns the Intersect::Type of this AudioVisual's bound and the
    // <camera>'s frustum. If <continueCulling> is false, no culling has to be done,
    // because the parent's bound was already completely in the frustum.
    // (If the parent's bount was completely outside the frustum, this function
    // wouldn't get called for a child).
    Intersect::Type cull(const Camera& camera, bool continueCulling) const;

    // Serves as a copy c'tor in the implementation of clone(). Doesn't make a
    // copy, but only copies relevant local data.
    void copyDataFrom(const AudioVisual&);

private:
    // Don't want a copy c'tor use clone() instead
    AudioVisual(const AudioVisual&);

    AudioVisual* mParent;
    LocalTransform mLocalTransform;
    WorldTransform mWorldTransform;
    StateBlock mStates;
    Material mMaterial;
    Sphere mBound;
    std::string mName;
    std::vector<Pointer<TimeController> > mControllers;
};

} // namespace Ge

#endif // ENGINE_AUDIO_VISUAL_H