#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Sep-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "BumpMaterial.h"
#include "Link.h"
#include "RenderCache.h"

#include <DirectXRenderer/DirectXRenderer.h>
#include <DirectXRenderer/DirectXHelpers.h>
#include <DirectXRenderer/DirectXTexture.h>

#include <d3d8.h>

namespace
{
    using namespace Ge;

    const Float SPECULAR_SCALE = 0.82f;

    // Self shadowing means that if tangent space light * vertex normal < 0 then
    // no lighting should be done even if the normal map normal would still
    // light some fragments.
    // <tangentLight> must be normalized
    inline
    //------------------------------------------------------------------------------
    void scaleSelfShadow(Vector3& tangentLight)
    //------------------------------------------------------------------------------
    {
        // The z-coordinate of the tangent space light vector contains N*L
        // where N is the vertex normal and L the tangent light vector.
        // For further explanation of this term see "A practical and robust bump..."
        // paper Equation 12 with c = 1/8.
        Float self = 8*tangentLight[2];
        clamp(self, 0.0f, 1.0f);
        tangentLight *= self;
    }

    inline
    //------------------------------------------------------------------------------
    void tangentLightToColor(Vector3& light, Color& color)
    //------------------------------------------------------------------------------
    {
        if (!light.isAlmost(Vector3::ORIGIN()))
        {
            light.normalize();
            scaleSelfShadow(light);
            convertVectorToColor(color, light, 1);
        }
        else
        {
            color = Color::BLACK;
        }
    }

} // namespace



namespace Ge
{

//------------------------------------------------------------------------------
BumpMaterial::BumpMaterial() :
//------------------------------------------------------------------------------
Material(2)
{
}

//------------------------------------------------------------------------------
void BumpMaterial::apply(UInt pass, const Matrix4x4& view) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(getLink() != 0 && pass < getPassCount());
    if (getLink() != 0 && pass < getPassCount())
    {
        switch (pass)
        {
        case 0:
            updateTangentLight();
            
            if (getPassCount() == 2)
            {
                updateHalfAngleLight(view);
            }

            if (getLink()->getMesh()->cache() != 0)
            {
                getLink()->getMesh()->cache()->update();
            }

            diffusePass();
            break;

        case 1:
            specularPass();
            break;

        default:
            GE_ASSERT_MSG(false, "BumpMaterial only has 2 passes");
        }
    }
}

//------------------------------------------------------------------------------
Material* BumpMaterial::clone() const
//------------------------------------------------------------------------------
{
    BumpMaterial* ret = GE_NEW BumpMaterial;
    ret->copyDataFrom(*this);
    return ret;
}

//------------------------------------------------------------------------------
void BumpMaterial::diffusePass() const
//------------------------------------------------------------------------------
{
    DirectXRenderer* dxr = reinterpret_cast<DirectXRenderer*>(&theRenderer());
    IDirect3DDevice8& device = dxr->getDirect3dDevice();

    D3DMATERIAL8 d3dMaterial;
    Dx::materialToD3dMaterial(*this, d3dMaterial);

    device.SetMaterial(&d3dMaterial);

    DX_CHECK(device.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE));
    DX_CHECK(device.SetTextureStageState(3, D3DTSS_COLOROP, D3DTOP_DISABLE));

    dxr->set(RenderState::LIGHTING_OFF);
    if (getTexture(0) != 0)
    {
        const DirectXTexture* t = reinterpret_cast<const DirectXTexture*>(getTexture(0));
        DX_CHECK(device.SetTexture(0, t->getDirect3dTexture()));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DOTPRODUCT3));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0));
    }
    else
    {
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE));
    }

    if (getTexture(1) != 0)
    {
        const DirectXTexture* t = reinterpret_cast<const DirectXTexture*>(getTexture(1));
        DX_CHECK(device.SetTexture(1, t->getDirect3dTexture()));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLOROP,  D3DTOP_MODULATE));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, 1));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_LINEAR));
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR));
    }
    else
    {
        DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE));
    }
}


//------------------------------------------------------------------------------
void BumpMaterial::activateSpecular(bool on)
//------------------------------------------------------------------------------
{
    setPassCount(on ? 2 : 1);
}



//------------------------------------------------------------------------------
void BumpMaterial::specularPass() const
//------------------------------------------------------------------------------
{
    DirectXRenderer* dxr = reinterpret_cast<DirectXRenderer*>(&theRenderer());
    IDirect3DDevice8& device = dxr->getDirect3dDevice();

    dxr->set(RenderState::DEPTH_READ_ONLY);
    dxr->set(RenderState::DEPTH_COMPARE_EQUAL);
    dxr->set(RenderState::BLEND_ADD_SQUARED);
    
    // Texture stage 0 gets half angles from specular color array.
    DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_SPECULAR));

    // Set dummy texture to activate stage 1.
    if (getTexture(1) == 0)
    {
        const DirectXTexture* t = reinterpret_cast<const DirectXTexture*>(getTexture(0));
        DX_CHECK(device.SetTexture(1, t->getDirect3dTexture()));
    }

    DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLOROP,  D3DTOP_MODULATE));
    DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_CURRENT));
    DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_CURRENT));
}

//------------------------------------------------------------------------------
void BumpMaterial::updateTangentLight() const
//------------------------------------------------------------------------------
{
    Mesh* mesh = getLink()->getMesh();
    bool ok = mesh->getActiveFrame().getTangents() != 0;
    GE_ASSERT(ok);
    if (!ok)
    {
        return;
    }

    // Reserve color array if not allocated
    if (mesh->getColors(0) == 0)
    {
        mesh->setColors(0, GE_NEW Color[mesh->getVertexCount()]);
    }

    Matrix3x4 inv = getLink()->getLocator()->getGlobal();
    invertAffine(inv);
    Vector3 objLight = inv*getLink()->getLight(0);
    
    // Directional light
    if (getLink()->getLight(0)[3] == 0)
    {
        Color* colors = mesh->getColors(0);
        objLight.normalize();
        Vector3 tangentLight;
        for (UInt i = 0; i<mesh->getVertexCount(); ++i)
        {
            tangentLight = mesh->getActiveFrame().getTangents()[i]*objLight;
            tangentLightToColor(tangentLight, colors[i]);
        }
    }
    // Point light
    else
    {
        Color* colors = mesh->getColors(0);
        Vector3 dir, tangentLight;
        for (UInt i = 0; i<mesh->getVertexCount(); ++i)
        {
            dir = objLight - mesh->getActiveFrame().getVertices()[i];
            dir.normalize();
            tangentLight = mesh->getActiveFrame().getTangents()[i]*dir;
            tangentLightToColor(tangentLight, colors[i]);
        }
    }
}


//------------------------------------------------------------------------------
void BumpMaterial::updateHalfAngleLight(const Matrix4x4& view) const
//------------------------------------------------------------------------------
{
    Mesh* mesh = getLink()->getMesh();
    bool ok = mesh->getActiveFrame().getTangents() != 0;
    GE_ASSERT(ok);
    if (!ok)
    {
        return;
    }

    // Reserve secondary color array if not allocated
    if (mesh->getColors(1) == 0)
    {
        mesh->setColors(1, GE_NEW Color[mesh->getVertexCount()]);
    }

//    GE_INTERVAL_TRACE(getLink()->getLight(0)<<',', 1);

    Matrix3x4 inv = getLink()->getLocator()->getGlobal();
    invertAffine(inv);
    Vector3 objLight = inv*getLink()->getLight(0);
    
    Matrix3x4 cameraLocator;
    view.getSubMatrix(cameraLocator);
    invertAffine(cameraLocator);
    Vector3 cameraPos;
    cameraLocator.getColumn(MatrixBase::TRANSLATION, cameraPos);    
    
//    GE_INTERVAL_TRACE(cameraPos<<std::endl, 1);

    cameraPos = inv*cameraPos;
    // Directional light
    if (getLink()->getLight(0)[3] == 0)
    {
        Color* colors = mesh->getColors(1);
        objLight.normalize();
        Vector3 halfAngle;
        for (UInt i = 0; i<mesh->getVertexCount(); ++i)
        {
            Vector3 cDir = cameraPos - mesh->getActiveFrame().getVertices()[i];
            cDir.normalize();
            halfAngle = cDir + objLight;
            halfAngle = mesh->getActiveFrame().getTangents()[i]*halfAngle;

            if (halfAngle != Vector3::ORIGIN())
            {
                halfAngle.normalize();
            }

            if (colors[i] != Color::BLACK)
            {
                convertVectorToColor(colors[i], halfAngle*SPECULAR_SCALE, 1);
            }
        }
    }
    // Point light
    else
    {
        Color* colors = mesh->getColors(1);
        Vector3 dir;
        Vector3 halfAngle;
        for (UInt i = 0; i<mesh->getVertexCount(); ++i)
        {
            Vector3 cDir = cameraPos - mesh->getActiveFrame().getVertices()[i];
            cDir.normalize();
                
            dir = objLight - mesh->getActiveFrame().getVertices()[i];
            dir.normalize();

            halfAngle = cDir + dir;
            halfAngle = mesh->getActiveFrame().getTangents()[i]*halfAngle;

            if (halfAngle != Vector3::ORIGIN())
            {
                halfAngle.normalize();
            }

            if (colors[i] != Color::BLACK)
            {
                convertVectorToColor(colors[i], halfAngle*SPECULAR_SCALE, 1);
            }
        }
    }
}


} // namespace Ge
