//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef ENGINE_CAMERA_H
#define ENGINE_CAMERA_H

#include "Locator.h"

#include <Math/Frustum.h>
#include <Math/Matrix.h>

#include <System/Pointer.h>

namespace Ge
{

// To implement some dynamic camera behaviour implement CameraController
// classes (e.g. look at ModelViewCameraController).
class Camera
{
public:
    // Default values:
    // - Aspect ratio is 4/3
    // - Horizontal field of view is PI_HALF
    // - Near plane is 1
    // - Far plane is 1000
    // The Locator must live on the heap.
    Camera(Locator&);

    //------------------------------------------------------------------------------
    // Accessors
    //------------------------------------------------------------------------------

    template <class MATRIX>
    void getViewMatrix(MATRIX& m) const
    {
        m = MATRIX::IDENTITY();
        m.setSubMatrix(mLocator->getGlobal());
        invertAffine(m);
    }

    template <class MATRIX>
    void getBillboardMatrix(MATRIX& m) const
    {
        m = MATRIX::IDENTITY();
        Matrix3x3 billboard;
        mLocator->getGlobal().getSubMatrix(billboard);
        m.setSubMatrix(billboard);
    }

    void getProjectionMatrix(Matrix4x4&) const;

    // Returns the frustum of the camera transformed into world space. Currently
    // I do visibility culling in world space, that's why I need this.
    const Frustum& getWorldFrustum() const;

    const Locator& getLocator() const;

    Locator& getLocator();

    // Returns horizontal/vertical (typically 4/3)
    Float getAspectRatio() const;

    // Set horizontal/vertical aspect ratio
    void setAspectRatio(Float);

    // Returns horizontal field of view in radians.
    Float getFieldOfView() const;

    // Set horizontal field of view in radians.
    void setFieldOfView(Float);

    // Returns the width of the viewport calculated from field of view
    // and near plane.
    Float getViewportWidth() const;

    Float getNearPlane() const;

    void setNearPlane(Float);

    Float getFarPlane() const;

    void setFarPlane(Float);

private:
    void recalculateFrustum() const;

    Pointer<Locator> mLocator;
    
    Float mAspectRatio;
    Float mFov;
    Float mNear;
    Float mFar;
    
    mutable Float mViewportWidth;
    mutable Frustum mFrustum;
    mutable bool mFrustumDirty;
};

} // namespace Ge

#endif // ENGINE_CAMERA_H