#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include "Color.h"

#include <Math/Interval.h>
#include <Math/Interpolation.h>

#include <System/Assert.h>
#include <System/BitManipulation.h>

namespace
{

using namespace Ge;

//------------------------------------------------------------------------------
bool isValidColor(Float r, Float g, Float b, Float a)
//------------------------------------------------------------------------------
{
    Interval range(0,1);
    return range.contains(r) && range.contains(g) &&
           range.contains(b) && range.contains(a);
}

} // namespace



namespace Ge
{

const Color Color::WHITE(1,1,1,1);
const Color Color::BLACK(0,0,0,1);
const Color Color::RED(1,0,0,1);
const Color Color::GREEN(0,1,0,1);
const Color Color::BLUE(0,0,1,0);
const Color Color::YELLOW(1,1,0,1);

//------------------------------------------------------------------------------
Color::Color() :
//------------------------------------------------------------------------------
mData(0xffffffff)
{
}

//------------------------------------------------------------------------------
Color::Color(Color::Format f, const Vector4& v)
//------------------------------------------------------------------------------
{
    set(f, v);
}

//------------------------------------------------------------------------------
Color::Color(Float r, Float g, Float b, Float a)
//------------------------------------------------------------------------------
{
    setFloat(r,g,b,a);
}

//------------------------------------------------------------------------------
Color::Color(const Color& color) :
//------------------------------------------------------------------------------
mData(color.mData)
{
}

//------------------------------------------------------------------------------
bool Color::operator ==(const Color& rhs) const
//------------------------------------------------------------------------------
{
    return mData == rhs.mData;
}

//------------------------------------------------------------------------------
bool Color::operator !=(const Color& rhs) const
//------------------------------------------------------------------------------
{
    return mData != rhs.mData;
}


//------------------------------------------------------------------------------
void Color::setFloat(Float r, Float g, Float b, Float a)
//------------------------------------------------------------------------------
{
    GE_ASSERT(isValidColor(r,g,b,a));
    mData = (UInt(r*255)&0xff)<<24 | (UInt(g*255)&0xff)<<16 |
            (UInt(b*255)&0xff)<<8  | (UInt(a*255)&0xff);
}


//------------------------------------------------------------------------------
void Color::setInt(UInt r, UInt g, UInt b, UInt a)
//------------------------------------------------------------------------------
{
    mData = r<<24 | g<<16 | b<<8 | a;
}

    
//------------------------------------------------------------------------------
void Color::set(Color::Format f, UInt32 color)
//------------------------------------------------------------------------------
{
    switch (f)
    {
    case RGBA:
        mData = color;
        break;

    case BGRA:
        mData = Bit::reflectByte(color);
        break;

    case ARGB:
        mData = ((color>>24)&0xff) |      // A
                ((color<<8)&0xffffff00);  // RGB
        break;

    default:
        GE_ASSERT(false);
    }
}

//------------------------------------------------------------------------------
void Color::set(Color::Format f, const Vector4& v)
//------------------------------------------------------------------------------
{
    switch (f)
    {
    case RGBA:
        setFloat(v[0], v[1], v[2], v[3]);
        break;

    case BGRA:
        setFloat(v[2], v[1], v[0], v[3]);
        break;

    case ARGB:
        setFloat(v[1], v[2], v[3], v[0]);
        break;
    }
}


//------------------------------------------------------------------------------
void Color::getFloat(Float& r, Float& g, Float& b, Float& a) const
//------------------------------------------------------------------------------
{
    r = Float((mData>>24)&0xff)/255;
    g = Float((mData>>16)&0xff)/255;
    b = Float((mData>>8)&0xff)/255;
    a = Float(mData&0xff)/255;
}

//------------------------------------------------------------------------------
UInt32 Color::get(Color::Format f) const
//------------------------------------------------------------------------------
{
    switch (f)
    {
    case RGBA:
        return mData;
        break;

    case BGRA:
        return Bit::reflectByte(mData);
        break;

    case ARGB:
        return ((mData<<24)&0xff000000) | // A
               ((mData>>8)&0x00ffffff);   // RGB
        break;

    default:
        GE_ASSERT(false);
        return mData;
    }
}

//------------------------------------------------------------------------------
UInt8 Color::get(Color::Channel c) const
//------------------------------------------------------------------------------
{
    switch (c)
    {
    case CH_RED:
        return (mData>>24)&0xff;
        break;

    case CH_GREEN:
        return (mData>>16)&0xff;
        break;

    case CH_BLUE:
        return (mData>>8)&0xff;
        break;

    case CH_ALPHA:
        return (mData)&0xff;
        break;

    default:
        GE_ASSERT(false);
        return 0;
    }
}

//------------------------------------------------------------------------------
Color lerp(const Color& c0, const Color& c1, Float value)
//------------------------------------------------------------------------------
{
    GE_ASSERT(0<=value && value<=1);
    Float r0, g0, b0, a0;
    c0.getFloat(r0, g0, b0, a0);
    Float r1, g1, b1, a1;
    c1.getFloat(r1, g1, b1, a1);
    return Color(
        lerp(r0, r1, value),
        lerp(g0, g1, value),
        lerp(b0, b1, value),
        lerp(a0, a1, value));
}

//------------------------------------------------------------------------------
void convertColorToVector(Vector3& v, const Color& c)
//------------------------------------------------------------------------------
{
    float alpha;
    c.getFloat(v[0], v[1], v[2], alpha);
    v -= Vector3(0.5f, 0.5f, 0.5f);
    v *= 2.0f;
}

//------------------------------------------------------------------------------
void convertVectorToColor(Color& c, const Vector3& v, Float alpha)
//------------------------------------------------------------------------------
{
//    GE_ASSERT_MSG(isAlmost(v.lengthSquare(), 1.0f, 0.001f),
//        "Vector is not normalized.");
    c.setFloat(v[0]*0.5f+0.5f, v[1]*0.5f+0.5f, v[2]*0.5f+0.5f, alpha);
}

} // namespace Ge
