#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Aug-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Emitter.h"
#include "Particle.h"

#include <Application/ScaledMetronome.h>

#include <Math/Random.h>

namespace Ge
{

//------------------------------------------------------------------------------
const Particle& Emitter::ConstIterator::operator *() const
//------------------------------------------------------------------------------
{
    return mEmitter->mParticles[mCurrent];
}

//------------------------------------------------------------------------------
const Particle* Emitter::ConstIterator::operator ->() const
//------------------------------------------------------------------------------
{
    return &(mEmitter->mParticles[mCurrent]);
}

//------------------------------------------------------------------------------
Emitter::ConstIterator& Emitter::ConstIterator::operator ++()
//------------------------------------------------------------------------------
{
    mCurrent = mEmitter->getNextParticle(mCurrent+1, true);
    return *this;
}

//------------------------------------------------------------------------------
bool Emitter::ConstIterator::operator ==(const Emitter::ConstIterator& rhs) const
//------------------------------------------------------------------------------
{
    return mCurrent == rhs.mCurrent && mEmitter == rhs.mEmitter;
}

//------------------------------------------------------------------------------
bool Emitter::ConstIterator::operator !=(const Emitter::ConstIterator& rhs) const
//------------------------------------------------------------------------------
{
    return !(*this == rhs);
}



//------------------------------------------------------------------------------
Emitter::Emitter() :
//------------------------------------------------------------------------------
mSpread(0),
mParticles(0),
mCreateCount(0)
{
    setSize(1, 1);
    setAngle(0, 0);
    setSpeed(Vector3(0,1,0));
}

//------------------------------------------------------------------------------
Emitter::Emitter(Float rate, Float lifetime) :
//------------------------------------------------------------------------------
mSpread(0),
mParticles(0),
mCreateCount(0)
{
    setSize(1, 1);
    setAmmount(rate, lifetime);
}

//------------------------------------------------------------------------------
Emitter::~Emitter()
//------------------------------------------------------------------------------
{
    if (mParticles != 0)
    {
        GE_DELETE_A mParticles;
        mParticles = 0;
    }
}

//------------------------------------------------------------------------------
void Emitter::setAmmount(Float rate, Float lifetime)
//------------------------------------------------------------------------------
{
    GE_ASSERT(rate > 0 && lifetime > 0);
    if (rate <= 0)
    {
        rate = 1;
    }

    if (lifetime <= 0)
    {
        lifetime = 1;
    }

    // Number of particles concurrently alive is rate*lifetime + 10% buffer.
    mParticleCount = rate*lifetime*1.1f;
    mRate = rate;
    mLifetime = lifetime;
    if (mParticles != 0)
    {
        GE_DELETE_A mParticles;
    }
    mParticles = GE_NEW Particle[mParticleCount];
}

//------------------------------------------------------------------------------
Float Emitter::getLifetime() const
//------------------------------------------------------------------------------
{
    return mLifetime;
}

//------------------------------------------------------------------------------
Emitter::ConstIterator Emitter::getBegin() const
//------------------------------------------------------------------------------
{
    ConstIterator begin;
    begin.mEmitter = this;
    begin.mCurrent = getNextParticle(0, true);
    return begin;
}

//------------------------------------------------------------------------------
Emitter::ConstIterator Emitter::getEnd() const
//------------------------------------------------------------------------------
{
    ConstIterator end;
    end.mEmitter = this;
    end.mCurrent = -1;
    return end;
}


//------------------------------------------------------------------------------
void Emitter::setPosition(const Vector3& p)
//------------------------------------------------------------------------------
{
    mPosition = p;
}

//------------------------------------------------------------------------------
void Emitter::setSpeed(const Vector3& s)
//------------------------------------------------------------------------------
{
    GE_ASSERT(!s.isAlmost(Vector3::ORIGIN()));
    mSpeed = s;
    mSpeedLength = s.length();
    mSpeedNormalized = s*(1/mSpeedLength);
    mSpeedQuaternion.makeRotation(Vector3(0,1,0), mSpeedNormalized);
}

//------------------------------------------------------------------------------
void Emitter::setSpread(Float s)
//------------------------------------------------------------------------------
{
    GE_ASSERT(s>=0);
    clamp(s, 0.0f, PI);
    mSpread = s;
}


//------------------------------------------------------------------------------
void Emitter::set(Material& m)
//------------------------------------------------------------------------------
{
    mMaterial = &m;
}


//------------------------------------------------------------------------------
void Emitter::set(const Color& start, const Color& end)
//------------------------------------------------------------------------------
{
    mColor.first = start;
    mColor.second = end;
}

//------------------------------------------------------------------------------
const std::pair<Color, Color>& Emitter::getColor() const
//------------------------------------------------------------------------------
{
    return mColor;
}

//------------------------------------------------------------------------------
void Emitter::setSize(Float start, Float end)
//------------------------------------------------------------------------------
{
    GE_ASSERT(start>0 && end>0);
    mSize.first = start;
    mSize.second = end;
}

//------------------------------------------------------------------------------
const std::pair<Float, Float>& Emitter::getSize() const
//------------------------------------------------------------------------------
{
    return mSize;
}

//------------------------------------------------------------------------------
void Emitter::setAngle(Float start, Float end)
//------------------------------------------------------------------------------
{
    mAngle.first = start;
    mAngle.second = end;
}

//------------------------------------------------------------------------------
const std::pair<Float, Float>& Emitter::getAngle() const
//------------------------------------------------------------------------------
{
    return mAngle;
}


//------------------------------------------------------------------------------
void Emitter::update()
//------------------------------------------------------------------------------
{
    if (mParticles == 0)
    {
        return;
    }

    UInt i = 0;

    // Update existing particles
    for (i = 0; i<mParticleCount; ++i)
    {
        mParticles[i].update();
    }

    mCreateCount += mRate*theScaledMetronome().getDelta()/1000;
    UInt createThisFrame = UInt(mCreateCount);
    mCreateCount -= createThisFrame;
    Int p = getNextParticle(0, false);
    while (createThisFrame > 0)
    {
        if (p >= 0)
        {
            GE_ASSERT(p<mParticleCount);
            
            // Calculate initial speed
            Float theta = frand(0, mSpread);
            Float phi = frand(0, TWO_PI);
            Float sinTheta = sin(theta);
            Vector3 speed(cos(phi)*sinTheta, cos(theta), sin(phi)*sinTheta);
            Vector3 particleSpeed;
            mSpeedQuaternion.transform(particleSpeed, speed);
            particleSpeed *= mSpeedLength;

            // Init particle
            mParticles[p].init(mPosition, particleSpeed, this);

            // Step to next particle
            --createThisFrame;
            p = getNextParticle(p+1, false);
        }
        else
        {
            createThisFrame = 0;
        }
    }
}

//------------------------------------------------------------------------------
const Material& Emitter::getMaterial() const
//------------------------------------------------------------------------------
{
    return *mMaterial;
}

//------------------------------------------------------------------------------
Int Emitter::getNextParticle(UInt current, bool alive) const
//------------------------------------------------------------------------------
{
    if (mParticles != 0)
    {
        for (UInt i = current; i<mParticleCount; ++i)
        {
            if (mParticles[i].isAlive() == alive)
            {
                return i;
            }
        }
    }

    return -1;
}

} // namespace Ge
