#ifndef ENGINE_EMITTER_H
#define ENGINE_EMITTER_H
#pragma once

//------------------------------------------------------------------------------
// (c) Aug-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Material.h"
#include "Color.h"

#include "Material.h"

#include <Math/Vector.h>
#include <Math/Quaternion.h>

#include <utility>

namespace Ge
{

class Particle;

// ToDo: If emitters can be bound to objects (e.g. space ship exhaust)
// then later add support to position them with locators.
class Emitter
{
public:
    class ConstIterator
    {
    public:
        const Particle& operator *() const;

        const Particle* operator ->() const;

        // Prefix
        ConstIterator& operator ++();

        bool operator ==(const ConstIterator&) const;

        bool operator !=(const ConstIterator&) const;

    private:
        friend class Emitter;

        const Emitter* mEmitter;
        Int mCurrent;
    };

    // Initial values:
    // - speed (0,1,0)
    // - spread = 0
    // - material = default Material
    // - size = 1

    Emitter();

    Emitter(Float rate, Float lifetime);

    ~Emitter();

    ConstIterator getBegin() const;

    ConstIterator getEnd() const;

    // Initial position of a Particle.
    void setPosition(const Vector3&);

    // Initial speed of a Particle.
    void setSpeed(const Vector3&);

    // Particle spread in radians.
    void setSpread(Float);
    
    // Material of generated Particles.
    void set(Material&);

    const Material& getMaterial() const;

    // Color of generated Particles (including start and end alpha)
    void set(const Color& start, const Color& end);

    const std::pair<Color, Color>& getColor() const;

    // Size of particle quad.
    void setSize(Float start, Float end);

    const std::pair<Float, Float>& getSize() const;

    // Angular rotation of quad.
    void setAngle(Float start, Float end);

    const std::pair<Float, Float>& getAngle() const;

    // Set the ammount of particles for this Emitter defined by
    // - rate: Particles per second.
    // - lifetime: Time to live for one particle.
    // All currently active particles for this Emitter are destroyed!
    void setAmmount(Float rate, Float lifetime);

    Float getLifetime() const;

    void update();

private:
    friend class ConstIterator;

    // Returns the next particle in the array that is <alive>, starting the search
    // at <current>. If no free particle is found then -1 is returned.
    Int getNextParticle(UInt current, bool alive) const;

    Float mRate;
    Vector3 mPosition;
    Vector3 mSpeed;
    Vector3 mSpeedNormalized;
    Float mSpeedLength;
    Quaternion mSpeedQuaternion;
    Float mSpread;
    Float mLifetime;
    Pointer<Material> mMaterial;
    std::pair<Float, Float> mSize;
    std::pair<Float, Float> mAngle;
    std::pair<Color, Color> mColor;
    Particle* mParticles;
    UInt mParticleCount;
    Float mCreateCount;
};

} // namespace Ge

#endif // !ENGINE_EMITTER_H