#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jul-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "FpsCameraController.h"

#include <Input/Mouse.h>

#include <Math/Matrix.h>

#include <Engine/Camera.h>

#include <Application/Metronome.h>

namespace
{
    const Ge::Float ROTATION_MULTIPLIER = 0.01f;
} // namespace





namespace Ge
{

//------------------------------------------------------------------------------
FpsCameraController::FpsCameraController(Camera& camera) :
//------------------------------------------------------------------------------
mSensitivity(1),
mUnitsPerSecond(2),
mCamera(camera),
mA(Keyboard::A),
mD(Keyboard::D),
mW(Keyboard::W),
mS(Keyboard::S),
mCursorLeft(Keyboard::LEFT),
mCursorRight(Keyboard::RIGHT),
mCursorUp(Keyboard::UP),
mCursorDown(Keyboard::DOWN),
mLeft("Left", mCursorLeft),
mRight("Right", mCursorRight),
mForward("Forward", mCursorUp),
mBackward("Backward", mCursorDown),
mC(Keyboard::C),
mCtrl(Keyboard::LEFT_CONTROL),
mSpace(Keyboard::SPACE),
mRightMouse(Mouse::RIGHT),
mCrouch("Crouch", mC),
mJump("Jump", mSpace)
{
    mLeft.map(mA);
    mRight.map(mD);
    mForward.map(mW);
    mBackward.map(mS);
    mCrouch.map(mCtrl);
    mJump.map(mRightMouse);

    setInvert(false);
}

//------------------------------------------------------------------------------
void FpsCameraController::update()
//------------------------------------------------------------------------------
{
    //------------------------------------------------------------------------------
    // Update rotation
    //------------------------------------------------------------------------------

    Vector2 pos;
    theMouse().getDelta(pos);
    pos *= mSensitivity;
    
    if (pos != Vector2::ORIGIN())
    {
        Matrix3x3 currentRotation;
        mCamera.getLocator().getGlobal().getSubMatrix(currentRotation);

        // Calculate current angle to y-axis
        Vector3 currentAt;
        mCamera.getLocator().getGlobal().getColumn(MatrixBase::AT, currentAt);
        Float currentAngle = currentAt.getAngleTo(Vector3(0,1,0));

        // If rotation would rotate over zenith then don't do up/down
        // part of the rotation.
        Matrix3x3 upDown;
        if (currentAngle + mInvert*pos[1]*ROTATION_MULTIPLIER < 0 ||
            currentAngle + mInvert*pos[1]*ROTATION_MULTIPLIER > PI)
        {
            upDown = Matrix3x3::IDENTITY();
        }
        else
        {
            makeXRotation(upDown, mInvert*pos[1]*ROTATION_MULTIPLIER);
        }

        Matrix3x3 leftRight;
        makeYRotation(leftRight, mInvert*pos[0]*ROTATION_MULTIPLIER);

        mCamera.getLocator().get().setSubMatrix(leftRight*currentRotation*upDown);
        mCamera.getLocator().update();
    }
    
    //------------------------------------------------------------------------------
    // Update translation
    //------------------------------------------------------------------------------

    mLeft.update();
    mRight.update();
    mForward.update();
    mBackward.update();
    mCrouch.update();
    mJump.update();

    Vector3 currentPos;
    mCamera.getLocator().getGlobal().getColumn(MatrixBase::TRANSLATION, currentPos);
    Vector3 currentRight;
    mCamera.getLocator().getGlobal().getColumn(MatrixBase::RIGHT, currentRight);
    Vector3 currentAt;
    mCamera.getLocator().getGlobal().getColumn(MatrixBase::AT, currentAt);
    
    Vector3 delta(0,0,0);
    Float factor = mUnitsPerSecond*(theMetronome().getDelta()/1000.0f);

    if (mLeft.isPressed())
    {    
        delta -= currentRight*factor;
    }

    if (mRight.isPressed())
    {    
        delta += currentRight*factor;
    }

    if (mForward.isPressed())
    {    
        delta += currentAt*factor;
    }

    if (mBackward.isPressed())
    {    
        delta -= currentAt*factor;
    }
    
    if (mCrouch.isPressed())
    {
        delta -= Vector3(0, factor, 0);
    }

    if (mJump.isPressed())
    {
        delta += Vector3(0, factor, 0);
    }

    if (delta != Vector3::ORIGIN())
    {
        mCamera.getLocator().get().setColumn(MatrixBase::TRANSLATION,
                                             currentPos+delta);
        mCamera.getLocator().update();
    }
}

//------------------------------------------------------------------------------
bool FpsCameraController::isInverted() const
//------------------------------------------------------------------------------
{
    return mInvert == -1 ? true : false;
}

//------------------------------------------------------------------------------
void FpsCameraController::setInvert(bool invert)
//------------------------------------------------------------------------------
{
    mInvert = invert ? -1 : 1;
}

//------------------------------------------------------------------------------
void FpsCameraController::setSensitivity(Float s)
//------------------------------------------------------------------------------
{
    mSensitivity = s;
}

//------------------------------------------------------------------------------
Float FpsCameraController::getSensitivity() const
//------------------------------------------------------------------------------
{
    return mSensitivity;
}

} // namespace Ge
