//------------------------------------------------------------------------------
// (c) 06-2001 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef ENGINE_HIERARCHY_ANIMATION_H
#define ENGINE_HIERARCHY_ANIMATION_H

#include "KeyframeController.h"

#include <System/Base.h>
#include <System/Pointer.h>
#include <System/Timer.h>

#include <string>

namespace Ge
{

// A hierarchy animation is a collection of KeyframeControllers that can be
// attached to a whole subtree. Every controller has to be identified by a
// prefix that is searched for, when the subtree is traversed. A controller
// will be attached to the first instance of a Locator in the subtree with a matching
// prefix -> so no multiple nodes with matching prefixes are allowed to exist.
class HierarchyAnimation
{
public:
    // Every HierarchyAnimation has a <name> like WALK, IDLE, ...
    HierarchyAnimation(const char* name);

    // Resets all attached controllers to the new offset <time>
    void reset(const Timer::Unit& time);
    
    // Ticks all AudioVisuals with controllers attached.
    // Returns false is the animation is over
    bool update(const Timer::Unit&);
    
    const Timer::Unit& getDuration() const;

    // Add a KeyframeController plus the <prefix> of the node it should be
    // attached to.
    void add(KeyframeController&, const char* prefix);

    void set(TimeController::Repeat);
    
    TimeController::Repeat getRepeatType() const;

    void attachTo(Locator&);

private:
    friend class AudioVisual;
    // Attaches all KeyframeControllers that match the AudioVisual's name
    void attachMatchingControllers(Locator&);

    // When a new KeyframeController was added recalculate the duration of the
    // whole animation.
    void recalcDuration(KeyframeController&);

    // Clear mWasMapped from all Entries
    void unmapAll();

    struct Entry
    {
        Entry() : mWasMapped(false) {}
        std::string mPrefix;
        Pointer<KeyframeController> mController;
        bool mWasMapped;
    };

    std::string mName;
    std::vector<Entry> mEntries;

    TimeController::Repeat mRepeatType;
    Timer::Unit mDuration; // The duration of this animation
    Timer::Unit mOffset;
};

} // namespace Ge

#endif // ENGINE_HIERARCHY_ANIMATION_H
