#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jul-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Locator.h"

namespace
{

using namespace Ge;

// This function creates a diagonal scaling matrix from an (x,y,z) scaling
// vector <v> and preapplies it to <m>.
template<typename T> inline
//------------------------------------------------------------------------------
void preScale(BasicMatrix<T, 3, 4>& m, const BasicVector<T, 3>& v)
//------------------------------------------------------------------------------
{
    m(0, 0) *= v[0];
    m(0, 1) *= v[0];
    m(0, 2) *= v[0];
    m(1, 0) *= v[1];
    m(1, 1) *= v[1];
    m(1, 2) *= v[1];
    m(2, 0) *= v[2];
    m(2, 1) *= v[2];
    m(2, 2) *= v[2];
}

} // namespace




namespace Ge
{

//------------------------------------------------------------------------------
// Locator::Iterator members
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
Locator::Iterator::Iterator()
//------------------------------------------------------------------------------
{
}

//------------------------------------------------------------------------------
Locator& Locator::Iterator::operator *()
//------------------------------------------------------------------------------
{
    GE_ASSERT(mCurrent != 0);
    return *mCurrent;
}

//------------------------------------------------------------------------------
Locator* Locator::Iterator::operator ->()
//------------------------------------------------------------------------------
{
    GE_ASSERT(mCurrent != 0);
    return mCurrent;
}

//------------------------------------------------------------------------------
Locator::Iterator& Locator::Iterator::operator ++()
//------------------------------------------------------------------------------
{
    // The current node has children
    if (!mCurrent->mChildren.empty())
    {
        mCurrent = mCurrent->mChildren[0];
        mStack.push(0);
    }
    // The current node has no children but the stack is empty. This means that
    // the starting node had no children.
    else if (mStack.empty())
    {
        mCurrent = 0;
    }
    else
    {
        UInt index = 0;
        bool foundNextChild = false;
        while (!mStack.empty() && !foundNextChild)
        {
            index = mStack.top();
            mStack.pop();
            Locator* parent = mCurrent->mParent;
            
            GE_ASSERT_MSG(parent != 0,
                "This case should never occur because the stack should be empty before.");

            // Current node is not last node in parent's child list.
            if (index < (parent->mChildren.size()-1))
            {
                ++index;
                mCurrent = parent->mChildren[index];
                mStack.push(index);
                foundNextChild = true;
            }
            // Current node was last node in parent's child list. Go further up.
            else
            {
                mCurrent = parent;
            }
        }

        // Reached the start = end node after traversal.
        if (mStack.empty())
        {
            mCurrent = 0;
        }
    }

    return *this;
}


//------------------------------------------------------------------------------
bool Locator::Iterator::operator ==(const Iterator& rhs) const
//------------------------------------------------------------------------------
{
    return mCurrent == rhs.mCurrent && mStack == rhs.mStack;
}


//------------------------------------------------------------------------------
bool Locator::Iterator::operator !=(const Iterator& rhs) const
//------------------------------------------------------------------------------
{
    return !((*this) == rhs);
}






//------------------------------------------------------------------------------
// Locator members
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
Locator* Locator::create(const Matrix3x4& m, const char* name)
//------------------------------------------------------------------------------
{
    Locator* l = GE_NEW Locator(name);
    l->get() = m;
    l->update();
    return l;
}

//------------------------------------------------------------------------------
Locator::Locator(const char* name) :
//------------------------------------------------------------------------------
mParent(0),
mScale(1,1,1)
{
    if (name != 0)
    {
        mName = name;
    }
}

//------------------------------------------------------------------------------
Locator::~Locator()
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<mChildren.size(); ++i)
    {
        mChildren[i]->mParent = 0;
    }
}

//------------------------------------------------------------------------------
void Locator::setName(const char* name)
//------------------------------------------------------------------------------
{
    mName = name;
}

//------------------------------------------------------------------------------
const std::string& Locator::getName() const
//------------------------------------------------------------------------------
{
    return mName;
}

//------------------------------------------------------------------------------
Locator::Iterator Locator::getBegin()
//------------------------------------------------------------------------------
{
    Iterator it;
    it.mCurrent = this;
    return it;
}

//------------------------------------------------------------------------------
Locator::Iterator Locator::getEnd() const
//------------------------------------------------------------------------------
{
    Iterator it;
    it.mCurrent = 0;
    return it;
}

//------------------------------------------------------------------------------
Matrix3x4& Locator::get()
//------------------------------------------------------------------------------
{
    return mLocal;
}

//------------------------------------------------------------------------------
const Matrix3x4& Locator::get() const
//------------------------------------------------------------------------------
{
    return mLocal;
}

//------------------------------------------------------------------------------
void Locator::setScale(const Vector3& s)
//------------------------------------------------------------------------------
{
    mScale = s;
}

//------------------------------------------------------------------------------
const Vector3& Locator::getScale() const
//------------------------------------------------------------------------------
{
    return mScale;
}


//------------------------------------------------------------------------------
const Matrix3x4& Locator::getGlobal() const
//------------------------------------------------------------------------------
{
    return mGlobal;
}

//------------------------------------------------------------------------------
void Locator::addChild(Locator& l)
//------------------------------------------------------------------------------
{
    l.mParent = this;
    mChildren.push_back(Pointer<Locator>(&l));
}

//------------------------------------------------------------------------------
void Locator::update()
//------------------------------------------------------------------------------
{
    if (mParent != 0)
    {
        mGlobal = mParent->mGlobal*mLocal;
    }
    else
    {
        mGlobal = mLocal;   
    }

    preScale(mGlobal, mScale);

    for (UInt i = 0; i<mChildren.size(); ++i)
    {
        mChildren[i]->update();
    }
}

//------------------------------------------------------------------------------
Locator* Locator::clone() const
//------------------------------------------------------------------------------
{
    Locator* ret = GE_NEW Locator;
    ret->get() = get();
    ret->mScale = mScale;
    ret->mName = mName;
    for (UInt i = 0; i<mChildren.size(); ++i)
    {
        ret->addChild(*mChildren[i]->clone());
    }

    ret->update();

    return ret;
}

//------------------------------------------------------------------------------
const Locator* Locator::getParent() const
//------------------------------------------------------------------------------
{
    return mParent;
}

} // namespace Ge
