#ifndef ENGINE_LOCATOR_H
#define ENGINE_LOCATOR_H
#pragma once

//------------------------------------------------------------------------------
// (c) Jul-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include <Math/Matrix.h>

#include <System/Object.h>
#include <System/Pointer.h>

#include <vector>
#include <stack>

namespace Ge
{

class Locator;

// Locators are used to place objects in the world. They can have children
// and form a hierarchical tree of Locators.
class Locator : public Object
{
public:
    // This Iterator iterates depth first through the hierarchy,
    // starting from this Locator.
    class Iterator
    {
    public:
        Locator& operator *();

        Locator* operator ->();

        // Prefix
        Iterator& operator ++();

        bool operator ==(const Iterator&) const;

        bool operator !=(const Iterator&) const;

    private:
        friend class Locator;
        Iterator();

        std::stack<UInt> mStack;
        Locator* mCurrent;
    };

    // Create a Locator on the heap with local and global set to <m>.
    static
    Locator* create(const Matrix3x4& m, const char* name = 0);

    // No initialisation is done! Use create instead if you want initialisation.
    // The <name> can be set optionally for hierarchies, where you want to
    // find a specific Locator in the hierarchy (e.g. find the "arm" in an animation).
    Locator(const char* name = 0);

    ~Locator();

    Iterator getBegin();

    Iterator getEnd() const;

    void setName(const char*);

    const std::string& getName() const;

    // Returns the local transformation
    Matrix3x4& get();

    // Returns the local transformation
    const Matrix3x4& get() const;

    // Set scale factor
    void setScale(const Vector3&);

    // Returns the scale factor
    const Vector3& getScale() const;

    // Returns the global transformation
    const Matrix3x4& getGlobal() const;

    // <l> must live on the heap.
    void addChild(Locator& l);

    // Update the global transformation recursively starting with this
    // locator.
    void update();

    // Create a copy of the complete subtree starting from this Locator.
    Locator* clone() const;

    const Locator* getParent() const;

private:
    friend class Iterator;

    Matrix3x4 mLocal;
    Vector3 mScale;
    Matrix3x4 mGlobal;
    Locator* mParent;
    std::vector<Pointer<Locator> > mChildren;
    std::string mName;
};

} // namespace Ge

#endif // !ENGINE_LOCATOR_H