#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 06-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include "LwsImporter.h"
#include "Cluster.h"
#include "Vector3Key.h"
#include "ModelPool.h"

#include <System/Base.h>
#include <System/FilePath.h>
#include <System/StringTools.h>
#include <System/Log.h>


namespace
{

//------------------------------------------------------------------------------
Ge::Int extractParentId(const std::string& lwsId)
//------------------------------------------------------------------------------
{
    std::string::size_type firstZero = lwsId.find_first_of('0');
    GE_ASSERT(firstZero != std::string::npos);
    std::string::size_type firstNonZero = lwsId.find_first_not_of('0', firstZero);
    Ge::Int ret = 0;
    if (std::string::npos != firstNonZero)
    {
        ret = Ge::StringTools::hexToInt(lwsId.substr(firstNonZero));
    }

    return ret;
}

} // namespace




namespace Ge
{

//------------------------------------------------------------------------------
LwsImporter::LwsImporter(const char* fileName, ModelPool& modelPool) :
//------------------------------------------------------------------------------
mModelPool(modelPool),
//mAnimationSet(0),
mCurrentEntry(0)
{
    bool ok = fileName != 0;
    GE_ASSERT_MSG(ok, "No file name provided");

    ok = (ok && FilePath(fileName).getExtension() == ".lws");
    GE_ASSERT_MSG(ok, "Wrong extension");

    if (ok)
    {
        std::ifstream file(fileName, std::ios::in);
        if (file)
        {
            if (verifyVersion(file))
            {
                while (waitForTag(file))
                {
                    while ("LoadObjectLayer" == mCurrentLine.getTag())
                    {
                        mEntries.push_back(Entry());
                        readObject(fileName);
                        readChannels(file);
                        readParent(file);
                        mCurrentEntry++;
                    }
//                    else
//                        GE_ASSERT_MSG(false, ".lws file corrupt?");

                    mCurrentLine.clear();
                }
            }
        }
        else
        {
            GE_ASSERT_MSG(false, ".lws file couldn't be opened");
        }
    }
}

//------------------------------------------------------------------------------
Cluster* LwsImporter::importModel() const
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(mEntries.size() > 0, ".lws file contained no models");
    Cluster* ret = 0;
    if (mEntries.size() > 1)
    {
        createHierarchyMap();
        ret = buildSubTree();
    }
    else if (mEntries.size() == 1)
    {
        ret = mEntries[0].mObject;
    }
    
    return ret;
}

//------------------------------------------------------------------------------
HierarchyAnimation LwsImporter::importAnimation() const
//------------------------------------------------------------------------------
{
    // All controllers with positive key count are added to the
    // HierarchyAnimation;
    HierarchyAnimation animation("NoName");
    PositionController* pc = 0;
    RotationController* rc = 0;
    ScaleController* sc = 0;
    for (UInt i = 0; i<mEntries.size(); i++)
    {
        if (mEntries[i].mPositionController.getKeyCount() > 0)
        {
            pc = GE_NEW PositionController(mEntries[i].mPositionController);
            animation.add(*pc, mEntries[i].mObject->getLocator()->getName().c_str());
        }

        if (mEntries[i].mRotationController.getKeyCount() > 0)
        {
            rc = GE_NEW RotationController(mEntries[i].mRotationController);
            animation.add(*rc, mEntries[i].mObject->getLocator()->getName().c_str());
        }

        if (mEntries[i].mScaleController.getKeyCount() > 0)
        {
            sc = GE_NEW ScaleController(mEntries[i].mScaleController);
            animation.add(*sc, mEntries[i].mObject->getLocator()->getName().c_str());
        }
    }

    return animation;
}


//------------------------------------------------------------------------------
void LwsImporter::createHierarchyMap() const
//------------------------------------------------------------------------------
{
    // Add all entries into the hierarchy map first
    for (UInt i = 0; i<mEntries.size(); i++)
    {
        mHierarchyMap[i] = mEntries[i].mObject;
    }

//    assureParentsAreNodes();

    // Add children to parents
    Int parentId(0);    
    for (i = 0; i<mEntries.size(); i++)
    {
        parentId = mEntries[i].mParent;
        // Does this entry have a parent? If yes -> make it a child of its parent
        if (parentId != -1)
        {
//            Node* parent = mHierarchyMap[parentId]->asNode();
            GE_ASSERT(mHierarchyMap[parentId]->getLocator() != 0);

            GE_LOG_DEBUG(mHierarchyMap[i]->getLocator()->getName()<<"("<<
                      mHierarchyMap[i]->getLocator()<<")"<<
                      " is child of "<<mHierarchyMap[parentId]->getLocator()->getName()<<"("<<
                      mHierarchyMap[parentId]->getLocator()<<")"<<std::endl);

            // Add this entry as a child to the parent
//            parent->addChild(*mHierarchyMap[i]);
            
            mHierarchyMap[parentId]->getLocator()->addChild(*mHierarchyMap[i]->getLocator());
        }
    }
}

////------------------------------------------------------------------------------
//void LwsImporter::assureParentsAreNodes() const
////------------------------------------------------------------------------------
//{
//    // Make all entries that are parents and not Nodes to Nodes and attach
//    // themselves as children.
//    Int parentId(0);
//    for (UInt i = 0; i<mEntries.size(); i++)
//    {
//        parentId = mEntries[i].mParent;
//        // Does this entry have a parent?
//        if (parentId != -1)
//        {
//            // Is the parent a Node?
//            Node* parent = mHierarchyMap[parentId]->asNode();
//            if (parent == 0)
//            {
//                // The parent is not a Node -> create a Node and add parent as
//                // child
//                parent = GE_NEW Node((mHierarchyMap[parentId]->getName()+"Node").c_str());
//                GE_LOG_DEBUG("Created new node: "<<parent->getName()<<std::endl);
//                parent->addChild(*mHierarchyMap[parentId]);
//                mHierarchyMap[parentId] = parent;
//            }
//        }
//    }
//}

//------------------------------------------------------------------------------
Cluster* LwsImporter::buildSubTree() const
//------------------------------------------------------------------------------
{
    Cluster* ret = 0;
    ret = GE_NEW Cluster;

    if (getRootObjectCount() >= 2)
    {
        // Make new root node and add all root nodes from the hierarchy map as
        // children (a root node is a node with getParent() == 0)    
        ret->set(*Locator::create(Matrix3x4::IDENTITY()));
        for (HierarchyMap::iterator j = mHierarchyMap.begin();
             j != mHierarchyMap.end();
             j++)
        {
            bool isRootNode = (j->second->getLocator()->getParent() == 0);
            if (isRootNode)
            {
                ret->getLocator()->addChild(*j->second->getLocator());
            }
        }
    }
    else
    {
        // Look for the one and only root node and return it
        for (HierarchyMap::iterator j = mHierarchyMap.begin();
             j != mHierarchyMap.end(); j++)
        {
            bool isRootNode(j->second->getLocator()->getParent() == 0);
            if (isRootNode)
            {
                ret->set(*j->second->getLocator());
                break;
            }
        }
    }

    // Copy links to new Cluster.
    for (HierarchyMap::iterator j = mHierarchyMap.begin();
         j != mHierarchyMap.end(); ++j)
    {
        const Cluster::Links& links = j->second->getLinks();
        for (UInt i = 0; i<links.size(); ++i)
        {
            ret->add(*links[i]);
        }
    }

    return ret;
}

//------------------------------------------------------------------------------
UInt LwsImporter::getRootObjectCount() const
//------------------------------------------------------------------------------
{
    UInt ret(0);
    for (HierarchyMap::iterator j = mHierarchyMap.begin();
         j != mHierarchyMap.end(); j++)
    {
        bool isRootNode(j->second->getLocator()->getParent() == 0);
        if (isRootNode)
        {
            ++ret;
        }
    }

    return ret;
}

//------------------------------------------------------------------------------
bool LwsImporter::verifyVersion(std::istream& file)
//------------------------------------------------------------------------------
{
    mCurrentLine.read(file);
    bool ret("LWSC" == mCurrentLine.getTag());
    if (ret)
    {
        mCurrentLine.read(file);
        ret = ("3" == mCurrentLine.getTag());
    }

    mCurrentLine.clear();
    GE_ASSERT_MSG(ret, "Only .lws version 3 can be imported");
    return ret;
}

//------------------------------------------------------------------------------
bool LwsImporter::waitForTag(std::istream& file)
//------------------------------------------------------------------------------
{
    while (!file.eof() &&
           "LoadObjectLayer" != mCurrentLine.getTag() &&
           "Channel" != mCurrentLine.getTag() &&
           "Key" != mCurrentLine.getTag() &&
           "{" != mCurrentLine.getTag() &&
           "ParentItem" != mCurrentLine.getTag())
    {
           mCurrentLine.read(file);
    }

    // !file.eof() means a tag was found
    return !file.eof();
}

//------------------------------------------------------------------------------
void LwsImporter::readObject(const char* fileName)
//------------------------------------------------------------------------------
{
    // Read the .obj file using the ObjImporter
    mCurrentLine.split();
    FilePath path(mCurrentLine.getRest().c_str());
    FilePath lwsFile(fileName);
    path.setDirectory(lwsFile.getDirectory().c_str());
    Cluster* obj = mModelPool.get(path.get().c_str());
    GE_ASSERT(obj != 0);
    if (obj != 0)
    {
        // Append entry number to link name so that if an animation
        // uses two objects with the same file name but different animation
        // keys the hierarchy is set up correctly.
        std::string name = path.getName()+StringTools::toString(mCurrentEntry);
        obj->getLocator()->setName(name.c_str());
        GE_TRACE("Imported "<<name<<std::endl);
        mEntries[mCurrentEntry].mObject = obj;
    }

    mCurrentLine.clear();
}

//------------------------------------------------------------------------------
void LwsImporter::readChannels(std::istream& file)
//------------------------------------------------------------------------------
{
    bool stillChannelsLeft(true);
    while (waitForTag(file) && stillChannelsLeft)
    {
        if ("Channel" == mCurrentLine.getTag() &&
            "0" == mCurrentLine.getRest())
        {
            readKeys(file, mEntries[mCurrentEntry].mPositionController);
        }
        else if ("Channel" == mCurrentLine.getTag() &&
                 "3" == mCurrentLine.getRest())
        {
            readKeys(file, mEntries[mCurrentEntry].mRotationController);
        }
        else if ("Channel" == mCurrentLine.getTag() &&
                 "6" == mCurrentLine.getRest())
        {
            readKeys(file, mEntries[mCurrentEntry].mScaleController);
            stillChannelsLeft = false;
        }
        else
        {
            GE_ASSERT_MSG(false, "Channels expected after loading .obj");
        }
    }
}

//------------------------------------------------------------------------------
void LwsImporter::readKeys(std::istream& file,
                           Vector3Controller& controller)
//------------------------------------------------------------------------------
{
    // Buffer values, because x,y,z channels are split up
    std::map<Timer::Unit, Vector3> valueBuffer;
    std::map<Timer::Unit, Vector3> tInBuffer;
    std::map<Timer::Unit, Vector3> tOutBuffer;
    UInt keyCount(0);
    Timer::Unit time(0);
    Float value(0), tIn(0), tOut(0);
    // Denotes reading the x, y, z coordinate
    UInt currentIndex(0);
    
    // Read 3 channels 0,1,2 (x,y,z or h,p,r channels).
    while (waitForTag(file) && currentIndex<3)
    {
        if ("{" == mCurrentLine.getTag() &&
            "Envelope" == mCurrentLine.getRest())
        {
            // Verify that x,y,z channels all have the same key count
            mCurrentLine.read(file);
            if (keyCount == 0)
                keyCount = StringTools::toInt(mCurrentLine.getTag());
            else
                GE_ASSERT(keyCount == StringTools::toInt(mCurrentLine.getTag()));

            // Read the keys for one channel
            while (waitForTag(file) &&
                   "Key" == mCurrentLine.getTag())
            {
                readKey(value, time, tIn, tOut);
                valueBuffer[time][currentIndex] = value;
                tInBuffer[time][currentIndex] = tIn;
                tOutBuffer[time][currentIndex] = tOut;
            }
            currentIndex++;
        }
        else if ("Channel" == mCurrentLine.getTag())
        {
            mCurrentLine.clear();
        }
        else
        {
            GE_ASSERT_MSG(false, "Wrong tag encountered during key import");
        }
    }

    copyBuffersToController(valueBuffer, tInBuffer, tOutBuffer, controller);
}

//------------------------------------------------------------------------------
void LwsImporter::readParent(std::istream& file)
//------------------------------------------------------------------------------
{
    if (waitForTag(file))
    {
        if ("ParentItem" == mCurrentLine.getTag())
        {
            mEntries[mCurrentEntry].mParent =
            extractParentId(mCurrentLine.getRest());
        }
    }
}

//------------------------------------------------------------------------------
void LwsImporter::copyBuffersToController(std::map<Timer::Unit, Vector3>& values,
                                          std::map<Timer::Unit, Vector3>& tIn,
                                          std::map<Timer::Unit, Vector3>& tOut,
                                          Vector3Controller& controller)
//------------------------------------------------------------------------------
{
    Vector3Key key;
    for (std::map<Timer::Unit, Vector3>::iterator i = values.begin();
         i != values.end();
         i++)
    {
        // Lightwave exports time in seconds -> *1000 to get ms
        key.setTime(i->first*1000);
        key.setValue(values[i->first]);
        key.setTangentIn(tIn[i->first]);
        key.setTangentOut(tOut[i->first]);
        controller.add(key);
    }
}

//------------------------------------------------------------------------------
void LwsImporter::readKey(Float& value, Timer::Unit& time,
                          Float& tangentIn, Float& tangentOut)
//------------------------------------------------------------------------------
{
    GE_ASSERT(mCurrentLine.getTag() == "Key");
    mCurrentLine.split();

    value = StringTools::toFloat(mCurrentLine.getTag());
    mCurrentLine.split();
    
    time = StringTools::toFloat(mCurrentLine.getTag());
    // Split 5 times, so that tag is at tangentIn
    for (UInt i = 0; i<5; i++)
    {
        mCurrentLine.split();
    }

    tangentIn = StringTools::toFloat(mCurrentLine.getTag());
    mCurrentLine.split();

    tangentOut = StringTools::toFloat(mCurrentLine.getTag());
    mCurrentLine.clear();
}

} // namespace Ge