#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jul-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Material.h"
#include "Link.h"

#include <DirectXRenderer/DirectXRenderer.h>
#include <DirectXRenderer/DirectXHelpers.h>
#include <DirectXRenderer/DirectXTexture.h>

#include <d3d8.h>
#include <d3dx8.h>


namespace Ge
{

//------------------------------------------------------------------------------
Material::Material(UInt passCount) :
//------------------------------------------------------------------------------
mColor(Color::WHITE),
mPassCount(passCount),
mLink(0),
mPreShadow(false),
mPostShadow(false)
{
}

//------------------------------------------------------------------------------
Color Material::getColor() const
//------------------------------------------------------------------------------
{
    return mColor;
}

//------------------------------------------------------------------------------
void Material::set(const Color& c)
//------------------------------------------------------------------------------
{
    mColor = c;
}

//------------------------------------------------------------------------------
const Texture* Material::getTexture(UInt index) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(index < MAX_TEXTURES);
    return mTextures[index].get();
}

//------------------------------------------------------------------------------
void Material::set(UInt index, const Texture& t)
//------------------------------------------------------------------------------
{
    GE_ASSERT(index < MAX_TEXTURES);
    mTextures[index] = &t;
}

//------------------------------------------------------------------------------
void Material::setLink(Link& l)
//------------------------------------------------------------------------------
{
    mLink = &l;
}

//------------------------------------------------------------------------------
void Material::setPreShadow(bool b)
//------------------------------------------------------------------------------
{
    mPreShadow = b;
}

//------------------------------------------------------------------------------
void Material::setPostShadow(bool b)
//------------------------------------------------------------------------------
{
    mPostShadow = b;
}

//------------------------------------------------------------------------------
bool Material::isPreShadow() const
//------------------------------------------------------------------------------
{
    return mPreShadow;
}

//------------------------------------------------------------------------------
bool Material::isPostShadow() const
//------------------------------------------------------------------------------
{
    return mPostShadow;
}

//------------------------------------------------------------------------------
void Material::apply(UInt pass, const Matrix4x4& view) const
//------------------------------------------------------------------------------
{
    // This is just some stupid single texture material for now.
    GE_ASSERT(pass < 1);

    D3DMATERIAL8 d3dMaterial;
    Dx::materialToD3dMaterial(*this, d3dMaterial);

    DirectXRenderer* dxr = reinterpret_cast<DirectXRenderer*>(&theRenderer());
    IDirect3DDevice8& device = dxr->getDirect3dDevice();

    device.SetMaterial(&d3dMaterial);

    if (isPreShadow())
    {
        dxr->set(RenderState::SHADE_FLAT);
        dxr->set(RenderState::BLEND_OFF);
        dxr->setAmbientLight(Color(0.2f,0.2f,0.2f,1));
        dxr->switchLights(false);
    }

    if (isPostShadow())
    {
        dxr->switchLights(true);
        dxr->setAmbientLight(Color::BLACK);
        dxr->set(RenderState::BLEND_ADD);

        dxr->set(RenderState::STENCIL_ON);
        dxr->set(RenderState::DEPTH_READ_ONLY);
        dxr->set(RenderState::DEPTH_COMPARE_EQUAL);
        
        // Write only on positive stencil values 
        device.SetRenderState(D3DRS_STENCILREF, 0);
        device.SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);
        device.SetRenderState(D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP);    
        device.SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
        device.SetRenderState(D3DRS_STENCILFUNC, D3DCMP_EQUAL);
    }

    DX_CHECK(device.SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE));
    DX_CHECK(device.SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE));
    DX_CHECK(device.SetTextureStageState(3, D3DTSS_COLOROP, D3DTOP_DISABLE));
    if (getTexture(0) != 0)
    {
        const DirectXTexture* t = reinterpret_cast<const DirectXTexture*>(getTexture(0));
        DX_CHECK(device.SetTexture(0, t->getDirect3dTexture()));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_CURRENT));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP));
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP));

        switch (t->getFiltering())
        {
        case Texture::POINT:
            device.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_POINT);
            device.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_POINT);
            break;
        case Texture::BILINEAR:
            device.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
            device.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
            break;

        default:
            GE_ASSERT_MSG(false, "Filtering mode unkown");
        }
    }
    else
    {
        DX_CHECK(device.SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE));
    }

    
#ifdef SHADER
//    Matrix4 transform(projection*view*world);
//    Vector4 cameraPos;
//    for (UInt i = 0; i<3; i++)
//    {
//        cameraPos[i] = mCamera->getPosition()[i];
//    }
//    
//    cameraPos[3] = 1;
//
//    device.SetVertexShaderConstant(1, &(world(0,0)), 4);
//    device.SetVertexShaderConstant(5, &(transform(0,0)), 4);
//    device.SetVertexShaderConstant(9, &(cameraPos[0]), 1);
//    device.SetVertexShader(shader);
#else

    if (getLink() != 0)
    {
        if (getLink()->getMesh()->getActiveFrame().getNormals() == 0)
        {
            dxr->set(RenderState::LIGHTING_OFF);
        }

        // Automatic uv generation for sphere maps.
        for (UInt set = 0; set<Mesh::UV_SET_COUNT; ++set)
        {
            switch(getLink()->getMesh()->getUvType(set))
            {
            case Mesh::UV_SPHERE:
                {
                    // Get inverted rotation part of modelview matrix
                    Matrix4x4 world(Matrix4x4::IDENTITY());
                    world.setSubMatrix(getLink()->getLocator()->getGlobal());
                    Matrix4x4 imv(view*world);
                    setColumn(imv, 3, 0.0f,0.0f,0.0f,1.0f);
                    setRow(imv, 3, 0.0f,0.0f,0.0f,1.0f);
                    invertAffine(imv);

                    // Texture transform matrix
                    Matrix4x4 t;
                    Ge::set(t, 0.5f,  0.0f, 0.0f, 0.5f,
                               0.0f,  0.5f, 0.0f, 0.5f,
                               0.0f,  0.0f, 0.0f, 0.0f,
                               0.0f,  0.0f, 0.0f, 1.0f);

                    t = t*imv;

                    // Set texture transform matrix
                    D3DXMATRIX mat;
                    Dx::matrixToD3dMatrix(t, mat);
                    device.SetTransform(
                        D3DTRANSFORMSTATETYPE(D3DTS_TEXTURE0 + set), &mat);
            
                    device.SetTextureStageState(
                        set, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);
            
                    device.SetTextureStageState(
                        set, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR);
                }
                break;

            case Mesh::UV_ARRAY:
                {
                    device.SetTextureStageState(
                        set, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);

                    device.SetTextureStageState(
                        set, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU);
                }
                break;

            case Mesh::UV_NONE:
                break;

            default:
                GE_ASSERT_MSG(false, "uv type not implemented yet.");
            }
        }
    }
#endif
}

//------------------------------------------------------------------------------
UInt Material::getPassCount() const
//------------------------------------------------------------------------------
{
    return isPreShadow() ? 1 : mPassCount;
}

//------------------------------------------------------------------------------
void Material::setPassCount(UInt pc)
//------------------------------------------------------------------------------
{
    mPassCount = pc;
}

//------------------------------------------------------------------------------
Material* Material::clone() const
//------------------------------------------------------------------------------
{
    Material* ret = GE_NEW Material;
    ret->copyDataFrom(*this);
    return ret;
}

//------------------------------------------------------------------------------
void Material::copyDataFrom(const Material& rhs)
//------------------------------------------------------------------------------
{
    mColor = rhs.mColor;
    for (UInt i = 0; i<MAX_TEXTURES; ++i)
    {
        mTextures[i] = rhs.mTextures[i];
    }
    mPassCount = rhs.mPassCount;
}

//------------------------------------------------------------------------------
const Link* Material::getLink() const
//------------------------------------------------------------------------------
{
    return mLink;
}

} // namespace Ge
