#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Aug-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Md3Importer.h"
#include "Md3Types.h"
#include "Md3Model.h"
#include "Cluster.h"
#include "Mesh.h"
#include "TexturePool.h"

#include <Image/Image.h>

#include <Math/Vector.h>

#include <Application/Settings.h>

#include <zipstream.h>

#include <string>
#include <fstream>


namespace
{

using namespace Ge;

//------------------------------------------------------------------------------
bool verifyHeader(const Md3Header& h)
//------------------------------------------------------------------------------
{
    // Check that ID is "IDP3" and file version is 15.
    bool id = h.mFileId == '3PDI';
    GE_ASSERT_MSG(id, "MD3 file ID is not 'IDP3'.");

    bool version = h.mVersion == 15;
    GE_ASSERT_MSG(version, "MD3 file version must be 15.");

    bool size = h.mHeaderSize == sizeof(h);
    GE_ASSERT_MSG(size, "MD3 header size mismatch.");

    return id && version && size;
}

//------------------------------------------------------------------------------
void getMaterialName(const char* skin, std::string& file)
//------------------------------------------------------------------------------
{
    std::string path(skin);
    Int slash = path.find_last_of('/');
    file = path.substr(slash+1);
}

} // namespace




namespace Ge
{

//------------------------------------------------------------------------------
// Helper structs
//------------------------------------------------------------------------------

// .md3 files store submeshes in one file. This struct holds one such submesh
struct SubMesh
{
    SubMesh() : mVertices(0), mUvs(0), mTriangles(0), mSkins(0) {}

    ~SubMesh()
    {
        if (mVertices != 0)
        {
            GE_DELETE_A mVertices;
        }

        if (mTriangles != 0)
        {
            GE_DELETE_A mTriangles;
        }

        if (mSkins != 0)
        {
            GE_DELETE_A mSkins;
        }

        // Uvs are not deleted because they are directly transfered into the
        // resulting mesh.
    }

    std::string mName;

    UInt mVertexCount;
    UInt mFrameCount;
    UInt mTriangleCount;

    Md3Vertex* mVertices;
    Vector2* mUvs;
    Md3Triangle* mTriangles;
    Md3Skin* mSkins;
};

// One body part of a model. Corresponds to one .md3 file (upper/lower/head).
struct Part
{
    Part() : mMeshes(0), mTags(0) {}

    ~Part()
    {
        if (mMeshes != 0)
        {
            GE_DELETE_A mMeshes;
        }
    }

    SubMesh* mMeshes;
    Md3Tag* mTags;
    UInt mTagCount;
    UInt mMeshCount;
    UInt mFrameCount;
};
    
//------------------------------------------------------------------------------
// Implementation class
//------------------------------------------------------------------------------

class Md3Importer::Impl
{
public:
    Impl(const char* path);

    Md3Model* import() const;

private:   
    void convert();

    void convert(Part& part, Md3Model::Part*& modelPart, Locator*& modelLocator);

    void readAnimationCfg(Md3Model::Part* lower, Md3Model::Part* upper);

    bool load(const char* file, Part& part);

    std::string mPath;
    zip::izipfile mPk3;

    Part mLower;
    Part mUpper;
    Part mHead;

    Cluster* mCluster;
    Md3Model* mModel;
};


//------------------------------------------------------------------------------
Md3Importer::Impl::Impl(const char* path) :
//------------------------------------------------------------------------------
mPath(path),
mCluster(0),
mModel(0),
mPk3((theSettings.getVariable("q3a").asString()+"baseq3/pak0.pk3").c_str())
{
    if (mPk3.isOk())
    {
        bool ok = load("lower.md3", mLower);
        if (ok)
        {
            ok = load("upper.md3", mUpper);  
        }

        if (ok)
        {
            ok = load("head.md3", mHead);
        }

        if (ok)
        {
            convert();
        }
    }
}



//------------------------------------------------------------------------------
bool Md3Importer::Impl::load(const char* file, Part& part)
//------------------------------------------------------------------------------
{
    zip::izipstream stream(mPk3, (mPath+'/'+file).c_str(), std::ios_base::binary);
    bool ok = stream.good();
    GE_ASSERT(ok);
    
    // Verify header.
    Md3Header header;
    if (ok)
    {
        stream.read(reinterpret_cast<char*>(&header), sizeof(header));
        ok = verifyHeader(header);
    }   
    
    // Read bones. (Currently they are just skipped).
    if (ok)
    {
        stream.ignore(sizeof(Md3Bone)*header.mFrameCount);
        ok = stream.good();
    }

    // Read joint animation data.
    if (ok)
    {
        part.mTags = GE_NEW Md3Tag[header.mFrameCount*header.mTagCount];
        part.mTagCount = header.mTagCount;
        part.mFrameCount = header.mFrameCount;

        stream.read(reinterpret_cast<char*>(part.mTags),
                    sizeof(Md3Tag)*header.mFrameCount*header.mTagCount);

#ifndef NDEBUG
        for (UInt i = 0; i<header.mTagCount; ++i)
        {
            GE_TRACE(part.mTags[i].mName<<"\n");
        }
#endif // !NDEBUG

        ok = stream.good();
    }   

    // Read meshes.
    if (ok)
    {
        part.mMeshes = GE_NEW SubMesh[header.mMeshCount];
        part.mMeshCount = header.mMeshCount;
        std::ifstream::pos_type current = stream.tellg();
        for (UInt i = 0; i<header.mMeshCount; ++i)
        {
            // Read mesh header.
            Md3MeshInfo info;
            stream.seekg(current);
            stream.read(reinterpret_cast<char*>(&info), sizeof(info));
            GE_ASSERT_MSG(info.mHeaderSize == sizeof(info),
                          "MD3 mesh header size mismatch.");
            
            // Copy header info.
            part.mMeshes[i].mVertexCount = info.mVertexCount;
            part.mMeshes[i].mTriangleCount = info.mTriangleCount;
            part.mMeshes[i].mFrameCount = info.mFrameCount;
            part.mMeshes[i].mName = reinterpret_cast<const char*>(info.mName);
            
            // Read skins.
            part.mMeshes[i].mSkins = GE_NEW Md3Skin[info.mSkinCount];
            stream.read(reinterpret_cast<char*>(part.mMeshes[i].mSkins),
                        sizeof(Md3Skin)*info.mSkinCount);

            // Read triangles.
            stream.seekg(current + std::ifstream::pos_type(info.mTriStart));
            part.mMeshes[i].mTriangles = GE_NEW Md3Triangle[info.mTriangleCount];
            stream.read(reinterpret_cast<char*>(part.mMeshes[i].mTriangles),
                        sizeof(Md3Triangle)*info.mTriangleCount);

            // Read uvs.
            stream.seekg(current + std::ifstream::pos_type(info.mUvStart));
            part.mMeshes[i].mUvs = GE_NEW Vector2[info.mVertexCount];
            stream.read(reinterpret_cast<char*>(part.mMeshes[i].mUvs),
                        sizeof(Vector2)*info.mVertexCount);

            // Read vertices.
            stream.seekg(current + std::ifstream::pos_type(info.mVertexStart));
            part.mMeshes[i].mVertices = GE_NEW Md3Vertex[info.mVertexCount*info.mFrameCount];
            stream.read(reinterpret_cast<char*>(part.mMeshes[i].mVertices),
                        sizeof(Md3Vertex)*info.mVertexCount*info.mFrameCount);

            // Move to next mesh for this body part.
            current += std::ifstream::pos_type(info.mMeshSize);
	    }
    }

    return ok;
}


//------------------------------------------------------------------------------
void Md3Importer::Impl::convert()
//------------------------------------------------------------------------------
{
    // Set up model container Cluster = root node.
    mCluster = GE_NEW Cluster;
    Locator* clumpLocator = GE_NEW Locator;
    makeXRotation(clumpLocator->get(), -PI_HALF);
    Matrix3x4 yRot;
    makeYRotation(yRot, PI_HALF);
    clumpLocator->get() = yRot*clumpLocator->get();
    mCluster->set(*clumpLocator);

    // Convert the individual parts.
    Md3Model::Part* lower;
    Locator* lowerLocator;
    convert(mLower, lower, lowerLocator);
    Md3Model::Part* upper;
    Locator* upperLocator;
    convert(mUpper, upper, upperLocator);
    Md3Model::Part* head;
    Locator* headLocator;
    convert(mHead, head, headLocator);

    // Assign the joints to the corresponding Model::Parts. Every Part contains
    // a map of all joints that it should animate during keyframe animation
    // (which should normally just be a list of its child frames).
    lower->mJoints["tag_torso"] = upperLocator;
    upper->mJoints["tag_head"] = headLocator;

    // Build model hierarchy.
    clumpLocator->addChild(*lowerLocator);
    lowerLocator->addChild(*upperLocator);
    upperLocator->addChild(*headLocator);

    readAnimationCfg(lower, upper);

    mModel = GE_NEW Md3Model(*mCluster, *lower, *upper, *head);

    // Do one update so that joints are set up correctly and bounding sphere
    // is calculated.
    mModel->update();
}


//------------------------------------------------------------------------------
void Md3Importer::Impl::convert(Part& part, Md3Model::Part*& modelPart,
                                Locator*& modelLocator)
//------------------------------------------------------------------------------
{
    modelPart = GE_NEW Md3Model::Part;
//    modelPart->mFrameCount = part.mFrameCount;
    modelPart->mTags = part.mTags;
    modelPart->mTagCount = part.mTagCount;
    
    // Create the locator for this part.
    modelLocator = Locator::create(Matrix3x4::IDENTITY());

    // Loop over all SubMeshes for this body part.
    for (UInt i = 0; i<part.mMeshCount; ++i)
    {
        SubMesh& mesh = part.mMeshes[i];
    
        // Create Mesh for SubMesh
        Mesh* geometry = GE_NEW Mesh(
            mesh.mFrameCount, mesh.mVertexCount, mesh.mTriangleCount*3);
        GE_ASSERT_MSG(geometry != 0, "Geometry creation failed.");

        // Morph meshes are dynamic.
        geometry->setDynamic(true);

        geometry->setUvs(0, Mesh::UV_ARRAY, mesh.mUvs);
        geometry->flipUvs(0);

        // Copy vertices and normals.
        for (UInt f = 0; f<mesh.mFrameCount; ++f)
        {
            MorphFrame* morphFrame = geometry->getMorphFrame(f);
            
            Vector3* vertices = GE_NEW Vector3[mesh.mVertexCount];
            morphFrame->setVertices(vertices);

            Vector3* normals = GE_NEW Vector3[mesh.mVertexCount];
            morphFrame->setNormals(normals);

            UInt offset = f*mesh.mVertexCount;
            Md3Vertex* md3Vertices = mesh.mVertices+offset;
            for (UInt v = 0; v<mesh.mVertexCount; ++v)
            {
                // >>6 because Q3 vertices are scaled by 64.
                set(vertices[v], Float(md3Vertices[v].mVertex[0])/64.0f,
                                 Float(md3Vertices[v].mVertex[1])/64.0f,
                                 Float(md3Vertices[v].mVertex[2])/64.0f);


                // Normal is given in sphere coordinates.
                Float lat, lng;       
                lat = Float(md3Vertices[v].mNormal[1])*PI/128.0f;
                lng = Float(md3Vertices[v].mNormal[0])*PI/128.0f;
                set(normals[v], cos(lat)*sin(lng), sin(lat)*sin(lng), cos(lng));
            }

            // Calculate the geomtry's bounding sphere.
            morphFrame->calculateBound(mesh.mVertexCount);
        }


        // Copy triangles.
        UInt16* triangles = GE_NEW UInt16[mesh.mTriangleCount*3];
        geometry->setIndices(triangles);
        for (UInt t = 0; t<mesh.mTriangleCount; ++t)
        {
            triangles[t*3] = mesh.mTriangles[t].mIndices[0];
            triangles[t*3+1] = mesh.mTriangles[t].mIndices[2];
            triangles[t*3+2] = mesh.mTriangles[t].mIndices[1];
        }

        // Create link, add the geometry and put it into clump.
        Link* link = GE_NEW Link;
        link->set(*modelLocator);
        link->set(*geometry);

        // Create material
        Material* material = GE_NEW Material;
        std::string file;
        getMaterialName(reinterpret_cast<char*>(&(mesh.mSkins[0].mName[1])), file);
        Texture* texture = theTexturePool().get(mPk3, (mPath+'/'+file).c_str());
        material->set(0, *texture);
        link->set(*material);

        modelPart->mLinks.push_back(link);
        mCluster->add(*link);
    }
}


//------------------------------------------------------------------------------
void Md3Importer::Impl::readAnimationCfg(Md3Model::Part* lower,
                                         Md3Model::Part* upper)
//------------------------------------------------------------------------------
{
    zip::izipstream file(mPk3, (mPath+"/animation.cfg").c_str());
    std::string line;
    while (file)
    {
        std::getline(file, line);
        Md3Animation animation;
        if (isdigit(line[0]))
        {
            std::istringstream stream(line);
            stream >> animation.mFrameStart;
            stream >> animation.mFrameCount;
            stream >> animation.mFrameLoop;
            stream >> animation.mFps;
            std::string dummy;
            stream >> dummy;
            stream >> dummy;
            if (dummy.find("BOTH") != std::string::npos)
            {
                lower->mAnimations[dummy] = animation;
                upper->mAnimations[dummy] = animation;
            }
            else if (dummy.find("LEGS") != std::string::npos)
            {
                lower->mAnimations[dummy] = animation;
            }
            else if (dummy.find("TORSO") != std::string::npos)
            {
                upper->mAnimations[dummy] = animation;
            }
        }
    }

    UInt skip = lower->mAnimations["LEGS_WALKCR"].mFrameStart -
                upper->mAnimations["TORSO_GESTURE"].mFrameStart;

    for (Md3Model::Part::Animations::iterator i = lower->mAnimations.begin();
         i != lower->mAnimations.end(); ++i)
    {
        if (i->first.find("LEGS") != std::string::npos)
        {
            i->second.mFrameStart -= skip;
        }
    }

    lower->mIdle = "LEGS_IDLE";
    upper->mIdle = "TORSO_STAND";
}

//------------------------------------------------------------------------------
Md3Model* Md3Importer::Impl::import() const
//------------------------------------------------------------------------------
{
    return mModel;
}





//------------------------------------------------------------------------------
// Md3Importer members
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
Md3Importer::Md3Importer(const char* path) :
//------------------------------------------------------------------------------
mImpl(GE_NEW Impl(path))
{
}

//------------------------------------------------------------------------------
Md3Importer::~Md3Importer()
//------------------------------------------------------------------------------
{
    if (mImpl != 0)
    {
        GE_DELETE mImpl;
    }
}

//------------------------------------------------------------------------------
Md3Model* Md3Importer::import() const
//------------------------------------------------------------------------------
{
    return mImpl->import();
}

} // namespace Ge

