#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Aug-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Md3Model.h"
#include "Md3Types.h"

#include "Cluster.h"

#include <Math/Tools.h>
#include <Math/Interpolation.h>

#include <Application/ScaledMetronome.h>

namespace
{

using namespace Ge;

// IN:
// <time> = delta time to the last frame in seconds.
// <fps> = frames per second for this animation.
// <currentDelta> = current delta in this frames interpolation sequence (in [0,1]).
//
// OUT:
// <delta> = new delta in this frames interpolation sequence (in [0,1]).
// <skip> = number of frames to be skipped (e.g. is 1 if the animation goes into
//          the next frame.
//------------------------------------------------------------------------------
void timeToDelta(Float time, UInt fps, Float currentDelta,
                 Float& delta, UInt& skip)
//------------------------------------------------------------------------------
{
    delta = time*fps + currentDelta;
    skip = 0;
    if (delta > 1.0f)
    {
        skip = delta; // skip is the integer part.
        delta -= skip; // delta is the remainder.
    }
}

} // namespace

namespace Ge
{

//------------------------------------------------------------------------------
Md3Model::Part::Part() :
//------------------------------------------------------------------------------
mTags(0)
{
}

//------------------------------------------------------------------------------
Md3Model::Part::~Part()
//------------------------------------------------------------------------------
{
    if (mTags != 0)
    {
        GE_DELETE_A mTags;
    }
}


//------------------------------------------------------------------------------
Md3Model::Md3Model(Cluster& c, Md3Model::Part& lower, Md3Model::Part& upper,
                   Md3Model::Part& head) :
//------------------------------------------------------------------------------
mCluster(&c),
mLower(&lower),
mUpper(&upper),
mHead(&head)
{
    setLowerAnimation(mLower->mIdle.c_str());
    setUpperAnimation(mUpper->mIdle.c_str());
}

//------------------------------------------------------------------------------
Md3Model::~Md3Model()
//------------------------------------------------------------------------------
{
    if (mLower != 0)
    {
        GE_DELETE mLower;
    }

    if (mUpper != 0)
    {
        GE_DELETE mUpper;
    }

    if (mHead != 0)
    {
        GE_DELETE mHead;
    }
}

//------------------------------------------------------------------------------
const Cluster& Md3Model::getCluster() const
//------------------------------------------------------------------------------
{
    return *mCluster;
}

//------------------------------------------------------------------------------
Cluster& Md3Model::getCluster()
//------------------------------------------------------------------------------
{
    return *mCluster;
}


//------------------------------------------------------------------------------
void Md3Model::update()
//------------------------------------------------------------------------------
{
    update(*mLower);
    update(*mUpper);

    //GE_TRACE(mLower->mJoints["tag_torso"]->get());
    mCluster->getLocator()->update();
    mCluster->calculateBound();
}

//------------------------------------------------------------------------------
void Md3Model::setLowerAnimation(const char* anim)
//------------------------------------------------------------------------------
{
    setAnimation(*mLower, anim);
}

//------------------------------------------------------------------------------
void Md3Model::setUpperAnimation(const char* anim)
//------------------------------------------------------------------------------
{
    setAnimation(*mUpper, anim);
}

//------------------------------------------------------------------------------
void Md3Model::setAnimation(Md3Model::Part& part, const char* anim)
//------------------------------------------------------------------------------
{
    part.mCurrentAnimation = anim;
    part.mCurrentFrame = part.mAnimations[part.mCurrentAnimation].mFrameStart;
    part.mCurrentDelta = 0;
}

//------------------------------------------------------------------------------
void Md3Model::update(Md3Model::Part& part)
//------------------------------------------------------------------------------
{
    Md3Animation& animation = part.mAnimations[part.mCurrentAnimation];
    if(animation.mFrameCount > 1)
    {
        Float time = theScaledMetronome().getDelta()*0.001f;

        Float delta;
        UInt skip;
        timeToDelta(time, animation.mFps, part.mCurrentDelta, delta, skip);
        part.mCurrentFrame += skip;
        

        // Animation end reached
        if (part.mCurrentFrame >= animation.mFrameStart+animation.mFrameCount-1)
        {
            if (animation.mFrameLoop != 0)
            {
                wrap(part.mCurrentFrame, 
                     animation.mFrameStart+animation.mFrameCount-animation.mFrameLoop,
                     animation.mFrameStart+animation.mFrameCount-1);
            }
            else
            {
                setAnimation(part, part.mIdle.c_str());
                update(part);
            }
        }

        UInt nextFrame = part.mCurrentFrame+1;
        if (part.mCurrentFrame == animation.mFrameStart+animation.mFrameCount-1)
        {
            nextFrame = animation.mFrameStart+animation.mFrameCount-animation.mFrameLoop;
        }

        updateMorphs(part, part.mCurrentFrame, nextFrame, delta);
        updateJoints(part, part.mCurrentFrame, nextFrame, delta);

        part.mCurrentDelta = delta;
    }
    else
    {
        part.mCurrentFrame = animation.mFrameStart;
//        GE_ASSERT_MSG(part.mCurrentFrame < animation.mFrameCount, "Invalid frame set.");
        updateMorphs(part, part.mCurrentFrame, 0, 0);
        updateJoints(part, part.mCurrentFrame, 0, 0);
    }
}

//------------------------------------------------------------------------------
void Md3Model::updateMorphs(const Md3Model::Part& part,
                            UInt frame, UInt nextFrame, Float delta)
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<part.mLinks.size(); ++i)
    {
        Mesh* g = part.mLinks[i]->getMesh();
        g->lerp(frame, nextFrame, delta);
    }
}

//------------------------------------------------------------------------------
void Md3Model::updateJoints(Md3Model::Part& part,
                            UInt frame, UInt nextFrame, Float delta)
//------------------------------------------------------------------------------
{
    for (UInt j = 0; j<part.mTagCount; ++j)
    {
        UInt baseTag = frame*part.mTagCount;
        const char* tagName = reinterpret_cast<const char*>(part.mTags[baseTag+j].mName);
        if (part.mJoints.find(tagName) != part.mJoints.end())
        {
            UInt nextTag = nextFrame*part.mTagCount;

            // Interpolate rotation.
            Matrix3x3 m1;
            Md3Tag& tag1 = part.mTags[baseTag+j];
            set(m1,
                tag1.mRotation[0][0], tag1.mRotation[1][0], tag1.mRotation[2][0],
                tag1.mRotation[0][1], tag1.mRotation[1][1], tag1.mRotation[2][1],
                tag1.mRotation[0][2], tag1.mRotation[1][2], tag1.mRotation[2][2]);

            Matrix3x3 m2;
            Md3Tag& tag2 = part.mTags[nextTag+j];
            set(m2,
                tag2.mRotation[0][0], tag2.mRotation[1][0], tag2.mRotation[2][0],
                tag2.mRotation[0][1], tag2.mRotation[1][1], tag2.mRotation[2][1],
                tag2.mRotation[0][2], tag2.mRotation[1][2], tag2.mRotation[2][2]);

            Quaternion q1, q2, q;
            q1.convertFrom(m1);
            q2.convertFrom(m2);
            q.slerp(q1, q2, delta);

            // Interpolate translation.
            Vector3 v1, v2, v;
            set(v1, tag1.mTranslation[0], tag1.mTranslation[1], tag1.mTranslation[2]);
            set(v2, tag2.mTranslation[0], tag2.mTranslation[1], tag2.mTranslation[2]);
            lerp(v, v1, v2, delta);

            // Update Locator.
            q.convertTo(part.mJoints[tagName]->get());
            part.mJoints[tagName]->get().setColumn(MatrixBase::TRANSLATION, v);
        }
    }
}

} // namespace Ge

