#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include "Mesh.h"
#include "Renderer.h"
#include "RenderCache.h"
#include "Color.h"

#include <Math/Vector.h>

#include <map>


namespace
{

using namespace Ge;

template<class T> inline
//------------------------------------------------------------------------------
void copy(Ge::UInt count, std::vector<T>& to, const T* from)
//------------------------------------------------------------------------------
{
    if (from != 0)
    {
        to.reserve(count);
        for (int i = 0; i<count; i++)
            to.push_back(from[i]);
    }
}

} // namespace






namespace Ge
{

const UInt Mesh::UV_SET_COUNT = 4;

//------------------------------------------------------------------------------
Mesh::Mesh(UInt frameCount,
           UInt vertexCount,
           Vector3* vertices,
           Vector3* normals,
           Vector2* textureUvs,
           Color* diffuse,
           UInt indexCount,
           UInt16* indices)
//------------------------------------------------------------------------------
{
    init();

    GE_ASSERT_MSG(frameCount > 0, "At least one frame needed");
    GE_ASSERT_MSG(vertexCount > 0, "Empty vertex list");
    GE_ASSERT_MSG(vertices != 0, "Invalid vertex pointer");
    GE_ASSERT_MSG(indexCount > 0, "Empty index list");
    GE_ASSERT_MSG(indices != 0, "Invalid index pointer");
    GE_ASSERT_MSG(vertexCount <= 0xffff, "No 32 bit indices supported");
        
    mFrameCount = frameCount;
    if (frameCount > 0)
    {
        mFrames = GE_NEW MorphFrame[frameCount];
        
        mVertexCount = vertexCount;
        mFrames[0].setVertices(vertices);
        mFrames[0].setNormals(normals);
    
        if (textureUvs != 0)
        {
            mTextureUvs[0].first = UV_ARRAY;
            mTextureUvs[0].second = textureUvs;
        }

        mColors[0] = diffuse;
        mIndexCount = indexCount;
        mIndices = indices;

        calculateBound();
        lerp(0,0,0);
    }
}

//------------------------------------------------------------------------------
Mesh::Mesh(UInt frameCount, UInt vertexCount, UInt indexCount)
//------------------------------------------------------------------------------
{
    init();

    mFrameCount = frameCount;
    mVertexCount = vertexCount;
    mIndexCount = indexCount;

    GE_ASSERT_MSG(frameCount > 0, "At least one frame needed");
    GE_ASSERT_MSG(vertexCount > 0, "Empty vertex list");
    GE_ASSERT_MSG(indexCount > 0, "Empty index list");
    GE_ASSERT_MSG(vertexCount <= 0xffff, "No 32 bit indices supported");

    if (frameCount > 0)
    {
        mFrames = GE_NEW MorphFrame[frameCount];
        lerp(0,0,0);
    }
}

//------------------------------------------------------------------------------
void Mesh::init()
//------------------------------------------------------------------------------
{
    mFrameCount = 0;
    mFrames = 0;
    mVertexCount = 0;
    mColors[0] = 0;
    mColors[1] = 0;
    UInt mIndexCount = 0;
    mIndices = 0;
    mConnectivity = 0;
    mDynamic = false;
}

//------------------------------------------------------------------------------
Mesh::~Mesh()
//------------------------------------------------------------------------------
{
    if (mFrames != 0)
    {
        GE_DELETE_A mFrames;
    }

    for (UInt i = 0; i<UV_SET_COUNT; ++i)
    {
        if (mTextureUvs[i].first == UV_ARRAY)
        {
            GE_DELETE_A mTextureUvs[i].second;
        }
    }

    for (UInt c = 0; c<2; ++c)
    {
        if (mColors[c] != 0)
        {
            GE_DELETE_A mColors[c];
        }
    }

    if (mIndices != 0)
    {
        GE_DELETE_A mIndices;
    }

    if (mConnectivity != 0)
    {
        GE_DELETE_A mConnectivity;
    }

    if (mActive.mLerpedFrame != 0)
    {
        GE_DELETE mActive.mLerpedFrame;
    }
}

////------------------------------------------------------------------------------
//void Mesh::render(Renderer& renderer) const
////------------------------------------------------------------------------------
//{
//    renderer.render(*this);
//}


//------------------------------------------------------------------------------
void Mesh::calculateNormals(bool smooth, bool weighByAngle)
//------------------------------------------------------------------------------
{   
    GE_ASSERT(mIndices != 0);
    if (!mIndices)
    {
        return;
    }

    for (UInt f = 0; f<mFrameCount; ++f)
    {
        MorphFrame& frame = mFrames[f];
        // Reserve space if mesh had no normals
        if (frame.mNormals == 0)
        {
            frame.mNormals = GE_NEW Vector3[getVertexCount()];
        }

        // Clear old normals.
        memset(frame.mNormals, 0, sizeof(Vector3)*getVertexCount());
    
        // If smooth, then create a vertex lookup map, so multiple vertices that
        // represent one vertex can be found fast
        std::map<Vector3, std::vector<UInt>, Vector3::Pred> vertices;
        UInt i(0);
        if (smooth)
        {
            for (i = 0; i<getVertexCount(); i++)
            {
                vertices[frame.getVertices()[i]].push_back(i);
            }
        }

        // Calculate new normals (calculate normal for every face, then sum
        // the normals up for each vertex)
        Vector3 faceNormal;
        Vector3 v0v1, v0v2, v1v2; // Edges
        UInt i0, i1, i2; // Indices
        Float a0(1), a1(1), a2(1); // Angles
        for (i = 0; i<getPrimitiveCount(); i++)
        {
            // Get triangle's vertices
            getTriangleVertices(i, i0, i1, i2);

            // Calculate face normal
            v0v1 = frame.getVertices()[i1];
            v0v1 -= frame.getVertices()[i0];
            v0v2 = frame.getVertices()[i2];
            v0v2 -= frame.getVertices()[i0];
            faceNormal = cross(v0v1, v0v2);

            // Weigh normal by angle?
            if (weighByAngle)
            {
                v1v2 = frame.getVertices()[i2];
                v1v2 -= frame.getVertices()[i1];
                a0 = v0v1.getAngleTo(v0v2);
                a1 = (-v0v1).getAngleTo(v1v2);
                a2 = (-v0v2).getAngleTo(-v1v2);
            }
        
            // Treat multiple vertices with same coordinates as one
            if (smooth)
            {
                UInt j(0);
                for (j = 0; j<vertices[frame.getVertices()[i0]].size(); j++)
                {
                    frame.mNormals[vertices[frame.getVertices()[i0]][j]] += a0*faceNormal;
                }

                for (j = 0; j<vertices[frame.getVertices()[i1]].size(); j++)
                {
                    frame.mNormals[vertices[frame.getVertices()[i1]][j]] += a1*faceNormal;
                }

                for (j = 0; j<vertices[frame.getVertices()[i2]].size(); j++)
                {
                    frame.mNormals[vertices[frame.getVertices()[i2]][j]] += a2*faceNormal;
                }
            }
            else
            {
                frame.mNormals[i0] += a0*faceNormal;
                frame.mNormals[i1] += a1*faceNormal;
                frame.mNormals[i2] += a2*faceNormal;
            }
        }

        // Normalize the result
        for (i = 0; i<getVertexCount(); i++)
        {
            // Some imported models had vertices that had exactly 0 here
            // (maybe unused vertices? check later.)
            if (frame.mNormals[i] != Vector3::ORIGIN())
            {
                frame.mNormals[i].normalize();
            }
        }
    }

    // Update active frame.
    lerp();

    // Update the cache
    if (mCache != 0)
    {
        mCache->update();
    }
}


//------------------------------------------------------------------------------
void Mesh::setDynamic(bool d)
//------------------------------------------------------------------------------
{
    mDynamic = d;
}

//------------------------------------------------------------------------------
bool Mesh::isDynamic() const
//------------------------------------------------------------------------------
{
    return mDynamic;
}

//------------------------------------------------------------------------------
void Mesh::setUvs(UInt set, UvType type, Vector2* array)
//------------------------------------------------------------------------------
{
    GE_ASSERT(set < UV_SET_COUNT);
    GE_ASSERT_MSG(mTextureUvs[set].first != UV_ARRAY ||
                  array != 0, "Uv type is UV_ARRAY but no coordinates provided.");

    if (set < UV_SET_COUNT)
    {
        // Delete old uv coordinates
        if (mTextureUvs[set].second != 0)
        {
            GE_DELETE_A mTextureUvs[set].second;
        }

        mTextureUvs[set].first = type;

        if (type == UV_ARRAY)
        {
            GE_ASSERT(array != 0);
            if (array == 0)
            {
                mTextureUvs[set].first = UV_NONE;
            }

            mTextureUvs[set].second = array;

            if (mCache != 0)
            {
                mCache->update();
            }
        }
    }
}

//------------------------------------------------------------------------------
void Mesh::flipUvs(UInt set)
//------------------------------------------------------------------------------
{
    if (getUvType(set) == UV_ARRAY)
    {
        for (UInt i = 0; i<getVertexCount(); i++)
        {
            mTextureUvs[set].second[i][1] = 1 - (getUvs(set)[i])[1];
        }

        if (mCache != 0)
        {
            mCache->update();
        }
    }
}

//------------------------------------------------------------------------------
void Mesh::flipTangents()
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<mFrameCount; ++i)
    {
        mFrames[i].flipTangents(getVertexCount());
    }
}

//------------------------------------------------------------------------------
void Mesh::setIndices(UInt16* i)
//------------------------------------------------------------------------------
{
    GE_ASSERT(i != 0);
    if (i != 0)
    {
        if (mIndices != 0)
        {
            GE_DELETE_A mIndices;
        }

        mIndices = i;
    }
}

//------------------------------------------------------------------------------
Color* Mesh::getColors(UInt set)
//------------------------------------------------------------------------------
{
    GE_ASSERT(set < 2);
    return set<2 ? mColors[set] : 0;
}

//------------------------------------------------------------------------------
void Mesh::setColors(UInt set, Color* c)
//------------------------------------------------------------------------------
{
    GE_ASSERT(set < 2);
    if (set<2 && mColors[set] != 0)
    {
        GE_DELETE_A mColors[set];
    }

    mColors[set] = c;
}

//------------------------------------------------------------------------------
const Color* Mesh::getColors(UInt set) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(set < 2);
    return set<2 ? mColors[set] : 0;
}

//------------------------------------------------------------------------------
void Mesh::getTriangleEdge(UInt triangleNumber, UInt edgeNumber, Edge& edge)
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(edgeNumber < 3, "Invalid edge number.");

    UInt16* triangle = &(mIndices[triangleNumber*3]);
   
    edge[0] = mFrames[0].getVertices()[triangle[edgeNumber]];
    edge[1] = mFrames[0].getVertices()[triangle[(edgeNumber+1)%3]];
}

//------------------------------------------------------------------------------
bool Mesh::triangleContainsEdge(UInt triangleNumber, const Edge& edge)
//------------------------------------------------------------------------------
{
    UInt16* triangle = &(mIndices[triangleNumber*3]);

    Edge triangleEdge;
    for (UInt e = 0; e<3; ++e)
    {
        getTriangleEdge(triangleNumber, e, triangleEdge);
        if (edge == triangleEdge)
        {
            return true;
        }
    }

    return false;
}

//------------------------------------------------------------------------------
bool Mesh::calculateConnectivity()
//------------------------------------------------------------------------------
{
    bool ret = true;

    if (mConnectivity != 0)
    {
        GE_DELETE_A mConnectivity;
    }

    mConnectivity = GE_NEW Neighbours[getPrimitiveCount()];
    Edge edge;
    for (UInt t = 0; t<getPrimitiveCount(); ++t)
    {
        // For all edges of this triangle
        for (UInt e = 0; e<3; ++e)
        {
            getTriangleEdge(t, e, edge);

            for (UInt other = 0; other<getPrimitiveCount(); ++other)
            {
                // Skip the current triangle.
                if (other == t)
                {
                    continue;
                }

                // Neighbour for this edge found. Quit searching.
                if (triangleContainsEdge(other, edge))
                {
                    mConnectivity[t][e] = other;
                    break;
                }
            }

            ret = false;
            GE_ASSERT_MSG(other<getPrimitiveCount(),
                "No neighbour found for this edge.");
        }
    }

    return ret;
}

//------------------------------------------------------------------------------
void Mesh::invert()
//------------------------------------------------------------------------------
{   
    UInt16* triangle = 0;
    for (UInt i = 0; i<getPrimitiveCount(); ++i)
    {
        triangle = &(mIndices[i*3]);
        std::swap(triangle[1], triangle[2]);
    }

    for (UInt f = 0; f<mFrameCount; ++f)
    {
        if (mFrames[f].getNormals() != 0)
        {
            for (UInt j = 0; j<getVertexCount(); ++j)
            {
                mFrames[f].mNormals[j] = -mFrames[f].mNormals[j];
            }
        }   
    }

    lerp();
}

//------------------------------------------------------------------------------
const Mesh::Neighbours* Mesh::getConnectivity() const
//------------------------------------------------------------------------------
{
    return mConnectivity;
}

//------------------------------------------------------------------------------
const UInt16* Mesh::getIndices() const
//------------------------------------------------------------------------------
{
    return mIndices;
}


//------------------------------------------------------------------------------
void Mesh::calculateFacePlanes()
//------------------------------------------------------------------------------
{
    for (UInt f = 0; f<mFrameCount; ++f)
    {
        mFrames[f].calculateFacePlanes(getPrimitiveCount(), mIndices);
    }

    lerp();
}


//------------------------------------------------------------------------------
const MorphFrame& Mesh::getActiveFrame() const
//------------------------------------------------------------------------------
{
    return *mActive.mFrame;
}

//------------------------------------------------------------------------------
void Mesh::calculateBound()
//------------------------------------------------------------------------------
{
    for (UInt f = 0; f<mFrameCount; ++f)
    {
        mFrames[f].calculateBound(getVertexCount());
    }

    lerp();
}


//------------------------------------------------------------------------------
void Mesh::lerp(UInt from, UInt to, Float s)
//------------------------------------------------------------------------------
{
    GE_ASSERT(from < mFrameCount);
    GE_ASSERT(to < mFrameCount);

    if (from < mFrameCount && to < mFrameCount)
    {
        mActive.mFrom = from;
        mActive.mTo = to;
        mActive.mS = s;
    }
    else
    {
        mActive.mFrom = 0;
        mActive.mTo = 0;
        mActive.mS = 0;
    }

    lerp();
    if (mCache != 0)
    {
        mCache->update();
    }
}

//------------------------------------------------------------------------------
void Mesh::lerp()
//------------------------------------------------------------------------------
{
    Float s = mActive.mS;
    if (s != 0 && s != 1)
    {
        if (mActive.mLerpedFrame == 0)
        {
            mActive.mLerpedFrame = GE_NEW MorphFrame;
        }

        mActive.mLerpedFrame->lerp(
            getVertexCount(), getPrimitiveCount(),
            mFrames[mActive.mFrom], mFrames[mActive.mTo], s);

        mActive.mFrame = mActive.mLerpedFrame;
    }
    else if (s == 0)
    {
        mActive.mFrame = &mFrames[mActive.mFrom];
    }
    else if (s == 1)
    {
        mActive.mFrame = &mFrames[mActive.mTo];
    }
}

//------------------------------------------------------------------------------
MorphFrame* Mesh::getMorphFrame(UInt frameNumber)
//------------------------------------------------------------------------------
{
    GE_ASSERT(frameNumber < mFrameCount && mFrames != 0);
    if (frameNumber < mFrameCount && mFrames != 0)
    {
        return &mFrames[frameNumber];
    }
    else
    {
        return 0;
    }
}

//------------------------------------------------------------------------------
Mesh::ActiveFrame::ActiveFrame() :
//------------------------------------------------------------------------------
mFrom(0),
mTo(0), 
mS(0), 
mFrame(0),
mLerpedFrame(0)
{
}

} // namespace Ge