//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

// 19.1.2002: The mesh now takes ownership of the vertex, normals,... memory
// that is passed to it and doesn't copy them to std::vectors anymore. This
// brought a big speed increase for dynamic meshes like water.


#ifndef ENGINE_MESH_H
#define ENGINE_MESH_H

#include "MorphFrame.h"
#include "Edge.h"

#include <System/Base.h>
#include <System/Pointer.h>

#include <Math/Vector.h>
#include <Math/Sphere.h>

#include <vector>
#include <utility>

namespace
{
    struct VertexInfo;
}

namespace Ge
{

class RenderCache;
class Color;

// Represents a modeled mesh. The Mesh takes ownership of the data it is passed
// (i.e. it will delete the vertex, normal,... arrays on destruction)
class Mesh : public Object
{
public:
    //------------------------------------------------------------------------------
    // Types & constants
    //------------------------------------------------------------------------------
    
    enum UvType
    {
        // Mesh has no uv coordinates
        UV_NONE,
        // uv coordinates are stored in an array
        UV_ARRAY,
        // uv coordinates are generated for a spherical environment map
        UV_SPHERE,
        // uv coordinates are generated for a cube environment map
        UV_CUBE
    };

    // Connectivity type.
    typedef UInt16 Neighbours[3];

    // The number of possible texture coordinate sets.
    static const UInt UV_SET_COUNT /* = 4 */;

    //------------------------------------------------------------------------------
    // Construction
    //------------------------------------------------------------------------------
    
    // - <frameCount>: Number of morph frames
    Mesh(UInt frameCount,
         UInt vertexCount,
         Vector3* vertices,
         Vector3* normals,
         Vector2* textureUvs,
         Color* colors,
         UInt indexCount,
         UInt16* indices);

    Mesh(UInt frameCount,
         UInt vertexCount,
         UInt indexCount);

    ~Mesh();

    MorphFrame* getMorphFrame(UInt frameNumber);

    // Set the Mesh's active frame. It lerps between the frames <from> and <to>.
    // <s> must be between 0 and 1.
    void lerp(UInt from, UInt to, Float s);

    const MorphFrame& getActiveFrame() const;

//    virtual
//    void render(Renderer& renderer) const;

    Pointer<RenderCache>& cache() const;
    
    // Vertex data
    void setDynamic(bool);

    bool isDynamic() const;

    UInt getVertexCount() const;
    
    // This member only makes sense for UV_ARRAY type
    // - <set>: texture coordinate set
    // - <index>: coordinate index
    const Vector2* getUvs(UInt set) const;

    // Set the texture coordinates for the corresponding coordinate <set>.
    // (There are four possible texture coordinate sets).
    // - <set>: coordinate set.
    // - <type>: coordinate type.
    // - <array>: if <type> is UV_ARRAY this must be != 0
    void setUvs(UInt set, UvType type, Vector2* array = 0);
    
    UvType getUvType(UInt set) const;

    const Color* getColors(UInt set) const;

    Color* getColors(UInt set);

    void setColors(UInt set, Color*);

    // Indices
    UInt getIndexCount() const;
    
    const UInt16* getIndices() const;

    void setIndices(UInt16*);

    UInt getPrimitiveCount() const;
    

    //------------------------------------------------------------------------------
    // Tools
    //------------------------------------------------------------------------------
    
    // Recalculates the mesh's vertex normals. If some vertices are doubled in the
    // mesh (e.g. triangle list with no index list) and <smooth> is true, then
    // vertices with the same coordinates are treated as one vertex.
    // If <weighByAngle> is true, then the face normal is weighed be the edge
    // angle at the vertex, giving an even more smooth appearance.
    // 
    // The RenderCache is updated afterwards.
    void calculateNormals(bool smooth = false, bool weighByAngle = false);

    // Flips the texture uv coordinates of <set> on the v axis (e.g. needed for Maya models)
    void flipUvs(UInt set);

    // Vertices with flipped uvs also need their tangents flipped.
    void flipTangents();

    // Calculates connectivity information (i.e. the 3 neighbouring triangles for
    // each triangle). Assumption: the mesh is closed. Returns false if the mesh
    // is not closed.
    bool calculateConnectivity();

    const Neighbours* getConnectivity() const;

    // Inverts the triangle winding and vertex normals (e.g. make a room out of a box).
    void invert();

    // Calculates the bounds of all MorphFrames.
    void calculateBound();

    // Calculate face planes of all MorphFrames.
    void calculateFacePlanes();

    // Calculate tangent space transforms of all MorphFrames. <uvSet> is the set
    // of uv coordinates that is taken into account for tangent space calculations.
    // If the <uvs> have been flipped with flipUvs() before, then the tangents
    // also have to be calculated with <flipped> == true.
    void calculateTangents(bool flipped = false, UInt uvSet = 0);

private:
    // Set pointer and so on to default values.
    void init();

    // Helper function.
    // For a given <triangleNumber> and <edgeNumber> within this triangle
    // return the <edge>.
    void getTriangleEdge(UInt triangleNumber, UInt edgeNumber, Edge& edge);

    // Helper function.
    // Does a given triangle contain an <edge>.
    bool triangleContainsEdge(UInt triangleNumber, const Edge& edge);

    // For the triangle number <triangle> retrieve its vertices. I.e. the
    // triangle will be v0, v1, v2 (in clockwise order). (If the mesh's vertices
    // are accessed via an index list, this is taken into account.) <triangle>
    // is NOT checked if within range.
    void getTriangleVertices(UInt triangle, UInt& v0, UInt& v1, UInt& v2);
    
    // Helper function for the above. Used to get the vertices starting from a
    // base vertex (= the first vertex in the triangle).
    void getVerticesStartingFromBase(UInt base, UInt& v0, UInt&v1, UInt& v2);

    void lerp();

    // Map old to new index after dupplication.
    typedef std::vector<std::pair<UInt, UInt16> > IndexMap;

    // Calculate tangent space transformations for one <frame>
//    void calculateTangents(UInt frame, Matrix3x3* tangents, UInt uvSet);
    void calculateTangents(UInt frame, std::vector<Matrix3x3>& tangents, UInt uvSet,
                           VertexInfo* vertexInfos, IndexMap&,
                           std::vector<UInt16>& duplicates, UInt& duplicateCount,
                           bool flipped);

    // Calculate texture gradients form one <triangle> in one <frame>
    void calculateGradients(UInt frame, UInt triangle,
                            Vector3& du, Vector3& dv, UInt uvSet,
                            bool flipped);


    //------------------------------------------------------------------------------
    // Data
    //------------------------------------------------------------------------------
    
    UInt mFrameCount;
    MorphFrame* mFrames;

    // Current lerped frame value.
    struct ActiveFrame
    {
        ActiveFrame();
        UInt mFrom;
        UInt mTo;
        Float mS;
        MorphFrame* mFrame;
        MorphFrame* mLerpedFrame;
    } mActive;
    
    UInt mVertexCount;

    struct UvPair
    {
        UvPair() : first(UV_NONE), second(0) {}

        UvType first;
        Vector2* second;
    } mTextureUvs[4];

    Color* mColors[2];

    UInt mIndexCount;
    UInt16* mIndices;
    
    Neighbours* mConnectivity;
    
    // Is mutable, since changing the cache doesn't change the Mesh
    mutable Pointer<RenderCache> mCache;

    bool mDynamic;
};

} // namespace Ge

#include <Engine/Mesh.inl>
#endif // ENGINE_MESH_H