#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Sep-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "Mesh.h"
#include "Color.h"

namespace
{

using namespace Ge;

enum TangentSpaceOrientation
{
    TS_NONE,
    TS_NORMAL,
    TS_FLIPPED
};

struct VertexInfo
{
    VertexInfo() : mOrientation(TS_NONE), mDuplicate(0) {}
    // Type for this vertex;
    TangentSpaceOrientation mOrientation;
    // Index of vertex with with different oriented tangent space
    // (that'll be a duplicated vertex).
    UInt16 mDuplicate;
};

//------------------------------------------------------------------------------
void addTangent(Matrix3x3& m, const Vector3& r0, const Vector3& r1, const Vector3& r2)
//------------------------------------------------------------------------------
{
    m.addRow(0, r0);
    m.addRow(1, r1);
    m.addRow(2, r2);
}

// Add tangent transformations for one vertex
// <index>: Index of vertex
// <du>, <dv>, <duXdv>: Tangent space transformation to be set
// <tangents>: Vector of tangents for this <frame>
// <frame>: Current frame
// <vertexInfos>: Array of vertex mapping infos
// <duplicates>: Vector of vertex indices to be duplicated
// <duplicateCount>: Current index in the duplicate array
//------------------------------------------------------------------------------
UInt16 addTangent(Mesh& mesh, std::vector<Matrix3x3>& tangents,
                  const Vector3& du, const Vector3& dv,
                  Vector3 duXdv, UInt16 index, UInt frame, VertexInfo* vertexInfos,
                  UInt& duplicateCount, std::vector<UInt16>& duplicates)
//------------------------------------------------------------------------------
{
    UInt16 ret = 0;
    const Vector3& normal = mesh.getMorphFrame(frame)->getNormals()[index];
    TangentSpaceOrientation toBeSet = normal*duXdv > 0 ? TS_NORMAL : TS_FLIPPED;
    TangentSpaceOrientation opposite = toBeSet == TS_NORMAL ? TS_FLIPPED : TS_NORMAL;

//    if (toBeSet == TS_FLIPPED)
//    {
//        duXdv = -duXdv;
//    }

    // This gives smoother results than taking duXdv/-duXdv
    duXdv = normal;

    if (frame != 0)
    {
        if (vertexInfos[index].mOrientation == toBeSet)
        {
            addTangent(tangents[index], du, dv, duXdv);
        }
        else
        {
            addTangent(tangents[vertexInfos[index].mDuplicate], du, dv, duXdv);
        }
    }
    // This is the first frame in the morph animation. We have to duplicate vertices
    // where the tangent space orientation is flipped (i.e. the texture is mirrored).
    else
    {
        // This vertex is visited the first time or is the same orientation as the
        // first time.
        if (vertexInfos[index].mOrientation != opposite)
        {
            vertexInfos[index].mOrientation = toBeSet;
            addTangent(tangents[index], du, dv, duXdv);
        }
        // This vertex was visited and was already duplicated
        else if (vertexInfos[index].mDuplicate > 0)
        {
            addTangent(tangents[vertexInfos[index].mDuplicate], du, dv, duXdv);
        }
        // This vertex was visited, now has differend orientation
        // and was not yet duplicated
        else
        {
            vertexInfos[index].mDuplicate = duplicateCount;
            tangents.push_back(Matrix3x3::ZERO());
            addTangent(tangents[vertexInfos[index].mDuplicate], du, dv, duXdv);
            duplicates.push_back(index);
            ++duplicateCount;

            ret = vertexInfos[index].mDuplicate;
        }
    }
    
    return ret;
}

} // namespace


namespace Ge
{


// Calculate tangent space transformations for one <frame>
//------------------------------------------------------------------------------
void Mesh::calculateTangents(UInt frame, std::vector<Matrix3x3>& tangents, UInt uvSet,
                             VertexInfo* vertexInfos, IndexMap& indexMap,
                             std::vector<UInt16>& duplicates, UInt& duplicateCount,
                             bool flipped)
//------------------------------------------------------------------------------
{    
    UInt triangleCount = getIndexCount()/3;
    for (UInt i = 0; i<triangleCount; ++i)
    {
        Vector3 du, dv;
        calculateGradients(frame, i, du, dv, uvSet, flipped);

        UInt v0, v1, v2;
        getTriangleVertices(i, v0, v1, v2);
        Vector3 duXdv = cross(du, dv);
        
        UInt16 n0, n1, n2;
        n0 = addTangent(*this, tangents, du, dv, duXdv, v0, frame,
                        vertexInfos, duplicateCount, duplicates);
        
        n1 = addTangent(*this, tangents, du, dv, duXdv, v1, frame,
                        vertexInfos, duplicateCount, duplicates);

        n2 = addTangent(*this, tangents, du, dv, duXdv, v2, frame,
                        vertexInfos, duplicateCount, duplicates);
        
        // We remap the indices for duplicated vertices for the first morph frame.
        if (frame == 0)
        {
            if (n0 != 0)
            {
                indexMap.push_back(std::make_pair(i*3, n0));
            }

            if (n1 != 0)
            {
                indexMap.push_back(std::make_pair(i*3+1, n1));
            }

            if (n2 != 0)
            {
                indexMap.push_back(std::make_pair(i*3+2, n2));
            }
        }
    }

    for (UInt j = 0; j<tangents.size(); ++j)
    {
#ifndef NDEBUG
        Matrix3x3& m = tangents[j];
        UInt size = tangents.size();
#endif
        Vector3 row;
        for (UInt r = 0; r<3; ++r)
        {
            tangents[j].getRow(r, row);
            if (row != Vector3::ORIGIN())
            {
                tangents[j].normalizeRow(r);
            }
        }

//        GE_ASSERT(tangents[j] != Matrix3x3::ZERO());
    }

    // Copy vertex data and so on.
    if (duplicateCount > getVertexCount())
    {
        GE_ASSERT(getMorphFrame(frame)->getVertices() != 0 &&
                  getMorphFrame(frame)->getNormals() != 0);

        // Copy vertices
        Vector3* newVertices = GE_NEW Vector3[duplicateCount];
        memcpy(newVertices,
               getMorphFrame(frame)->getVertices(),
               getVertexCount()*sizeof(Vector3));
        getMorphFrame(frame)->setVertices(newVertices);
        for (UInt i = 0; i<duplicates.size(); ++i)
        {
            newVertices[getVertexCount()+i] = newVertices[duplicates[i]];
        }

        // Copy normals
        Vector3* newNormals = GE_NEW Vector3[duplicateCount];
        memcpy(newNormals,
               getMorphFrame(frame)->getNormals(),
               getVertexCount()*sizeof(Vector3));
        getMorphFrame(frame)->setNormals(newNormals);
        for (UInt j = 0; j<duplicates.size(); ++j)
        {
            newNormals[getVertexCount()+j] = newNormals[duplicates[j]];
        }

        if (frame == 0)
        {
            // Reserve color arrays (don't copy, since colors are reset every frame
            // with light direction anyway).
            for (UInt c = 0; c<2; ++c)
            {
                if (mColors[c] != 0)
                {
                    Color* newColors = GE_NEW Color[duplicateCount];
                    GE_DELETE_A mColors[c];
                    mColors[c] = newColors;
                    memset(mColors[c], 0, sizeof(Color)*duplicateCount);
                }
            }

            // Copy texture coordinates
            for (UInt i = 0; i<4; ++i)
            {
                if (mTextureUvs[i].first == UV_ARRAY && mTextureUvs[i].second != 0)
                {
                    Vector2* newUvs = GE_NEW Vector2[duplicateCount];
                    memcpy(newUvs, mTextureUvs[i].second, sizeof(Vector2)*getVertexCount());
                    GE_DELETE_A mTextureUvs[i].second;
                    mTextureUvs[i].second = newUvs;
                    for (UInt i = 0; i<duplicates.size(); ++i)
                    {
                        newUvs[getVertexCount()+i] = newUvs[duplicates[i]];
                    }
                }
            }
        }
    }
}

// Calculate texture gradients for one <triangle> in one <frame>
//------------------------------------------------------------------------------
void Mesh::calculateGradients(UInt frame, UInt triangle,
                              Vector3& du, Vector3& dv,
                              UInt uvSet, bool flipped)
//------------------------------------------------------------------------------
{
    UInt v0, v1, v2;
    getTriangleVertices(triangle, v0, v1, v2);
    const Vector3* vertices = mFrames[frame].getVertices();

    for (UInt i = 0; i<3; ++i)
    {
        Vector3 uv01(vertices[v1][i] - vertices[v0][i],
                     (mTextureUvs[uvSet].second[v1])[0] - (mTextureUvs[uvSet].second[v0])[0],
                     (mTextureUvs[uvSet].second[v1])[1] - (mTextureUvs[uvSet].second[v0])[1]);

        Vector3 uv02(vertices[v2][i] - vertices[v0][i],
                     (mTextureUvs[uvSet].second[v2])[0] - (mTextureUvs[uvSet].second[v0])[0],
                     (mTextureUvs[uvSet].second[v2])[1] - (mTextureUvs[uvSet].second[v0])[1]);

        Vector3 cp = cross(uv01, uv02);

        if (fabs(cp[0]) > 0)
        {
            du[i] = -cp[1]/cp[0];
            dv[i] = (flipped ? cp[2] : -cp[2])/cp[0];
        }
        else
        {
            du[i] = 0;
            dv[i] = 0;
        }
    }
}

// for all frames
//------------------------------------------------------------------------------
void Mesh::calculateTangents(bool flipped, UInt uvSet)
//------------------------------------------------------------------------------
{
    GE_ASSERT(mTextureUvs[uvSet].first == UV_ARRAY && mTextureUvs[uvSet].second != 0);
    if (mTextureUvs[uvSet].first == UV_ARRAY && mTextureUvs[uvSet].second != 0)
    {
        VertexInfo* vertexInfos = GE_NEW VertexInfo[getVertexCount()];
        IndexMap indexMap;
        std::vector<UInt16> duplicates;
        UInt duplicateCount = getVertexCount();
        for (UInt f = 0; f<mFrameCount; ++f)
        {
            std::vector<Matrix3x3> tangents;
            tangents.resize(duplicateCount, Matrix3x3::ZERO());
            calculateTangents(f, tangents, uvSet, vertexInfos, indexMap, duplicates,
                              duplicateCount, flipped);
            Matrix3x3* t = GE_NEW Matrix3x3[tangents.size()];
            Int s = tangents.size();
            memcpy(t, &(tangents[0]), tangents.size()*sizeof(Matrix3x3));
            mFrames[f].setTangents(t);
        }

        // Remap indices to duplicated vertices
        for (UInt i = 0; i<indexMap.size(); ++i)
        {
            mIndices[indexMap[i].first] = indexMap[i].second;
        }

        // Set vertex count to new value after vertex duplication
        mVertexCount = duplicateCount;

        // The current lerped frame is not correct anymore because it still has
        // the old vertex count.
        if (mActive.mLerpedFrame != 0)
        {
            GE_DELETE mActive.mLerpedFrame;
            mActive.mLerpedFrame = 0;
        }

        GE_DELETE_A vertexInfos;
    }
}

} // namespace Ge
