#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Mai-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "ModelViewCameraController.h"

#include <Input/Mouse.h>

#include <Math/Matrix.h>

#include <Engine/Camera.h>

namespace
{
    const Ge::Float ROTATION_MULTIPLIER = 0.01f;
    const Ge::Float TRANSLATION_MULTIPLIER = 0.2f;
    const Ge::Float WHEEL_ZOOM_MULTIPLIER = 0.001f;
    const Ge::Float MOUSE_ZOOM_MULTIPLIER = 2.0f;
} // namespace





namespace Ge
{

//------------------------------------------------------------------------------
ModelViewCameraController::ModelViewCameraController(Camera& camera) :
//------------------------------------------------------------------------------
mDistance(10),
mSensitivity(1),
mCamera(camera)
{
    setInvert(false);
    recalculate();
}

//------------------------------------------------------------------------------
void ModelViewCameraController::update()
//------------------------------------------------------------------------------
{
    bool changed = false;

    Vector2 pos;
    theMouse().getDelta(pos);
    pos *= mSensitivity;

    if (theMouse().isPressed(Mouse::LEFT))
    {
        // Common model view cameras rotate up/down around the camera coordinate
        // system's x-axis but left/right around the world y-axis.

        Matrix3x3 currentRotation;
        mCamera.getLocator().get().getSubMatrix(currentRotation);

        // Calculate current angle to y-axis
        Vector3 currentAt;
        mCamera.getLocator().get().getColumn(MatrixBase::AT, currentAt);
        Float currentAngle = currentAt.getAngleTo(Vector3(0,1,0));
    
        // If rotation would rotate over zenith then don't do up/down
        // part of the rotation.
        Matrix3x3 upDown;
        if (currentAngle + mInvert*pos[1]*ROTATION_MULTIPLIER < 0 ||
            currentAngle + mInvert*pos[1]*ROTATION_MULTIPLIER > PI)
        {
            upDown = Matrix3x3::IDENTITY();
        }
        else
        {
            makeXRotation(upDown, mInvert*pos[1]*ROTATION_MULTIPLIER);
        }
    
        Matrix3x3 leftRight;
        makeYRotation(leftRight, mInvert*pos[0]*ROTATION_MULTIPLIER);

        mCamera.getLocator().get().setSubMatrix(leftRight*currentRotation*upDown);

        changed = true;
    }

    if (theMouse().isPressed(Mouse::RIGHT))
    {
        Vector3 relative(-mInvert*pos[0]*TRANSLATION_MULTIPLIER,
                         mInvert*pos[1]*TRANSLATION_MULTIPLIER, 0);
        Matrix3x3 currentRotation;
        mCamera.getLocator().get().getSubMatrix(currentRotation);
        relative = currentRotation*relative;
        mCenter += relative;

        changed = true;
    }

    Float zoom = theMouse().getWheelDelta()*mSensitivity;
    
    if (zoom == 0 && theMouse().isPressed(Mouse::MIDDLE))
    {
        zoom = pos[0]*MOUSE_ZOOM_MULTIPLIER;
    }

    if (zoom != 0)
    {
        mDistance *= (1+WHEEL_ZOOM_MULTIPLIER*mInvert*zoom);
        if (mDistance < 1.0f)
        {
            mDistance = 1.0f;
        }
        changed = true;
    }

    if (changed)
    {
        Vector3 at;
        mCamera.getLocator().get().getColumn(MatrixBase::AT, at);

        Vector3 translation = mCenter - at*mDistance;
        mCamera.getLocator().get().setColumn(MatrixBase::TRANSLATION, translation);

        mCamera.getLocator().update();
    }
}

//------------------------------------------------------------------------------
void ModelViewCameraController::recalculate()
//------------------------------------------------------------------------------
{
    Vector3 from;
    mCamera.getLocator().get().getColumn(MatrixBase::TRANSLATION, from);

    Vector3 at;
    mCamera.getLocator().get().getColumn(MatrixBase::AT, at);

    mCenter = from + at*mDistance;
}


//------------------------------------------------------------------------------
Float ModelViewCameraController::getDistance() const
//------------------------------------------------------------------------------
{
    return mDistance;
}


//------------------------------------------------------------------------------
void ModelViewCameraController::setDistance(Float d)
//------------------------------------------------------------------------------
{
    mDistance = d;
}

//------------------------------------------------------------------------------
bool ModelViewCameraController::isInverted() const
//------------------------------------------------------------------------------
{
    return mInvert == 1 ? true : false;
}

//------------------------------------------------------------------------------
void ModelViewCameraController::setInvert(bool invert)
//------------------------------------------------------------------------------
{
    mInvert = invert ? 1 : -1;
}

//------------------------------------------------------------------------------
void ModelViewCameraController::setSensitivity(Float s)
//------------------------------------------------------------------------------
{
    mSensitivity = s;
}

//------------------------------------------------------------------------------
Float ModelViewCameraController::getSensitivity() const
//------------------------------------------------------------------------------
{
    return mSensitivity;
}

} // namespace Ge
