#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Aug-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "MorphFrame.h"

#include <Math/Interpolation.h>

namespace Ge
{

//------------------------------------------------------------------------------
MorphFrame::MorphFrame() :
//------------------------------------------------------------------------------
mVertices(0),
mNormals(0),
mFacePlanes(0),
mBound(Vector3(0,0,0), 0),
mTangents(0)
{
}

//------------------------------------------------------------------------------
MorphFrame::~MorphFrame()
//------------------------------------------------------------------------------
{
    if (mVertices != 0)
    {
        GE_DELETE_A mVertices;
    }

    if (mNormals != 0)
    {
        GE_DELETE_A mNormals;
    }

    if (mFacePlanes != 0)
    {
        GE_DELETE_A mFacePlanes;
    }

    if (mTangents != 0)
    {
        GE_DELETE_A mTangents;
    }
}

//------------------------------------------------------------------------------
void MorphFrame::setVertices(Vector3* v)
//------------------------------------------------------------------------------
{
    if (mVertices != 0)
    {
        GE_DELETE_A mVertices;
    }

    mVertices = v;
}


//------------------------------------------------------------------------------
const Vector3* MorphFrame::getVertices() const
//------------------------------------------------------------------------------
{
    return mVertices;
}


//------------------------------------------------------------------------------
void MorphFrame::setNormals(Vector3* n)
//------------------------------------------------------------------------------
{
    if (mNormals != 0)
    {
        GE_DELETE_A mNormals;
    }

    mNormals = n;
}

//------------------------------------------------------------------------------
const Vector3* MorphFrame::getNormals() const
//------------------------------------------------------------------------------
{
    return mNormals;
}

//------------------------------------------------------------------------------
void MorphFrame::calculateBound(UInt vertexCount)
//------------------------------------------------------------------------------
{
    mBound.merge(mVertices[0], vertexCount);
}

//------------------------------------------------------------------------------
const Sphere& MorphFrame::getBound() const
//------------------------------------------------------------------------------
{
    return mBound;
}

//------------------------------------------------------------------------------
void MorphFrame::calculateFacePlanes(UInt triangleCount, UInt16* indices)
//------------------------------------------------------------------------------
{
    if (mFacePlanes != 0)
    {
        GE_DELETE_A mFacePlanes;
    }

    mFacePlanes = GE_NEW Vector4[triangleCount];

    Vector3 faceNormal;
    Vector3 v0v1, v0v2; // Edges
    UInt16* triangle;
    for (UInt i = 0; i<triangleCount; ++i)
    {
        triangle = &(indices[i*3]);

        // Calculate face normal
        v0v1 = getVertices()[triangle[1]];
        v0v1 -= getVertices()[triangle[0]];
        v0v2 = getVertices()[triangle[2]];
        v0v2 -= getVertices()[triangle[0]];
        faceNormal = cross(v0v1, v0v2);
        set(mFacePlanes[i], faceNormal[0], faceNormal[1], faceNormal[2],
                            -faceNormal*getVertices()[triangle[0]]);
    }
}

//------------------------------------------------------------------------------
void MorphFrame::flipTangents(UInt vertexCount)
//------------------------------------------------------------------------------
{
    if (mTangents != 0)
    {
        for (UInt i = 0; i<vertexCount; ++i)
        {
            mTangents[i](1,0) = -mTangents[i](1,0);
            mTangents[i](1,1) = -mTangents[i](1,1);
            mTangents[i](1,2) = -mTangents[i](1,2);
        }
    }
}

//------------------------------------------------------------------------------
const Vector4* MorphFrame::getFacePlanes() const
//------------------------------------------------------------------------------
{
    return mFacePlanes;
}

//------------------------------------------------------------------------------
void MorphFrame::setTangents(Matrix3x3* t)
//------------------------------------------------------------------------------
{
    GE_ASSERT(t != 0);
    if (t != 0)
    {
        if (mTangents != 0)
        {
            GE_DELETE_A mTangents;
        }

        mTangents = t;
    }
}

//------------------------------------------------------------------------------
void MorphFrame::lerp(UInt vertexCount, UInt triangleCount,
                      const MorphFrame& from, const MorphFrame& to, Float s)
//------------------------------------------------------------------------------
{
    // Reserve space for vertices if necessary
    if (mVertices == 0)
    {
        mVertices = GE_NEW Vector3[vertexCount];
    }

// This macro checks whether a specific component should be lerped and reserves
// memory for it if there isn't one already.
// in: <component>: component to check
//     <Type>: type of component
//     <count>: number of elements for component
// out: <shouldLerp>: should this component be lerped?
#define reserveMemoryFor(component, Type, count, shouldLerp) \
{ \
    bool fromHas##component = from.get##component() != 0; \
    bool toHas##component = to.get##component() != 0; \
    GE_ASSERT( (fromHas##component && toHas##component ) || (!fromHas##component && !toHas##component) ); \
    bool lerp##component = fromHas##component && toHas##component; \
    if (lerp##component && m##component == 0) \
    { \
        m##component = GE_NEW Type[count]; \
    } \
    shouldLerp = lerp##component; \
}

    bool shouldLerpNormals;
    reserveMemoryFor(Normals, Vector3, vertexCount, shouldLerpNormals);

    bool shouldLerpFacePlanes;
    reserveMemoryFor(FacePlanes, Vector4, triangleCount, shouldLerpFacePlanes);
    
    bool shouldLerpTangents;
    reserveMemoryFor(Tangents, Matrix3x3, vertexCount, shouldLerpTangents);

    GE_ASSERT(0<=s && s<=1);
    clamp(s, 0.0f, 1.0f);
    if (s == 0)
    {
        copyFrom(vertexCount, triangleCount, from,
                 shouldLerpNormals, shouldLerpFacePlanes, shouldLerpTangents);
    }
    else if (s == 1)
    {
        copyFrom(vertexCount, triangleCount, to,
                 shouldLerpNormals, shouldLerpFacePlanes, shouldLerpTangents);
    }
    else
    {
        // Lerp vertices
        for (UInt i = 0; i<vertexCount; ++i)
        {
            ::Ge::lerp(mVertices[i], from.getVertices()[i], to.getVertices()[i], s);
        }

        // Lerp normals
        if (shouldLerpNormals)
        {
            for (UInt i = 0; i<vertexCount; ++i)
            {
                ::Ge::lerp(mNormals[i], from.getNormals()[i], to.getNormals()[i], s);
            }
        }

        // Lerp face planes
        if (shouldLerpFacePlanes)
        {
            for (UInt i = 0; i<triangleCount; ++i)
            {
                ::Ge::lerp(mFacePlanes[i], from.getFacePlanes()[i], to.getFacePlanes()[i], s);
            }
        }

        // Lerp bounding sphere
        Vector3 center;
        ::Ge::lerp(center, from.getBound().getCenter(), to.getBound().getCenter(), s);
        mBound.setCenter(center);
        mBound.setRadius(::Ge::lerp(from.getBound().getRadius(), to.getBound().getRadius(), s));

        // Lerp tangents
        if (shouldLerpTangents)
        {
            for (UInt i = 0; i<vertexCount; ++i)
            {
                ::Ge::lerp(mTangents[i], from.getTangents()[i], to.getTangents()[i], s);
            }
        }
    }
}

//------------------------------------------------------------------------------
void MorphFrame::copyFrom(UInt vertexCount, UInt triangleCount,
                          const MorphFrame& source,
                          bool normals, bool facePlanes, bool tangents)
//------------------------------------------------------------------------------
{
    memcpy(mVertices, source.mVertices, vertexCount*sizeof(Vector3));
    
    if (normals)
    {
        memcpy(mNormals, source.mNormals, vertexCount*sizeof(Vector3));
    }

    if (facePlanes)
    {
        memcpy(mFacePlanes, source.mFacePlanes, triangleCount*sizeof(Vector4));
    }

    if (tangents)
    {
        memcpy(mTangents, source.mTangents, vertexCount*sizeof(Matrix3x3));
    }

    mBound = source.mBound;
}

//------------------------------------------------------------------------------
const Matrix3x3* MorphFrame::getTangents() const
//------------------------------------------------------------------------------
{
    return mTangents;
}

} // namespace Ge
