#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 06-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include "Cluster.h"
#include "Link.h"

#include <Engine/ObjImporter.h>
#include <Engine/Mesh.h>

#include <System/StringTools.h>
#include <System/FilePath.h>
#include <System/Log.h>

#include <fstream>
#include <map>

namespace
{

using namespace Ge;

//------------------------------------------------------------------------------
Vector3 stringToVector3(const std::string& str)
//------------------------------------------------------------------------------
{
    std::string x;
    std::string y;
    std::string z;
    std::string buffer;
    StringTools::split(str, x, buffer);
    StringTools::split(buffer, y, z);
    // Reverse z-coordinate, because .obj uses a right handed coordinate system
    return Vector3(StringTools::toFloat(x), StringTools::toFloat(y),
                   -(StringTools::toFloat(z)));
}

//------------------------------------------------------------------------------
Vector2 stringToVector2(const std::string& str)
//------------------------------------------------------------------------------
{
    std::string x;
    std::string y;
    std::string z;
    std::string buffer;
    StringTools::split(str, x, buffer);
    StringTools::split(buffer, y, z);
    return Vector2(StringTools::toFloat(x), StringTools::toFloat(y));
}

// Face entries in an .obj file are of the form
// - vertex or
// - vertex//normal or
// - vertex/texture uv/normal
// This function splits an entry into the components
//------------------------------------------------------------------------------
void entryToIndices(const std::string& entry,
                    UInt& vIndex, UInt& nIndex, UInt& tIndex)
//------------------------------------------------------------------------------
{
    Int pos(entry.find('/'));
    if (pos == -1)
        vIndex = StringTools::toInt(entry) - 1;
    else
    {
        std::string v;
        std::string t;
        std::string n;
        v = entry.substr(0, pos);
        t = entry.substr(pos+1);
        vIndex = StringTools::toInt(v) - 1;

        pos = t.find('/');
        // Either there is no slash or two slashes
        GE_ASSERT(pos != -1);
        n = t.substr(pos+1);

        if (pos != 0)
        {
            t = t.substr(0, pos);
            tIndex = StringTools::toInt(t) - 1;
        }

        nIndex = StringTools::toInt(n) - 1;
    }
}       

// This is used by splitData(). <mMapped> specifies, whether this vertex
// (= vertex/normal/texture uv) was already mapped to a new index due to
// the data split. <mIndex> contains the new index that the old index was
// mapped to. If an index wasn't mapped, then the vertex, normal,... has
// to be added to the new MeshData 
struct Index
{
    Index() : mMapped(false), mIndex(0) {}

    bool mMapped;
    UInt16 mIndex;
};

// This struct is used in splitData() to index the std::map of already mapped
// indices. A vertex has to be duplicated, if either a new normal or a new
// texture coordinate was encountered (e.g. mapping of a texture to all 6 sides
// of a cube would need to have triple copies of each vertex, because for
// each side of the cube other texture coordinates are needed at the vertices)
struct VertexData
{
    Vector3 mVertex;
    Vector3 mNormal;
    Vector2 mTextureUv;

    struct Pred
    {
        inline
        bool operator ()(const VertexData& lhs, const VertexData& rhs) const
        {
             Vector3::Pred pred3;
             Vector2::Pred pred2;
             if (pred3(lhs.mVertex, rhs.mVertex))
                 return true;
             if (pred3(lhs.mNormal, rhs.mNormal))
                 return true;
             if (pred2(lhs.mTextureUv, rhs.mTextureUv))
                 return true;

             return false;
        }
    };
};

} // namespace











namespace Ge
{

//------------------------------------------------------------------------------
ObjImporter::ObjImporter(const char* fn, TexturePool& pool) :
//------------------------------------------------------------------------------
mNormals(0),
mTextureUvs(0),
mMaterialLib(pool),
mLocator(Locator::create(Matrix3x4::IDENTITY()))
{
    FilePath fileName(fn);
    GE_ASSERT_MSG(fn != 0, "No file name provided");
    GE_ASSERT_MSG(fileName.getExtension() == ".obj", "Wrong extension");

    
    // Some .obj files have no mtllib tag -> try <materlialLibName> if present or
    // lib with same name as .obj
    bool materialsLoaded(false);

    std::ifstream file(fileName.get().c_str(), std::ios::in);
    Int index(-1);
    // Could file be opened?
    if (file)
    {
        mName = fileName.getName();
        while (waitForTag(file))
        {
            if ("usemtl" == mCurrentLine.getTag())
            {
                // "usemtl" means a new sub mesh was encountered
                index++;
                mSubsets.push_back(Subset(mCurrentLine.getRest().c_str()));
                mCurrentLine.clear();
            }
            else if ("v" == mCurrentLine.getTag())
                readVertices(file);
            else if ("vn" == mCurrentLine.getTag())
                readNormals(file);
            else if ("vt" == mCurrentLine.getTag())
                readTextureUvs(file);
            else if ("f" == mCurrentLine.getTag())
                readIndices(index, file);
            else if ("mtllib" == mCurrentLine.getTag())
            {
                mMaterialLib.import((fileName.getDirectory() +
                                     mCurrentLine.getRest()).c_str());
                materialsLoaded = true;
                mCurrentLine.clear();
            }
            else
                GE_ASSERT(false);
        }
    }
    else
    {
        GE_ASSERT_MSG(false, "File could not be accessed");
        GE_LOG_ERROR("File "<<fn<<" could not be accessed");
    }

    if (index > -1 && !materialsLoaded)
    {
        fileName.setExtension(".mtl");
        mMaterialLib.import(fileName.get().c_str());
    }

    splitData();
}

//------------------------------------------------------------------------------
Cluster* ObjImporter::import(bool importMaterialAndTexture,
                             bool recalculateNormals,
                             bool flipUvs) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(mMeshes.size() > 0);
    if (mMeshes.size() > 0)
    {
        Cluster* ret = GE_NEW Cluster;
        // For each material create a mesh and group them under a Node
        for (UInt i = 0; i<mMeshes.size(); i++)
        {
            ret->add(*getLink(i, importMaterialAndTexture,
                              recalculateNormals, flipUvs));
        }

        ret->set(*mLocator);
        return ret;
    }
    else
    {
        return 0;
    }
}

//------------------------------------------------------------------------------
void ObjImporter::splitData()
//------------------------------------------------------------------------------
{
    // This map stores a mapping of old to new index
    std::map<VertexData, Index, VertexData::Pred> indexMap;
    VertexData current;
    // For each subset build a MeshData struct that holds the vertex data and
    // remapped indices
    for (UInt subset = 0; subset<mSubsets.size(); subset++)
    {
        mMeshes.push_back(MeshData());
        indexMap.clear();
        for (UInt j = 0; j<mSubsets[subset].mVertexIndices.size(); j++)
        {
            // Remember the indices of the currently visited vertex
            current.mVertex = mVertices[mSubsets[subset].mVertexIndices[j]];
            if (mNormals.size() > 0)
                current.mNormal = mNormals[mSubsets[subset].mNormalIndices[j]];
            if (mTextureUvs.size() > 0)
                current.mTextureUv = mTextureUvs[mSubsets[subset].mTextureUvIndices[j]];

            // If an index wasn't mapped yet this means, that the vertex is
            // encountered the first time -> add it to mMeshes
            if (!indexMap[current].mMapped)
            {
                GE_ASSERT(mVertices.size() > 0);
                mMeshes[subset].mVertices.push_back(
                    current.mVertex);
                if (mNormals.size() > 0)
                    mMeshes[subset].mNormals.push_back(
                        current.mNormal);
                if (mTextureUvs.size() > 0)
                    mMeshes[subset].mTextureUvs.push_back(
                        current.mTextureUv);
                indexMap[current].mIndex =
                    mMeshes[subset].mVertices.size()-1;
                indexMap[current].mMapped = true;
            }
            
            mMeshes[subset].mIndices.push_back(indexMap[current].mIndex);
        }
    }
}

//------------------------------------------------------------------------------
Link* ObjImporter::getLink(UInt index,
                           bool importMaterialAndTexture,
                           bool recalculateNormals,
                           bool flipUvs) const
//------------------------------------------------------------------------------
{
    // Reserve memory
    Vector3* normals = 0;
    Vector2* textureUvs = 0;
    if (mMeshes[index].mNormals.size() > 0)
    {
        UInt normalCount = mMeshes[index].mNormals.size();
        normals = GE_NEW Vector3[normalCount];
        memcpy(normals, &mMeshes[index].mNormals[0], sizeof(Vector3)*normalCount);
    }
    if (mMeshes[index].mTextureUvs.size() > 0)
    {
        UInt uvCount = mMeshes[index].mTextureUvs.size();
        textureUvs = GE_NEW Vector2[uvCount];
        memcpy(textureUvs, &mMeshes[index].mTextureUvs[0], sizeof(Vector2)*uvCount);
    }
    Vector3* vertices = GE_NEW Vector3[mMeshes[index].mVertices.size()];
    memcpy(vertices, &mMeshes[index].mVertices[0],
           sizeof(Vector3)*mMeshes[index].mVertices.size());
    UInt16* indices = GE_NEW UInt16[mMeshes[index].mIndices.size()];
    memcpy(indices, &mMeshes[index].mIndices[0],
           sizeof(UInt16)*mMeshes[index].mIndices.size());

    
    // Construct the mesh
    Mesh* mesh = GE_NEW Mesh(
        1,
        mMeshes[index].mVertices.size(),
        vertices,
        normals,
        textureUvs,
        0,
        mMeshes[index].mIndices.size(),
        indices);

    // Calculate bounding volume
    mesh->calculateBound();

    // Recalculate normals
    if (recalculateNormals)
    {
        mesh->calculateNormals(true, true);
    }

    // Add textures and materials to the mesh
    Link* ret = GE_NEW Link(*mLocator, *mesh);
    if (mSubsets[index].mName != "Default")
    {
        Material* material =
            mMaterialLib.getMaterial(mSubsets[index].mName.c_str());
        
        if (material != 0)
        {
            Material& mat = *material;
            if (mat.getTexture(0) != 0)
            {
                if (!importMaterialAndTexture)
                {
                    mat.set(Color::WHITE);
                }

                if (mat.getTexture(0)->hasAlpha())
                {
                    StateBlock b = ret->getStateBlock();
                    b.set(RenderState::BLEND_SOURCE_ALPHA);
                    ret->set(b);
                }
            }

            ret->set(mat);
        }
    }

    if (flipUvs)
    {
        mesh->flipUvs(0);
    }

    return ret;
}

//------------------------------------------------------------------------------
bool ObjImporter::waitForTag(std::istream& file)
//------------------------------------------------------------------------------
{
    // Wait for a known tag
    while (!file.eof() &&
           "usemtl" != mCurrentLine.getTag() &&
           "v" != mCurrentLine.getTag() &&
           "vn" != mCurrentLine.getTag() &&
           "vt" != mCurrentLine.getTag() &&
           "f" != mCurrentLine.getTag() &&
           "mtllib" != mCurrentLine.getTag())
           mCurrentLine.read(file);

    // !file.eof() means a tag was found
    return !file.eof();
}

//------------------------------------------------------------------------------
void ObjImporter::readVertices(std::istream& file)
//------------------------------------------------------------------------------
{
    if (waitForTag(file))       
        while ("v" == mCurrentLine.getTag() && !file.eof())
        {
            mVertices.push_back(
                stringToVector3(mCurrentLine.getRest()));
            mCurrentLine.read(file);
        }
}

//------------------------------------------------------------------------------
void ObjImporter::readNormals(std::istream& file)
//------------------------------------------------------------------------------
{
    if (waitForTag(file))
        while ("vn" == mCurrentLine.getTag() && !file.eof())
        {
            mNormals.push_back(
                stringToVector3(mCurrentLine.getRest()));
            mCurrentLine.read(file);
        }
}

//------------------------------------------------------------------------------
void ObjImporter::readTextureUvs(std::istream& file)
//------------------------------------------------------------------------------
{
    if (waitForTag(file))
        while ("vt" == mCurrentLine.getTag() && !file.eof())
        {
            mTextureUvs.push_back(
                stringToVector2(mCurrentLine.getRest()));
            mCurrentLine.read(file);
        }
}

//------------------------------------------------------------------------------
void ObjImporter::readIndices(UInt index, std::istream& file)
//------------------------------------------------------------------------------
{
    if (waitForTag(file))
        while ("f" == mCurrentLine.getTag() && !file.eof())
        {
            parseFace(index, mCurrentLine.getRest());
            mCurrentLine.read(file);
        }   

    int a = mSubsets[index].mVertexIndices.size();
}

//------------------------------------------------------------------------------
void ObjImporter::pushIndices(UInt index, UInt vertex, UInt normal, UInt texture)
//------------------------------------------------------------------------------
{
    mSubsets[index].mVertexIndices.push_back(vertex);

    if (mNormals.size() > 0)
        mSubsets[index].mNormalIndices.push_back(normal);

    if (mTextureUvs.size() > 0)
        mSubsets[index].mTextureUvIndices.push_back(texture);
}

//------------------------------------------------------------------------------
void ObjImporter::parseFace(UInt index, const std::string& face)
//------------------------------------------------------------------------------
{
    std::string entry;
    std::string rest(face);

    // These buffers store the vertex, normal and textureUv index of a triangle
    UInt vBuffer[3], nBuffer[3], tBuffer[3];
    for (UInt i = 0; i<3; i++)
    {
        GE_ASSERT_MSG(rest != "",
            "A face with less then 3 vertices is not possible");
        StringTools::split(rest, entry, rest);
        entryToIndices(entry, vBuffer[i], nBuffer[i], tBuffer[i]);
    }
    
    // Because .obj uses a right handed system -> reverse the triangle order
    pushIndices(index, vBuffer[0], nBuffer[0], tBuffer[0]);
    pushIndices(index, vBuffer[2], nBuffer[2], tBuffer[2]);
    pushIndices(index, vBuffer[1], nBuffer[1], tBuffer[1]);

    // If there still is a rest, then it's a multi-vertex face
    while (rest != "")
    {
        StringTools::split(rest, entry, rest);
        entryToIndices(entry, vBuffer[1], nBuffer[1], tBuffer[1]);

        // Triangulate the face (I build a triangle fan for multi-vertex faces
        // taking the first vertex as a starting point)
        pushIndices(index, vBuffer[0], nBuffer[0], tBuffer[0]);
        pushIndices(index, vBuffer[1], nBuffer[1], tBuffer[1]);
        pushIndices(index, vBuffer[2], nBuffer[2], tBuffer[2]);
        vBuffer[2] = vBuffer[1];
        nBuffer[2] = nBuffer[1];
        tBuffer[2] = tBuffer[1];
    }
}

} // namespace Ge






//------------------------------------------------------------------------------
// unitTest()
//------------------------------------------------------------------------------

#include <iostream>

namespace Ge
{

//------------------------------------------------------------------------------
void ObjImporter::unitTest()
//------------------------------------------------------------------------------
{
    std::cout<<"ObjImporter::unitTest()\n";

    std::string face("1/2/3");
    UInt v(-1), n(-1), t(-1);
    entryToIndices(face, v, n, t);
    GE_ASSERT(v == 0);
    GE_ASSERT(n == 2);
    GE_ASSERT(t == 1);

    face = "1//3";
    v = n = t = -1;
    entryToIndices(face, v, n, t);
    GE_ASSERT(v == 0);
    GE_ASSERT(n == 2);
    GE_ASSERT(t == -1);

    face = "1";
    v = n = t = -1;
    entryToIndices(face, v, n, t);
    GE_ASSERT(v == 0);
    GE_ASSERT(n == -1);
    GE_ASSERT(t == -1);

    std::cout<<"Test OK\n\n";
}

} // namespace Ge
