//------------------------------------------------------------------------------
// (c) 06-2001 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef ENGINE_OBJ_IMPORTER_H
#define ENGINE_OBJ_IMPORTER_H

#include "MtlImporter.h"
#include "Locator.h"

#include <Math/Vector.h>

#include <System/TagLine.h>

#include <vector>
#include <string>

namespace Ge
{

class Cluster;
class Link;
class TexturePool;

// This class cannot be used directly. Use the ModelPool instead.
// .obj files contain a global list of vertices, normals and texture
// coordinates. Further separate lists of faces for each of material are
// provided. Faces that have more than three vertices are allowed. The
// file format uses a right handed coordinate system that has to be converted
// into the engine's left handed system. The face lists index into the global
// vertex/normal/texture arrays and sharing of e.g. a vertex across different
// sub meshes belonging to different materials is allowed.
// So reading an .obj file needs reading the vector arrays and face lists
// and then splitting the data up into one Mesh per material, where shared
// vertices/normals/texture coordinates have to be duplicated.
class ObjImporter
{
public:
    // Transforms to a model. If <importMaterialAndTexture> is true, then material
    // information will also be imported, when there is a texture.
    // Otherwise only the texture will be imported.
    Cluster* import(bool importMaterialAndTexture,
                    bool recalculateNormals,
                    bool flipUvs) const;

    // For testing only
    static
    void unitTest();

private:
    friend class ModelPool;
    
    ObjImporter(const char* fileName, TexturePool&);

    struct Subset
    {
        Subset(const char* name) : mName(name) {}
        std::string mName;
        std::vector<UInt16> mVertexIndices;
        std::vector<UInt16> mNormalIndices;
        std::vector<UInt16> mTextureUvIndices;
    };

    struct MeshData
    {
        std::vector<Vector3> mVertices;
        std::vector<Vector3> mNormals;
        std::vector<Vector2> mTextureUvs;
        std::vector<UInt16> mIndices;
    };

    void readVertices(std::istream&);
    void readNormals(std::istream&);
    void readTextureUvs(std::istream&);
    void readIndices(UInt index, std::istream&);
    void parseFace(UInt index, const std::string&);
    void pushIndices(UInt index,
                     UInt vertex, UInt normal, UInt texture);
    // Returns false if eof reached. Otherwise reads line after line
    // until one of the known tags is reached in the file
    bool waitForTag(std::istream&);
    Link* getLink(UInt index, bool importMaterialAndTexture,
                  bool recalculateNormals,
                  bool flipUvs) const;
    // Split vertex data into MeshData according to the indices in
    // mSubsets
    void splitData();

    std::vector<Subset> mSubsets;
    std::vector<MeshData> mMeshes;
    std::vector<Vector3> mVertices;
    std::vector<Vector3> mNormals;
    std::vector<Vector2> mTextureUvs;
    TagLine mCurrentLine;
    std::string mName;
    Pointer<Locator> mLocator;
    // Must be mutable, because there is no const reference std::map::operator[]
    mutable MtlImporter mMaterialLib;
};

} // namespace Ge

#endif // ENGINE_OBJ_IMPORTER_H