#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include "Renderer.h"
#include "World.h"
#include "Camera.h"
#include "Link.h"
#include "Cluster.h"
#include "Light.h"

#include <Math/Matrix.h>

#include <System/Log.h>

#include <utility>


namespace Ge
{

//------------------------------------------------------------------------------
Pointer<Renderer>& Renderer::instancePtr()
//------------------------------------------------------------------------------
{
    static Pointer<Renderer> r(0);
    return r;
}

//------------------------------------------------------------------------------
Renderer::Renderer(/*Sorter& sorter, */const Color& background) :
//------------------------------------------------------------------------------
mCamera(0),
//mSorter(&sorter),
mBackgroundColor(background)
{
}

//------------------------------------------------------------------------------
void Renderer::render(const World& world)
//------------------------------------------------------------------------------
{
    // ToDo: Actually the algorithm should be:
    // - Find meshes in frustum
    // - Find lights for those meshes
    // - Sort meshes by e.g. material
    // - Render meshes with lights
    // - Render shadow volumes
    // - Render shadow quad over screen

    // Set the camera

    if (world.getCamera() != 0)
    {
        set(*world.getCamera());
    }
    else
    {
        static Camera camera(*Locator::create(Matrix3x4::IDENTITY()));
        set(camera);
        GE_ASSERT_MSG(false, "Every world needs a camera");
    }
    

    // Set the lights

    for (UInt l = 0; l<world.getLights().size(); ++l)
    {
        add(*world.getLights()[l]);
    }

    
    // Collect all visible Links into a list.
    
    const std::vector<Pointer<const Cluster> >& clusters = world.getClusters();
    VolumeVector volumes;
    std::vector<Link*> visibleLinks;
    for (UInt c = 0; c<clusters.size(); ++c)
    {
        bool clusterVisible =
            mCamera->getWorldFrustum().intersects(clusters[c]->getBound()) !=
            Intersect::DISJOINT;
        
        const LinkVector& links = clusters[c]->getLinks();
        collect(links, volumes, clusterVisible, visibleLinks);
    }
    
    const LinkVector& links = world.getLinks();
    collect(links, volumes, true, visibleLinks);

    // Render the visible Links.

    if (volumes.size() == 0)
    {
        render(visibleLinks);
//        render(visibleLinks, volumes);
    }
    else
    {
//        render(visibleLinks);
        render(visibleLinks, volumes);
    }
}

//------------------------------------------------------------------------------
void Renderer::collect(const LinkVector& links, VolumeVector& volumes,
                       bool clusterVisible,
                       std::vector<Link*>& visibleLinks)
//------------------------------------------------------------------------------
{
    for (UInt l = 0; l<links.size(); ++l)
    {
        if (clusterVisible &&
            mCamera->getWorldFrustum().intersects(links[l]->getBound()))
        {
            visibleLinks.push_back(links[l].get());
        }
        
        if (links[l]->isShadowCaster())
        {
            links[l]->updateShadowVolume();
            volumes.push_back(std::make_pair(
                &links[l]->getLocator()->getGlobal(),
                links[l]->getShadowVolume()
                ));
        }
    }
}

//------------------------------------------------------------------------------
void Renderer::render(const std::vector<Link*>& visibleLinks)
//------------------------------------------------------------------------------
{
    static Material defaultMaterial;
    for (UInt l = 0; l<visibleLinks.size(); ++l)
    {
        setWorldMatrix(visibleLinks[l]->getLocator()->getGlobal());
        
        Material* material = visibleLinks[l]->getMaterial() != 0 ?
            visibleLinks[l]->getMaterial() : 
            &defaultMaterial;
        material->setLink(*visibleLinks[l]);
        set(*material);
        
        set(visibleLinks[l]->getStateBlock());
        
        render(*visibleLinks[l]->getMesh());
    }
}

//------------------------------------------------------------------------------
void Renderer::render(const std::vector<Link*>& visibleLinks,
                      const VolumeVector& volumes)
//------------------------------------------------------------------------------
{
    static Material defaultMaterial;
    for (UInt l = 0; l<visibleLinks.size(); ++l)
    {
        setWorldMatrix(visibleLinks[l]->getLocator()->getGlobal());
        
        Material* material = visibleLinks[l]->getMaterial() != 0 ?
            visibleLinks[l]->getMaterial() : 
            &defaultMaterial;
        material->setLink(*visibleLinks[l]);
        material->setPreShadow(true);
        set(*material);
        
        set(visibleLinks[l]->getStateBlock());
        render(*visibleLinks[l]->getMesh());

        material->setPreShadow(false);
    }

    // Render the shadow volumes
    
    for (UInt j = 0; j<volumes.size(); ++j)
    {
        setWorldMatrix(*(volumes[j].first));
        render(*(volumes[j].second));
    }

    {for (UInt l = 0; l<visibleLinks.size(); ++l)
    {
        setWorldMatrix(visibleLinks[l]->getLocator()->getGlobal());
        
        Material* material = visibleLinks[l]->getMaterial() != 0 ?
            visibleLinks[l]->getMaterial() : 
            &defaultMaterial;
        material->setLink(*visibleLinks[l]);
        material->setPostShadow(true);
        set(*material);
        
        set(visibleLinks[l]->getStateBlock());
        render(*visibleLinks[l]->getMesh());

        material->setPostShadow(false);
    }}
}

//------------------------------------------------------------------------------
const Camera* Renderer::getCamera() const
//------------------------------------------------------------------------------
{
    return mCamera;
}

//------------------------------------------------------------------------------
void Renderer::set(const Camera& camera)
//------------------------------------------------------------------------------
{
    mCamera = &camera;
    Matrix4x4 m;
    getCamera()->getViewMatrix(m);
    setViewMatrix(m);
    getCamera()->getProjectionMatrix(m);
    setProjectionMatrix(m);
}

//------------------------------------------------------------------------------
const Color& Renderer::getBackgroundColor() const
//------------------------------------------------------------------------------
{
    return mBackgroundColor;
}

} // namespace Ge
