//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef ENGINE_RENDERER_H
#define ENGINE_RENDERER_H

#include <System/Object.h>
#include <System/Pointer.h>

#include <Engine/Color.h>
#include <Engine/StateBlock.h>
#include <Engine/Link.h>

#include <Math/Vector.h>
#include <Math/Matrix.h>

namespace Ge
{

class World;
class Light;
class Material;
class Mesh;
class Lines;
class DisplayStatistics;
class Camera;
class ShadowVolume;
class Emitter;

class Renderer;
Renderer& theRenderer();

// Abstrace renderer base class. Derive your own renderer class from it.
class Renderer : public Object
{
public:

    // Has to be set by the framework. The Renderer must live on the heap and
    // is the one and only Renderer instance to be used from the application.
    static Pointer<Renderer>& instancePtr();

    //------------------------------------------------------------------------------
    // Rendering
    //------------------------------------------------------------------------------
    virtual void beginScene() = 0;
    virtual void endScene() = 0;
    virtual void render(const World&);
    virtual void render(const Mesh&) = 0;
    virtual void render(const Lines&) = 0;
    virtual void render(const ShadowVolume&) = 0;
    virtual void render(const Emitter&) = 0;


    //------------------------------------------------------------------------------
    // Lights
    //------------------------------------------------------------------------------
    virtual void add(const Light&) = 0;
    virtual UInt getPossibleNumberOfLights() const = 0;
    virtual void setAmbientLight(const Color&) = 0;

    virtual void setWorldMatrix(const Matrix4x4&) = 0;
    virtual void setWorldMatrix(const Matrix3x4&) = 0;
    void set(const Camera&);


    //------------------------------------------------------------------------------
    // Debugging
    //------------------------------------------------------------------------------
    virtual void print(UInt x, UInt y, const char* text) = 0;
    virtual void print(const DisplayStatistics&) = 0;

    //------------------------------------------------------------------------------
    // Render states
    //------------------------------------------------------------------------------

    virtual void set(const StateBlock&) = 0;
    virtual void set(const Material&) = 0;

protected:
    //------------------------------------------------------------------------------
    // Construction
    //------------------------------------------------------------------------------
    // <sorter> must live on the heap
    Renderer(const Color& background = Color(0.2f,0.2f,0.2f,1));
    virtual ~Renderer() {}

    const Color& getBackgroundColor() const;

    const Camera* getCamera() const;

    virtual void setProjectionMatrix(const Matrix4x4&) = 0;

    virtual void setViewMatrix(const Matrix4x4&) = 0;

private:
    typedef std::vector<Pointer<Link> > LinkVector;
    typedef std::vector<std::pair<const Matrix3x4*, const ShadowVolume*> > VolumeVector;
    
    // Collect all visible Links in the world into a <visibleLinks> list and add
    // shadow casters into the <volumes> list.
    void collect(const LinkVector& links, VolumeVector& volumes,
                 bool clusterVisible, std::vector<Link*>& visibleLinks);

    // Render path if there are no shadow casters in the scene.
    void render(const std::vector<Link*>& visibleLinks);

    // Render path if there are shadow casters in the scene.
    void render(const std::vector<Link*>& visibleLinks, const VolumeVector&);

    Color mBackgroundColor;
    const Camera* mCamera;
};

inline
//------------------------------------------------------------------------------
Renderer& theRenderer()
//------------------------------------------------------------------------------
{
    return *(Renderer::instancePtr());
}

} // namespace Ge

#endif // ENGINE_AUDIO_VISUAL_H