#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Okt-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "SphereMesh.h"

#include "Mesh.h"

#include <Math/Vector.h>

namespace Ge
{

//------------------------------------------------------------------------------
Mesh* SphereMesh::create(Float radius, UInt tesselation)
//------------------------------------------------------------------------------
{
    GE_ASSERT(radius > 0);
    if (radius <= 0)
    {
        radius = 1;
    }

    // Take next even number
    tesselation = tesselation&1 ? tesselation+1 : tesselation;
    Float t = tesselation;

    // Number of circles (=t*0.5-1) + north and south pole.
    UInt vertexCount = t*(t*0.5f - 1) + 2;
    GE_ASSERT_MSG(vertexCount < 0xffff, "Tesselation to high.");
    if (vertexCount >= 0xffff)
    {
        return 0;
    }

    // Create normals
    Vector3* normals = GE_NEW Vector3[vertexCount];
    UInt current = 0;
    set(normals[current++], 0.0f, 1.0f, 0.0f);
    Float step = TWO_PI/t;
    for (UInt p = 0; p<(t*0.5-1); ++p)
    {
        Float phi = (p+1)*step;
        Float sp = sin(phi);
        Float cp = cos(phi);
        Float theta = 0;
        for (UInt t = 0; t<tesselation; ++t)
        {
            Float st = sin(theta);
            Float ct = cos(theta);
            set(normals[current++],
                ct*sp,
                cp,
                st*sp);
            theta += step;
        }
    }
    set(normals[current], 0.0f, -1.0f, 0.0f);
    GE_ASSERT(current == vertexCount-1);

    // Create vertices (= normals scaled by the radius)
    Vector3* vertices = GE_NEW Vector3[vertexCount];
    for (UInt v = 0; v<vertexCount; ++v)
    {
        vertices[v] = radius*normals[v];
    }

    // tesselation*#of circles - 1 (= t*0.5-2) quads +
    // triangles at north and south pole
    UInt indexCount = 3*(2*t*(t*0.5f - 2) + 2*t);
    UInt16* indices = GE_NEW UInt16[indexCount];
    UInt i = 0;
    UInt index = 0;

    // Northpole triangles
    for (i = 0; i<t; ++i)
    {
        indices[index++] = 0;
        indices[index++] = 1 + ((i+1)%tesselation);
        indices[index++] = 1 + i;
    }

    // Create quads
    for (i = 0; i<t*0.5f - 2; ++i)
    {
        UInt start = 1 + i*tesselation;
        for (UInt j = 0; j<t; ++j)
        {
            UInt vertex = start + j;
            UInt next = start + ((j+1)%tesselation);
            // Create upper quad triangle
            indices[index++] = vertex;
            indices[index++] = next;
            indices[index++] = next + tesselation;
            // Create lower quad triangle
            indices[index++] = vertex;
            indices[index++] = next + tesselation;
            indices[index++] = vertex + tesselation;
        }
    }

    // Southpole triangles are the reverse of the north pole
    for (i = 0; i<3*t; ++i)
    {
        indices[index++] = vertexCount-1 - indices[i];
    }

    GE_ASSERT(index == indexCount);

    Mesh* sphere = GE_NEW Mesh(
        1,
        vertexCount,
        vertices,
        normals,
        0,
        0,
        indexCount,
        indices);

    sphere->calculateBound();
    return sphere;
}

} // namespace Ge
