#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 06-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include <Engine/Vector3Controller.h>
#include <Math/Interpolation.h>


namespace Ge
{

//------------------------------------------------------------------------------
Vector3Controller::Vector3Controller() :
//------------------------------------------------------------------------------
mCurrentKey(0),
mCurrentTime(0)
{
    mData = GE_NEW Vector3ControllerData;
}

//------------------------------------------------------------------------------
void Vector3Controller::add(const Vector3Key& key)
//------------------------------------------------------------------------------
{
    if (verifyKey(key))
    {
        // If the first added key is not at zero then add a dummy key at zero.
        // Modify first key so that tangentIn is zero (that's the Lightwave
        // behaviour)
        if (mData->getKeyCount() == 0 && key.getTime() != 0)
        {
            Vector3Key startKey(key);
            Vector3 tangentOut(startKey.getTangentOut());
            startKey.setTime(0);
            startKey.setTangentIn(Vector3::ORIGIN());
            startKey.setTangentOut(Vector3::ORIGIN());
            mData->add(startKey);
            startKey.setTangentIn(Vector3::ORIGIN());
            startKey.setTangentOut(tangentOut);
        }
        else
            mData->add(key);
    }
}

//------------------------------------------------------------------------------
const Timer::Unit& Vector3Controller::getDuration() const
//------------------------------------------------------------------------------
{
    static Timer::Unit zero(0);
    // There have to be at least two keys for having a non static animation
    if (getData().getKeyCount() > 1)
        return getData().getKey(getData().getKeyCount()-1).getTime();
    else
        return zero;
}

//------------------------------------------------------------------------------
const Vector3ControllerData& Vector3Controller::getData() const
//------------------------------------------------------------------------------
{
    return *mData;
}

//------------------------------------------------------------------------------
const Timer::Unit Vector3Controller::getCurrentAnimationTime(const Timer::Unit& time)
//------------------------------------------------------------------------------
{
    Timer::Unit ret(time - getOffset());
    GE_ASSERT(ret >= 0);
    if (isReverse())
    {
        ret = getDuration() - ret;
        if (ret < 0)
        {
            reset(time);
            ret = 0;
            flipReverse();
        }
    }
    else
    {
        if (ret > getDuration())
        {
            switch (getRepeatType())
            {
            case REVERSE:
                reset(time);
                flipReverse();
                ret = getDuration();
                break;

            case LOOP:
                reset(time);
                ret = 0;
                break;

            case ONE_SHOT:
                ret = -1;
                break;
            }
        }
    }

    return ret;
}

//------------------------------------------------------------------------------
UInt Vector3Controller::getCurrentKey(const Timer::Unit& animationTime)
//------------------------------------------------------------------------------
{
    if (animationTime < getData().getKey(mCurrentKey).getTime() ||
        animationTime > getData().getKey(mCurrentKey+1).getTime())
    {
        bool keepLooking(true);
        UInt i(0);
        for (i = 0; i<getData().getKeyCount()-1 && keepLooking; i++)
        {
            keepLooking =
                (animationTime < getData().getKey(i).getTime()) ||
                (animationTime > getData().getKey(i+1).getTime());
        }

        mCurrentKey = i-1;
    }
             
    return mCurrentKey;
}

//------------------------------------------------------------------------------
bool Vector3Controller::calcCurrentVector(const Timer::Unit& time,
                                          Vector3& vector)
//------------------------------------------------------------------------------
{
    GE_ASSERT(getData().getKeyCount() > 0);
    if (getData().getKeyCount() == 1)
    {
        vector = getData().getKey(0).getValue();
        return false;
    }
    else
    {
        Timer::Unit animationTime(getCurrentAnimationTime(time));
        if (animationTime != -1)
        {
            UInt current(getCurrentKey(animationTime));
            const Vector3Key& k0 = getData().getKey(current);
            const Vector3Key& k1 = getData().getKey(current+1);
            Float t0(k0.getTime());
            Float t1(k1.getTime());
            Float s(toUnitInterval(t0, t1, animationTime));
            for (UInt i = 0; i<3; i++)
            {
                vector[i] = getInterpolator().interpolate(
                    (k0.getValue())[i], (k1.getValue())[i],
                    (k0.getTangentOut())[i], (k1.getTangentIn())[i],
                    s);
            }
        }
        else
            vector = getData().getKey(getKeyCount()-1).getValue();

        return animationTime != -1;
    }
}

//------------------------------------------------------------------------------
UInt Vector3Controller::getKeyCount() const
//------------------------------------------------------------------------------
{
    return getData().getKeyCount();
}

} // namespace Ge
