//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include <Image/Image.h>

#include <Engine/Color.h>

#include <System/Log.h>
#include <System/FilePath.h>
#include <System/StringTools.h>

#include <zipstream.h>

#include <il/ilu.h>
#include <set>

#ifdef _WIN32
#   include <Windows.h>
#endif

namespace
{

    static struct Il
    {
        Il()
        {
            ilInit();

#ifndef NDEBUG
            mSet.insert(IL_RGB);
            mSet.insert(IL_BGR);
            mSet.insert(IL_RGBA);
            mSet.insert(IL_BGRA);
            mSet.insert(IL_LUMINANCE);
#endif // !NDEBUG
        }

#ifndef NDEBUG
        bool isSupportedFormat(int format) { return mSet.find(format) != mSet.end(); }
        std::set<int> mSet;
#endif // !NDEBUG
        
    } il;

} // namespace

namespace Ge
{

//------------------------------------------------------------------------------
Image::Image(unsigned int width, unsigned int height,
             Format format,
             unsigned char* data) :
//------------------------------------------------------------------------------
mImage(0),
mWidth(width),
mHeight(height),
mFormat(format)
{
    switch (format)
    {
    case RGB_888:
    case BGR_888:
        mBitDepth = 24;
        break;

    case RGBA_8888:
    case BGRA_8888:
        mBitDepth = 32;
        break;

    case LUMINANCE_8:
        mBitDepth = 8;
        break;

    default:
        GE_ASSERT_MSG(false, "Unknown image format");
        mBitDepth = 32;
    }
    
    // Generate image
    ilGenImages(1, &mImage);

    // Bind image name
    ilBindImage(mImage);
    
    ILboolean ok = ilTexImage(width, height, 1, mBitDepth/8, format,
                              IL_UNSIGNED_BYTE, data);

    GE_ASSERT_MSG(ok == IL_TRUE, "Image creation failed");
    if (ok != IL_TRUE)
    {
        GE_LOG_ERROR("Image creation failed"<<std::endl);
    }
}

//------------------------------------------------------------------------------
Image::Image(const char* file) :
//------------------------------------------------------------------------------
mImage(0),
mWidth(0),
mHeight(0),
mBitDepth(0)
{
    // Generate image
    ilGenImages(1, &mImage);

    // Bind image name
    ilBindImage(mImage);

    // Load the file
    ILboolean ok = ilLoadImage(const_cast<char*>(file));
    GE_ASSERT(ok == IL_TRUE);
    if (ok == IL_TRUE)
    {
        mWidth = ilGetInteger(IL_IMAGE_WIDTH);
        mHeight = ilGetInteger(IL_IMAGE_HEIGHT);
        mBitDepth = ilGetInteger(IL_IMAGE_BITS_PER_PIXEL);

        // ### ToDo: This cast is not safe, change it later.
        GE_ASSERT_MSG(il.isSupportedFormat(ilGetInteger(IL_IMAGE_FORMAT)),
            "Picture format not supported yet");
        mFormat = static_cast<Format>(ilGetInteger(IL_IMAGE_FORMAT));
    }
    else
    {
        GE_LOG_ERROR("\""<<file<<"\" not found."<<std::endl);
    }
}

//------------------------------------------------------------------------------
Image::Image(zip::izipfile& zipfile, const char* file) :
//------------------------------------------------------------------------------
mImage(0),
mWidth(0),
mHeight(0),
mBitDepth(0)
{
    zip::izipstream stream(zipfile, file, std::ios_base::binary);
    if (stream)
    {
        Int size = stream.getSize();
        if (size != -1)
        {
            char* buffer = GE_NEW char[size];
            stream.read(buffer, size);

            // Generate image
            ilGenImages(1, &mImage);

            // Bind image name
            ilBindImage(mImage);

            // Load the file
            ILboolean ok = ilLoadL(IL_TYPE_UNKNOWN, buffer, size);
            GE_ASSERT(ok == IL_TRUE);
            if (ok == IL_TRUE)
            {
                mWidth = ilGetInteger(IL_IMAGE_WIDTH);
                mHeight = ilGetInteger(IL_IMAGE_HEIGHT);
                mBitDepth = ilGetInteger(IL_IMAGE_BITS_PER_PIXEL);

                // ### ToDo: This cast is not safe, change it later.
                GE_ASSERT_MSG(il.isSupportedFormat(ilGetInteger(IL_IMAGE_FORMAT)),
                    "Picture format not supported yet");
                mFormat = static_cast<Format>(ilGetInteger(IL_IMAGE_FORMAT));
            }
            else
            {
                GE_LOG_ERROR("\""<<file<<"\" not found."<<std::endl);
            }

            GE_DELETE_A buffer;
        }
    }
}

//------------------------------------------------------------------------------
Image::Image(const Image& im) :
//------------------------------------------------------------------------------
mImage(0),
mWidth(im.mWidth),
mHeight(im.mHeight),
mBitDepth(im.mBitDepth)
{
    ilBindImage(im.mImage);
    mImage = ilCloneCurImage();
}


//------------------------------------------------------------------------------
Image::~Image()
//------------------------------------------------------------------------------
{
    // Delete image
    if (mImage != 0)
        ilDeleteImages(1, &mImage);
}

//------------------------------------------------------------------------------
void Image::save(FileType ft, const char* fileName) const
//------------------------------------------------------------------------------
{
#ifdef _WIN32
    DeleteFile(fileName);
#else
    GE_ASSERT(false);
#endif

    ilBindImage(mImage);
    ILboolean ok = ilSave(ft, const_cast<char*>(fileName));
    GE_ASSERT_MSG(ok != 0, "Error during image saving");
}

//------------------------------------------------------------------------------
void Image::flipX()
//------------------------------------------------------------------------------
{
    ilBindImage(mImage);
    iluFlipImage();
}

//------------------------------------------------------------------------------
UInt8* Image::getData()
//------------------------------------------------------------------------------
{
    ilBindImage(mImage);
    return reinterpret_cast<UInt8*>(ilGetData());
}

//------------------------------------------------------------------------------
const UInt8* Image::getData() const
//------------------------------------------------------------------------------
{
    ilBindImage(mImage);
    return reinterpret_cast<const UInt8*>(ilGetData());
}


//------------------------------------------------------------------------------
void Image::copyTo(UInt8* destination, UInt pitch) const
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(getBitDepth()%8 == 0,
        "An image's bit depth must be divisible by 8");
    UInt bytesPerPixel(getBitDepth()/8);
    GE_ASSERT_MSG(pitch >= getWidth()*bytesPerPixel,
        "The pitch of the destination buffer must at least fit one horizontal line");

    if (pitch >= getWidth()*bytesPerPixel)
    {
        unsigned int bytesPerLine(getWidth()*bytesPerPixel);
        const UInt8* data = getData(); 
        for (UInt i = 0; i<getHeight(); i++)
        {
            memcpy(destination, data, bytesPerLine);
            destination += pitch;
            data += bytesPerLine;
        }
    }
}

//------------------------------------------------------------------------------
void Image::copyFrom(const UInt8* source, UInt pitch)
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(getBitDepth()%8 == 0,
        "An image's bit depth must be divisible by 8");
    UInt bytesPerPixel(getBitDepth()/8);
    GE_ASSERT_MSG(pitch >= getWidth()*bytesPerPixel,
        "The pitch of the destination buffer must at least fit one horizontal line");

    if (pitch >= getWidth()*bytesPerPixel)
    {
        unsigned int bytesPerLine(getWidth()*bytesPerPixel);
        UInt8* data = getData(); 
        for (UInt i = 0; i<getHeight(); i++)
        {
            memcpy(data, source, bytesPerLine);
            source += pitch;
            data += bytesPerLine;
        }
    }
}

//------------------------------------------------------------------------------
void Image::convertTo(Image::Format format)
//------------------------------------------------------------------------------
{
    if (mFormat != format)
    {
        if (mFormat != HSL)
        {
            if (format != HSL)
            {
                ilBindImage(mImage);
                ilConvertImage(format, IL_UNSIGNED_BYTE);
                mBitDepth = ilGetInteger(IL_IMAGE_BITS_PER_PIXEL);
                // ### ToDo: This cast is not safe, change it later.
                mFormat = static_cast<Format>(ilGetInteger(IL_IMAGE_FORMAT));
            }
            else
            {
                switch (format)
                {
                case HSL:
                    convertToHsl();
                    break;

                default:
                    GE_ASSERT_MSG(false, "Unknown image format.");
                }
            }
        }
        else
        {
            switch (mFormat)
            {
            case HSL:
                convertFromHsl(format);
                break;

            default:
                GE_ASSERT_MSG(false, "Unknown image format.");
            }
        }
    }
}

//------------------------------------------------------------------------------
void Image::convertToHsl()
//------------------------------------------------------------------------------
{
    UInt step = getBitDepth()/8;
    UInt8* data = getData();
    for (UInt y = 0; y<getHeight(); ++y)
    {  
        for (UInt x = 0; x<getWidth(); ++x)
        {
            double h(0),s(0),l(0);
            double r,g,b;
            r = double(*data)/255.0;
            g = double(*(data+1))/255.0;
            b = double(*(data+2))/255.0;
       
            if (getFormat() == BGR_888 || getFormat() == BGRA_8888)
            {
                std::swap(r,b);
            }

            RGB_to_HSL(r,g,b,&h,&s,&l);
            
            *data = UInt8(h*255);
            *(data+1) = UInt8(s*255);
            *(data+2) = UInt8(l*255);

            data += step;
        }
    }

    mFormat = HSL;
}

//------------------------------------------------------------------------------
void Image::convertFromHsl(Format format)
//------------------------------------------------------------------------------
{
    UInt step = getBitDepth()/8;
    UInt8* data = getData();
    for (UInt y = 0; y<getHeight(); ++y)
    {
        for (UInt x = 0; x<getWidth(); ++x)
        {
            double h,s,l;
            double r(0),g(0),b(0);
            h = double(*data)/255.0;
            s = double(*(data+1))/255.0;
            l = double(*(data+2))/255.0;
            
            HSL_to_RGB(h,s,l,&r,&g,&b);

            if (format == BGR_888 || format == BGRA_8888)
            {
                std::swap(r,b);
            }
            
            *data = UInt8(r*255);
            *(data+1) = UInt8(g*255);
            *(data+2) = UInt8(b*255);

            data += step;
        }
    }

    mFormat = format;
}

//------------------------------------------------------------------------------
void Image::setLuminance(UInt8 l)
//------------------------------------------------------------------------------
{
    GE_ASSERT(mFormat == HSL);
    if (mFormat == HSL)
    {
        UInt step = getBitDepth()/8;
        UInt8* data = getData();
        data += 2; // jump to first luminance value
        for (UInt y = 0; y<getHeight(); ++y)
        {
            for (UInt x = 0; x<getWidth(); ++x)
            {
                *data = l;
                data += step;
            }
        }
    }
}

//------------------------------------------------------------------------------
Image* Image::createNormalMap(Float unit, Image::Format format) const
//------------------------------------------------------------------------------
{
    Image* ret = GE_NEW Image(getWidth(), getHeight(), RGB_888);

    const UInt8* s = 0;
    Image* newImage = 0;
    if (getFormat() == LUMINANCE_8)
    {
        s = getData();
    }
    else
    {
        newImage = GE_NEW Image(*this);
        newImage->convertTo(LUMINANCE_8);
        s = newImage->getData();
    }

    GE_ASSERT_MSG(s != 0, "No image data for normal map creation");
    if (s)
    {
        UInt8* d = ret->getData();
        // On y axis we take -unit because image data starts in the
        // lower left corner which is (0,1) in texture space. Thus
        // we move up the texture from the bottom in direction (0,0).
        Vector3 p0(0,0,0), px(unit,0,0), py(0,-unit,0);
        Vector3 normal;
        Color pixel;
        for (UInt y = 0; y<getHeight(); ++y)
        {
            for (UInt x = 0; x<getWidth(); ++x)
            {
                if (x == getWidth()-1 || y == getHeight()-1)
                {
                    *d = 0x80; ++d;
                    *d = 0x80; ++d;
                    *d = 0xff; ++d;
                }
                else
                {
                    p0[2] = *s;
                    px[2] = *(s+1);
                    py[2] = *(s+getWidth());
                    normal = cross(py-p0, px-p0);
                    normal.normalize();
                    convertVectorToColor(pixel, normal);

                    *d = pixel.get(Color::CH_RED); ++d;
                    *d = pixel.get(Color::CH_GREEN); ++d;
                    *d = pixel.get(Color::CH_BLUE); ++d;
                }

                ++s;
            }
        }
    }

    if (newImage)
    {
        GE_DELETE newImage;
    }

    ret->convertTo(format);
    return ret;
}

} // namespace Ge