//------------------------------------------------------------------------------
// (c) 04-2001 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef IMAGE_IMAGE_H
#define IMAGE_IMAGE_H

#include <il/il.h>
#include <System/Base.h>
#include <System/Object.h>

namespace zip
{
    class izipfile;
}

namespace Ge
{

// Image has copy semantics and not reference semantics. I.e. copying an image
// copies the image data.
class Image : public Object
{
public:
    // Seems that OpenIL doesn't support 16 bit formats
    enum Format
    {
        RGB_888 = IL_RGB,
        BGR_888 = IL_BGR,
        RGBA_8888 = IL_RGBA,
        BGRA_8888 = IL_BGRA,
        LUMINANCE_8 = IL_LUMINANCE,
        HSL // can also be 32 bit with alpha channel
    };

    enum FileType
    {
        BMP = IL_BMP,
        JPG = IL_JPG,
        TGA = IL_TGA
    };

    // Creates a new image. If data != 0 image data will be copied into the
    // newly created image.
    Image(unsigned int width, unsigned int height,
          Format, unsigned char* data = 0);

    explicit
    Image(const char* file);

    // Load an image out of a .zip-file
    Image(zip::izipfile&, const char* file);
    
    // This copies all image data and is therefore very expensive!
    explicit
    Image(const Image&);

    ~Image();

    void save(FileType, const char* fileName) const;

    UInt getWidth() const;
    
    UInt getHeight() const;
    
    UInt getBitDepth() const;

    UInt8* getData();
    
    const UInt8* getData() const;

    // Copies the image to <destination>. <pitch> denotes the size in bytes
    // of one horizontal line (+stride) in the <destination> buffer.
    void copyTo(UInt8* destination, UInt pitch) const;

    void copyFrom(const UInt8* source, UInt pitch);

    // Flip image at x-axis
    void flipX();

    void convertTo(Format);

    // Create a normal map on the heap using interpreting this image as
    // a height map. If this image is not a greyscale image then it'll be
    // converted to greyscale. You can specify the resulting image <format>.
    // <unit> specifies the steepnes of the resulting normal map.
    // (e.g. unit = 1 interprets the distance in x/y between two pixels
    // as one unit).
    Image* createNormalMap(Float unit, Format = RGB_888) const;
    
    Format getFormat() const;

    // If the format is HSL then set the luminance value to a constant
    void setLuminance(UInt8);

private:
    Image& operator =(const Image&);

    void convertToHsl();

    void convertFromHsl(Format format);

    ILuint mImage;
    UInt mWidth;
    UInt mHeight;
    UInt mBitDepth;
    Format mFormat;
};

// Taken from Graphics Gems
void RGB_to_HSL(double r, double g, double b, double* h, double* s, double* l);

void HSL_to_RGB(double h, double sl, double l, double* r, double* g, double* b);

} // namespace Ge

#include <Image/Image.inl>
#endif // IMAGE_IMAGE_H
