//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef FFT_ROUTINES_H
#define FFT_ROUTINES_H


// This file contains my FFT implementations. You shouldn't use them directly,
// but rather use the Fft classes. Like that you can take advantage of the
// FFTW library.

#include <System/Types.h>

namespace Ge
{

// A collection of fast fourier transform functions.
template<class T>
class FftRoutines
{
public:
    // This function does FFT in one dimension. I've implemented it acording to
    // a paper by Paul Heckbert.
    // <data> contains 2^<sizePower> elements. <stride> is the stride between
    // two array elements.
    // <sign> is the sign of 2*PI*i in the fourier transformation.
    static
    void fft(std::complex<T>* data, unsigned int sizePower, int sign = 1, unsigned int stride = 1);
    // 2D FFT. <data> is assumed to be an array of rows with 2^<rowPower> rows and
    // 2^<columnPower> columns.
    // <sign> is the sign of 2*PI*i in the fourier transformation.
    static
    void fft2D(std::complex<T>* data, unsigned int rowPower, unsigned int columnPower, int sign = 1);


    // This routine calculates a DFT using ordinary sums. I've used it to verify,
    // that my FFT routine is correct. <stride> is the stride between
    // two array elements.
    // <sign> is the sign of 2*PI*i in the fourier transformation.
    static
    void dft(std::complex<T>* data, std::complex<T>* result, unsigned int sizePower,
             int sign = 1, unsigned int stride = 1);
    // 2D DFT using ordinary sums.
    // <sign> is the sign of 2*PI*i in the fourier transformation.
    static
    void dft2D(std::complex<T>* data, std::complex<T>* result,
               unsigned int rowPower, unsigned int columnPower, int sign = 1);


    // This routine is taken from "Numerical recipes in C" and does FFT in one
    // dimension. Data indices start at 1!! I didn't really understand how it
    // works and have used it only for verifying the results of my FFT routine.
    static
    void four1(T* data, unsigned int nn, int isign);

    static
    void unitTest();

private:
    // <value> = 00000110, <limit> = 3 returns 00000011.
    static
    unsigned int reflect(unsigned int value, unsigned int limit);

    // Reflect the indices of the elements stored in array <data> and
    // reorder the entries according to the reflected indices.
    static
    void reorder(std::complex<T>* data, unsigned int sizePower, unsigned int stride = 1);
};

} // namespace Ge

#include "FftRoutines.inl"
#endif // FFT_ROUTINES_H