//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------


#include "UnitRoot.h"
#include <System/Trace.h>

namespace Ge
{

template<class T> inline
//------------------------------------------------------------------------------
unsigned int FftRoutines<T>::reflect(unsigned int value, unsigned int limit)
//------------------------------------------------------------------------------
{
    unsigned int ret(0);
    for (unsigned int i = 0; i<limit; ++i)
    {
        ret <<= 1;
        ret |= 1&value;
        value >>= 1;
    }

    return ret;
}

template<class T> inline
//------------------------------------------------------------------------------
void FftRoutines<T>::reorder(std::complex<T>* data, unsigned int sizePower, unsigned int stride)
//------------------------------------------------------------------------------
{
    unsigned int n(1<<sizePower);
    unsigned int j;
    for (unsigned int i = 0; i<n; ++i)
    {
        j = reflect(i, sizePower);
        if (i < j)
            std::swap(data[i*stride], data[j*stride]);
    }
}

template<class T>
//------------------------------------------------------------------------------
void FftRoutines<T>::dft(std::complex<T>* data, std::complex<T>* result, unsigned int sizePower,
                 int sign, unsigned int stride)
//------------------------------------------------------------------------------
{
    unsigned int n = 1<<sizePower;
    UnitRoot<T> w(n);
    
    for (unsigned int i = 0; i<n; ++i)
    {
        result[i*stride] = 0;
        for (unsigned int j = 0; j<n; ++j)
        {
            result[i*stride] += w.pow((sign*i*j)&(n-1))*data[j*stride];
        }
    }
}


template<class T>
//------------------------------------------------------------------------------
void FftRoutines<T>::fft(std::complex<T>* data, unsigned int sizePower, int sign,
                 unsigned int stride)
//------------------------------------------------------------------------------
{
    reorder(data, sizePower, stride);

    unsigned int n(1<<sizePower);
    UnitRoot<T> w(n);
    
    unsigned int bd(1); // Butterfly delta
    unsigned int bs(2); // Butterfly step
    unsigned int wm(n>>1); // Unit root multiplier
    
    std::complex<T> temp;
    unsigned int ci, ni; // Current index, next index in the butterfly
    for (unsigned int level = 0; level < sizePower; ++level)
    {
        for (unsigned int i = 0; i<n; i += bs)
        {
            for(unsigned int j = 0; j<bd; ++j)
            {
                ci = stride*(i+j);
                ni = stride*(i+j+bd);
                
                // (j*wm)&(n-1) = (j*wm)%n if n is a power of 2
                temp = w.pow((sign*j*wm)&(n-1))*data[ni];
                
                data[ni] = data[ci] - temp;
                data[ci] += temp;    
            }
        }
        bd <<= 1;
        bs <<= 1;
        wm >>= 1;
    }
}

template<class T>
//------------------------------------------------------------------------------
void FftRoutines<T>::four1(T* data, unsigned int nn, int isign)
//------------------------------------------------------------------------------
{
    unsigned long n, mmax, m, j, istep, i;
    double wtemp, wr, wpr, wpi, wi, theta;
    T tempr, tempi;

    n = nn<<1;
    j = 1;
    for (i = 1; i<n; i+=2)
    {
        if (j>i)
        {
            swap(data[j], data[i]);
            swap(data[j+1], data[i+1]);
        }
        m = n>>1;
        while(m>=2 && j>m)
        {
            j -= m;
            m>>=1;
        }
        j+=m;
    }

    mmax = 2;
    while (n>mmax)
    {
        istep = mmax<<1;
        theta = isign*(6.283185307179586476925286766559/mmax);
        wtemp = sin(0.5*theta);
        wpr = -2.0*wtemp*wtemp;
        wpi = sin(theta);
        wr = 1.0;
        wi = 0.0;
        for (m = 1; m<mmax; m += 2)
        {
            for(i = m; i<=n; i += istep)
            {
                j = i+mmax;
                tempr = wr*data[j] - wi*data[j+1];
                tempi = wr*data[j+1] + wi*data[j];
                data[j] = data[i] - tempr;
                data[j+1] = data[i+1] - tempi;
                data[i] += tempr;
                data[i+1] += tempi;
            }
            wr = (wtemp = wr)*wpr - wi*wpi + wr;
            wi = wi*wpr + wtemp*wpi + wi;
        }
        mmax = istep;
    }
}


template<class T>
//------------------------------------------------------------------------------
void FftRoutines<T>::dft2D(std::complex<T>* data, std::complex<T>* result,
                   unsigned int rowPower, unsigned int columnPower,
                   int sign)
//------------------------------------------------------------------------------
{
    unsigned int rows = 1<<rowPower;
    unsigned int columns = 1<<columnPower;
    
    // Do FFT for all rows.
    std::complex<T>* s = data;
    std::complex<T>* d = result;
    unsigned int i;
    for (i = 0; i<rows; ++i)
    {
        dft(s, d, columnPower, sign);
        s += columns;
        d += columns;
    }

    // Do FFT for all columns.
    s = data;
    d = result;
    for (i = 0; i<columns; ++i)
    {
        dft(s, d, rowPower, sign, columns);
        ++s; ++d;
    }
}

template<class T>
//------------------------------------------------------------------------------
void FftRoutines<T>::fft2D(std::complex<T>* data, unsigned int rowPower, unsigned int columnPower,
                   int sign)
//------------------------------------------------------------------------------
{
    unsigned int rows = 1<<rowPower;
    unsigned int columns = 1<<columnPower;
    
    // Do FFT for all rows.
    std::complex<T>* ptr = data;
    unsigned int i;
    for (i = 0; i<rows; ++i)
    {
        fft(ptr, columnPower, sign);
        ptr += columns;
    }

    // Do FFT for all columns.
    ptr = data;
    for (i = 0; i<columns; ++i)
    {
        fft(ptr, rowPower, sign, columns);
        ++ptr;
    }
}


//template<class T>
////------------------------------------------------------------------------------
//void FftRoutines<T>::unitTest()
////------------------------------------------------------------------------------
//{    
//    std::complex<float> an[128][128];
//    std::complex<float> rn[128][128];
//    fftw_complex in[128][128];
//
//    using namespace std;
//
//    // Create arrays
//    for (unsigned int i = 0; i<128; ++i)
//    {
//        for (unsigned int j = 0; j<128; ++j)
//        {
//            an[i][j] = complex<float>(::rand(), ::rand());
//            in[i][j].re = an[i][j].real();
//            in[i][j].im = an[i][j].imag();
//        }
//    }
//
//    // Dft
//    FftRoutines<float>::dft2D((complex<float>*)an, (complex<float>*)rn, 7, 7, -1);
//    // FftRoutines
//    FftRoutines<float>::fft2D((complex<float>*)an, 7, 7, -1);
//    // Fftw
//    fftwnd_plan p;
//    p = fftw2d_create_plan(128, 128, FFTW_FORWARD, FFTW_ESTIMATE|FFTW_IN_PLACE);
//    fftwnd_one(p, (fftw_complex*)in, 0);
//    fftwnd_destroy_plan(p); 
//
//    for (i = 0; i<128; ++i)
//    {
//        for (unsigned int j = 0; j<128; ++j)
//        {
//            GE_ASSERT((fabs(an[i][j].real()-in[i][j].re)/an[i][j].real() < 0.01));
//            GE_ASSERT(((fabs(an[i][j].imag()-in[i][j].im)/an[i][j].imag() < 0.01));
//            GE_ASSERT(((fabs(an[i][j].real()-rn[i][j].real())/an[i][j].imag() < 0.01));
//            GE_ASSERT(((fabs(an[i][j].imag()-rn[i][j].imag())/an[i][j].imag() < 0.01));
//        }
//    }
//}

} // namespace Ge