#ifndef MATH_MATRIX_H
#define MATH_MATRIX_H
#pragma once

//------------------------------------------------------------------------------
// (c) Apr-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include <System/Base.h>

#include "Vector.h"

#include <ostream>

namespace Ge
{

class MatrixBase
{
public:
    // Use this enum if you want to access matrix columns in a more readable way.
    enum Column
    {
        RIGHT = 0,
        UP = 1,
        AT = 2,
        TRANSLATION = 3
    };
};

template<typename T, UInt R, UInt C>
class BasicMatrix : public MatrixBase
{
public:

    static BasicMatrix& IDENTITY();

    static BasicMatrix& ZERO();

    //------------------------------------------------------------------------------
    // Constructors
    //------------------------------------------------------------------------------

    BasicMatrix();

    BasicMatrix(const BasicMatrix&);
    
    //------------------------------------------------------------------------------
    // Accessors
    //------------------------------------------------------------------------------
    
    const T& operator()(UInt row, UInt column) const;

    T& operator()(UInt row, UInt column);

    // Returns a sub matrix of *this starting at <rowStart> and <colStart>. The
    // dimensions of the sub matrix must be small enough.
    template<typename T1, UInt R1, UInt C1>
    void getSubMatrix(BasicMatrix<T1, R1, C1>& m,
                      UInt rowStart = 0, UInt colStart = 0) const
    {
        GE_ASSERT_MSG(rowStart+R1 <= R && colStart+C1 <= C,
            "Invalid dimensions.");

        for (UInt i = rowStart; i<R1; ++i)
        {
            for (UInt j = colStart; j<C1; ++j)
            {
                m(i,j) = (*this)(rowStart+i,colStart+j);
            }
        }
    }

    // Sets sub matrix of *this starting at <rowStart> and <colStart>. The
    // dimensions of the sub matrix must be small enough.
    template<typename T1, UInt R1, UInt C1>
    void setSubMatrix(const BasicMatrix<T1, R1, C1>& m,
                      UInt rowStart = 0, UInt colStart = 0)
    {
        GE_ASSERT_MSG(rowStart+R1 <= R && colStart+C1 <= C,
            "Invalid dimensions.");

        for (UInt i = rowStart; i<R1; ++i)
        {
            for (UInt j = colStart; j<C1; ++j)
            {
                (*this)(rowStart+i,colStart+j) = m(i,j);
            }
        }
    }

    void getRow(UInt row, BasicVector<T, C>&) const;

    void setRow(UInt row, const BasicVector<T, C>&);

    void addRow(UInt row, const BasicVector<T, C>&);

    void normalizeRow(UInt row);

    void getColumn(UInt column, BasicVector<T, R>&) const;

    void setColumn(UInt column, const BasicVector<T, R>&);

    void addColumn(UInt column, const BasicVector<T, R>&);

    void normalizeColumn(UInt column);
    
    //------------------------------------------------------------------------------
    // Operators
    //------------------------------------------------------------------------------
    
    // *= for Matrices is not implemented intentionally, since matrix
    // multiplication is not commutative

    // Can't be put into the .inl file on VC6.
    BasicVector<T, R> operator *(const BasicVector<T, C>& rhs) const
    {
        BasicVector<T, R> ret;
        for (UInt i = 0; i<R; ++i)
        {
            ret[i] = 0;
            for (UInt j = 0; j<C; ++j)
            {
                ret[i] += (*this)(i,j)*rhs[j];
            }
        }

        return ret;
    }

    // Can't be put into the .inl file on VC6. This will only compile iff
    // C=R1 and sizeof(T) == sizeof(T1). Don't use this function but use the operator *.
    // It is only needed due to problems with templates under VC6.
    template<typename T1, UInt R1, UInt C1>
    BasicMatrix<T, R, C1> multiplyHelper(const BasicMatrix<T1, R1, C1>& rhs) const
    {
        GE_CTASSERT( (sizeof(T) == sizeof(T1)) && (C == R1) );
        
        BasicMatrix<T, R, C1> ret;

        if (C == R1)
        {
            for (UInt r = 0; r<R; ++r)
            {
                for (UInt c = 0; c<C1; ++c)
                {
                    ret(r,c) = 0;
                    for (UInt rc = 0; rc<R1; ++rc)
                    {
                        ret(r,c) += (*this)(r,rc)*rhs(rc,c);
                    }
                }
            }
        }

        return ret;
    }

    BasicMatrix& operator =(const BasicMatrix&);

    bool operator ==(const BasicMatrix&) const;

    bool operator !=(const BasicMatrix&) const;
    
    //------------------------------------------------------------------------------
    // Tools
    //------------------------------------------------------------------------------
    
    bool isAlmost(const BasicMatrix&, Float tolerance = EPSILON) const;

private:
    T mData[R][C];
};


//------------------------------------------------------------------------------
// Users should use these types
//------------------------------------------------------------------------------

typedef BasicMatrix<Float, 3, 3> Matrix3x3;

typedef BasicMatrix<Float, 3, 4> Matrix3x4;

typedef BasicMatrix<Float, 4, 4> Matrix4x4;




//------------------------------------------------------------------------------
// Global operators
//------------------------------------------------------------------------------

// The w coordinate of the vector is assumed to be 1. After multiplying with
// the matrix the vector components are divided by the resulting w.
template<typename T>
BasicVector<T, 3> operator *(const BasicMatrix<T, 4, 4>&, const BasicVector<T, 3>&);

// The w coordinate of the vector is assumed to be 1.
template<typename T>
BasicVector<T, 3> operator *(const BasicMatrix<T, 3, 4>&, const BasicVector<T, 3>&);

template<typename T>
BasicMatrix<T, 3, 3> operator *(const BasicMatrix<T, 3, 3>&, const BasicMatrix<T, 3, 3>&);

template<typename T>
BasicMatrix<T, 4, 4> operator *(const BasicMatrix<T, 4, 4>&, const BasicMatrix<T, 4, 4>&);

template<typename T>
BasicMatrix<T, 3, 4> operator *(const BasicMatrix<T, 3, 4>&, const BasicMatrix<T, 3, 4>&);

template<typename T>
void setColumn(BasicMatrix<T, 4, 4>& m, UInt column,
               const T& x, const T& y, const T& z, const T& w);

template<typename T>
void setRow(BasicMatrix<T, 4, 4>& m, UInt column,
            const T& x, const T& y, const T& z, const T& w);

template<typename T>
void setColumn(BasicMatrix<T, 3, 4>& m, UInt column,
               const T& x, const T& y, const T& z);

template<typename T>
void setColumn(BasicMatrix<T, 3, 3>& m, UInt column,
               const T& x, const T& y, const T& z);

template<typename T>
void set(BasicMatrix<T, 3, 3>& m,
         const T& _00, const T& _01, const T& _02,
         const T& _10, const T& _11, const T& _12,
         const T& _20, const T& _21, const T& _22);

template<typename T>
void set(BasicMatrix<T, 4, 4>& m,
         const T& _00, const T& _01, const T& _02, const T& _03,
         const T& _10, const T& _11, const T& _12, const T& _13,
         const T& _20, const T& _21, const T& _22, const T& _23,
         const T& _30, const T& _31, const T& _32, const T& _33);

template<typename T, UInt R, UInt C>
std::ostream& operator <<(std::ostream&, const BasicMatrix<T, R, C>&);

//------------------------------------------------------------------------------
// 3D engine helpers
//------------------------------------------------------------------------------

// Create a projection matrix.
template<typename T>
void makeProjection(BasicMatrix<T, 4, 4>&,
                    const T& nearPlane, const T& farPlane,
                    const T& fovHorizontal, const T& aspectRatio);

// Create a projection matrix with far plane at infinity
template<typename T>
void makeProjection(BasicMatrix<T, 4, 4>&,
                    const T& nearPlane,
                    const T& fovHorizontal, const T& aspectRatio);


// This will only compile for Matrix3x4 and Matrix4x4.
// If the parameters don't allow the construction of the matrix, then false
// will be returned.
// <from>: look position
// <at>: point you are looking at
// <up>: up direction (doesn't need to be orthogonal to <from>-<at>).
// The viewer is assumed to look down the positive z-axis in object space.
template<typename T, UInt R, UInt C>
bool makeLookAt(BasicMatrix<T, R, C>&, const BasicVector<T, 3>& from,
                const BasicVector<T, 3>& at, const BasicVector<T, 3>& up);

// This will only compile for Matrix3x4 and Matrix4x4.
// If the parameters don't allow the construction of the matrix, then false
// will be returned.
// <from>: position of camera
// <at>: point the camera is looking at
// <up>: up direction of camera (doesn't need to be orthogonal to <from>-<at>).
// The camera is assumed to look down the positive z-axis in object space.
template<typename T, UInt R, UInt C>
bool makeView(BasicMatrix<T, R, C>&, const BasicVector<T, 3>& from,
              const BasicVector<T, 3>& at, const BasicVector<T, 3>& up);

// Construct a billboard matrix with a given view matrix.
template<typename T, UInt R, UInt C>
void makeBillboard(BasicMatrix<T, R, C>& out, const BasicMatrix<T, 3, 4>& view);

template<typename T, UInt R, UInt C>
void makeBillboard(BasicMatrix<T, R, C>& out, const BasicMatrix<T, 4, 4>& view);




//------------------------------------------------------------------------------
// Rotations
//------------------------------------------------------------------------------

// Rotation is clockwise when looking down the positive rotation axis
// towards the origin.
template<typename T, UInt R, UInt C>
void makeXRotation(BasicMatrix<T, R, C>&, const T& angle);

template<typename T, UInt R, UInt C>
void makeYRotation(BasicMatrix<T, R, C>&, const T& angle);

template<typename T, UInt R, UInt C>
void makeZRotation(BasicMatrix<T, R, C>&, const T& angle);


// Create rotation from Euler angles
template<typename T, UInt R, UInt C>
void makeRotation(BasicMatrix<T, R, C>&,
                  const T& head, const T& pitch, const T& roll);

// Create rotation from axis/angle.
template<typename T, UInt R, UInt C>
void makeRotation(BasicMatrix<T, R, C>&,
                  const BasicVector<T, 3>& axis, const T& angle);




//------------------------------------------------------------------------------
// Inversion of affine matrices
//------------------------------------------------------------------------------

// Returns true if the upper left sub Matrix3x3 is orthogonal. (Affine matrices
// represent a rotation + translation). I don't check whether the last row
// of a Matrix4x4 is (0,0,0,1). It's up to the user to verify this.
template<typename T, UInt R, UInt C>
bool isAffine(const BasicMatrix<T, R, C>&);

template<typename T, UInt R, UInt C>
void invertAffine(BasicMatrix<T, R, C>&);

} // namespace Ge

#include "Matrix.inl"
#endif // !MATH_MATRIX_H