#ifndef QUATERNION_H
#define QUATERNION_H
#pragma once

//------------------------------------------------------------------------------
// (c) Mai-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include <System/Base.h>

#include "Vector.h"

namespace Ge
{

// Don't use BasicQuaternion directly. Use the Quaternion typedef below.
template<typename T>
class BasicQuaternion
{
public:
    //------------------------------------------------------------------------------
    // Construction
    //------------------------------------------------------------------------------
    
    BasicQuaternion();

    BasicQuaternion(const T& real, const T& x, const T& y, const T& z);

    BasicQuaternion(const T&, const BasicVector<T, 3>&);

    //------------------------------------------------------------------------------
    // Accessors
    //------------------------------------------------------------------------------

    T& real();

    const T& real() const;

    BasicVector<T, 3>& imaginery();

    const BasicVector<T, 3>& imaginery() const;

    void set(const T& real, const T& x, const T& y, const T& z);
    
    //------------------------------------------------------------------------------
    // Operators
    //------------------------------------------------------------------------------
    
    BasicQuaternion& operator =(const BasicQuaternion&);

    BasicQuaternion& operator +=(const BasicQuaternion&);

    BasicQuaternion& operator -=(const BasicQuaternion&);

    BasicQuaternion& operator *=(const T&);

    BasicQuaternion operator *(const BasicQuaternion&) const;

    bool operator ==(const BasicQuaternion&) const;

    bool operator !=(const BasicQuaternion&) const;

    BasicQuaternion operator -() const;

    //------------------------------------------------------------------------------
    // Tools
    //------------------------------------------------------------------------------
    
    // Conjugate this quaternion.
    void conjugate();

    // Get the conjugate of this quaternion.
    BasicQuaternion getConjugate() const;

    // Rotates <in> using this quaternion and stores the result in <out>.
    void transform(BasicVector<T, 3>& out, const BasicVector<T, 3>& in) const;

    T getNorm() const;

    // Convert the quaternion to a corresponding rotation matrix.
    template<class MATRIX>
    void convertTo(MATRIX& m) const
    {
        GE_ASSERT(::Ge::isAlmost(getNorm(), 1.0f, 0.001f));
        T wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2;

        // Calculate coefficients
        x2 = imaginery()[0] + imaginery()[0];
        y2 = imaginery()[1] + imaginery()[1]; 
        z2 = imaginery()[2] + imaginery()[2];
        xx = imaginery()[0] * x2;
        xy = imaginery()[0] * y2;
        xz = imaginery()[0] * z2;
        yy = imaginery()[1] * y2;
        yz = imaginery()[1] * z2;
        zz = imaginery()[2] * z2;
        wx = real() * x2;
        wy = real() * y2;
        wz = real() * z2;

        m = MATRIX::IDENTITY();
        m(0, 0) = 1 - (yy + zz);
        m(0, 1) = xy - wz;
        m(0, 2) = xz + wy;

        m(1, 0) = xy + wz;
        m(1, 1) = 1 - (xx + zz);
        m(1, 2) = yz - wx;

        m(2, 0) = xz - wy;
        m(2, 1) = yz + wx;
        m(2, 2) = 1 - (xx + yy);
    }

    // Convert <m> to a quaternion.
    template<class MATRIX>
    void convertFrom(const MATRIX& m)
    {
        static UInt nxt[3] = {1, 2, 0};
    
        Float trace, s, q[4];
        trace = m(0, 0) + m(1, 1) + m(2, 2);

        // check the diagonal
        if (trace > 0)
        {
            s = sqrt(trace + T(1.0));
            real() = -s / T(2.0);
            s = T(0.5) / s;
            imaginery()[0] = (m(1, 2) - m(2, 1)) * s;
            imaginery()[1] = (m(2, 0) - m(0, 2)) * s;
            imaginery()[2] = (m(0, 1) - m(1, 0)) * s;
        }
        else
        {		
            UInt i = 0;
            // diagonal is negative
            if (m(1, 1) > m(0, 0))
            {
                i = 1;
            }

            if (m(2, 2) > m(i, i))
            {
                i = 2;
            }

            UInt j = nxt[i];
            UInt k = nxt[j];

            s = sqrt ((m(i, i) - (m(j, j) + m(k, k))) + 1);
  
            q[i] = s * 0.5;

            if (s != 0.0)
            {
                s = 0.5 / s;
            }

            q[3] = (m(j, k) - m(k, j)) * s;
            q[j] = (m(i, j) + m(j, i)) * s;
            q[k] = (m(i, k) + m(k, i)) * s;

            imaginery()[0] = q[0];
            imaginery()[1] = q[1];
            imaginery()[2] = q[2];
            real() = -q[3];
        }

        GE_ASSERT(::Ge::isAlmost(getNorm(), 1.0f));
    }

    // Make a quaternion using the input Euler angles. (Parameters are passed
    // by value, because a copy is needed anyway).
    void makeRotation(T head, T pitch, T roll);

    // Make a quaternion from a rotation axis and rotation angle.
    void makeRotation(const BasicVector<T, 3>& axis, const T& angle);

    // Make a quaternion that rotates a unit vector <from> to a unit vector <to>.
    void makeRotation(const BasicVector<T, 3>& from, const BasicVector<T, 3>& to);

    bool isAlmost(const BasicQuaternion<T>&, Float tolerance = EPSILON) const;

    // Spherical linear interpolation. <t> must be in [0,1].
    void slerp(const BasicQuaternion<T>& q0, const BasicQuaternion<T>& q1, Float t);

private:
    T mReal;
    BasicVector<T, 3> mImaginery;
};

typedef BasicQuaternion<Float> Quaternion;

template<class T>
BasicQuaternion<T> operator *(const BasicQuaternion<T>&, const T&);

template<class T>
BasicQuaternion<T> operator *(const T&, const BasicQuaternion<T>&);

} // namespace Ge

#include "Quaternion.inl"
#endif // !QUATERNION_H