//------------------------------------------------------------------------------
// (c) Mai-2002 Gottfried Chen
//------------------------------------------------------------------------------

namespace Ge
{

//------------------------------------------------------------------------------
// Construction
//------------------------------------------------------------------------------

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>::BasicQuaternion()
//------------------------------------------------------------------------------
{
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>::BasicQuaternion(const T& real, const BasicVector<T, 3>& im) :
//------------------------------------------------------------------------------
mReal(real),
mImaginery(im)
{
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>::BasicQuaternion(const T& real, const T& x, const T& y, const T& z) :
//------------------------------------------------------------------------------
mReal(real),
mImaginery(x,y,z)
{
}

//------------------------------------------------------------------------------
// Accessors
//------------------------------------------------------------------------------

template<typename T> inline
//------------------------------------------------------------------------------
T& BasicQuaternion<T>::real()
//------------------------------------------------------------------------------
{
    return mReal;
}

template<typename T> inline
//------------------------------------------------------------------------------
const T& BasicQuaternion<T>::real() const
//------------------------------------------------------------------------------
{
    return mReal;
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicVector<T, 3>& BasicQuaternion<T>::imaginery()
//------------------------------------------------------------------------------
{
    return mImaginery;
}


template<typename T> inline
//------------------------------------------------------------------------------
const BasicVector<T, 3>& BasicQuaternion<T>::imaginery() const
//------------------------------------------------------------------------------
{
    return mImaginery;
}


template<typename T> inline
//------------------------------------------------------------------------------
void BasicQuaternion<T>::set(const T& w, const T& x, const T& y, const T& z)
//------------------------------------------------------------------------------
{
    mReal = w;
    set(mImaginery, x,y,z);
}

//------------------------------------------------------------------------------
// Operators
//------------------------------------------------------------------------------

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>& BasicQuaternion<T>::operator =(const BasicQuaternion<T>& rhs)
//------------------------------------------------------------------------------
{
    if (this != &rhs)
    {
        real() = rhs.real();
        imaginery() = rhs.imaginery();
    }
    return *this;
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>& BasicQuaternion<T>::operator +=(const BasicQuaternion<T>& rhs)
//------------------------------------------------------------------------------
{
    real() += rhs.real();
    imaginery() += rhs.imaginery();
    return *this;
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>& BasicQuaternion<T>::operator -=(const BasicQuaternion<T>& rhs)
//------------------------------------------------------------------------------
{
    real() -= rhs.real();
    imaginery() -= rhs.imaginery();
    return *this;
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T>& BasicQuaternion<T>::operator *=(const T& rhs)
//------------------------------------------------------------------------------
{
    real() *= rhs;
    imaginery() *= rhs;
    return *this;
}

template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T> BasicQuaternion<T>::operator *(const BasicQuaternion<T>& rhs) const
//------------------------------------------------------------------------------
{
    return BasicQuaternion<T>(// real
                              real()*rhs.real()-imaginery()*rhs.imaginery(),
                              // imaginery
                              cross(imaginery(), rhs.imaginery()) +
                              rhs.real()*imaginery() + real()*rhs.imaginery());
}


template<typename T> inline
//------------------------------------------------------------------------------
bool BasicQuaternion<T>::operator ==(const BasicQuaternion<T>& rhs) const
//------------------------------------------------------------------------------
{
    return real() == rhs.real() && imaginery() == rhs.imaginery();
}


template<typename T> inline
//------------------------------------------------------------------------------
bool BasicQuaternion<T>::operator !=(const BasicQuaternion<T>& rhs) const
//------------------------------------------------------------------------------
{
    return !(*this == rhs);
}


template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T> BasicQuaternion<T>::operator -() const
//------------------------------------------------------------------------------
{
    return BasicQuaternion<T>(-real(), -imaginery());
}



//------------------------------------------------------------------------------
// Tools
//------------------------------------------------------------------------------


template<typename T> inline
//------------------------------------------------------------------------------
void BasicQuaternion<T>::conjugate()
//------------------------------------------------------------------------------
{
    imaginery() = -imaginery();
}


template<typename T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T> BasicQuaternion<T>::getConjugate() const
//------------------------------------------------------------------------------
{
    return BasicQuaternion<T>(real(), -imaginery());
}


template<typename T> inline
//------------------------------------------------------------------------------
void BasicQuaternion<T>::transform(BasicVector<T, 3>& out,
                                   const BasicVector<T, 3>& v) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(::Ge::isAlmost(getNorm(), T(1)));
    BasicQuaternion<T> qv(0, v);
    out = ((*this)*qv*(this->getConjugate())).imaginery();
}


template<typename T> inline
//------------------------------------------------------------------------------
T BasicQuaternion<T>::getNorm() const
//------------------------------------------------------------------------------
{
    return square(real()) + imaginery().lengthSquare();
}


template<typename T>
//------------------------------------------------------------------------------
void BasicQuaternion<T>::makeRotation(T head, T pitch, T roll)
//------------------------------------------------------------------------------
{
    // This can later be optimized by premultiplying the quaternions to
    // get a formula where h,p,r only need to be filled in.
    head *= 0.5f;
    pitch *= 0.5f;
    roll *= 0.5f;
    BasicQuaternion<T> qh(cos(head), 0, sin(head), 0);
    BasicQuaternion<T> qp(cos(pitch), sin(pitch), 0, 0);
    BasicQuaternion<T> qr(cos(roll), 0, 0, sin(roll));
    *this = qr*qp*qh;
}


template<typename T> inline
//------------------------------------------------------------------------------
void BasicQuaternion<T>::makeRotation(const BasicVector<T, 3>& axis, const T& angle)
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(::Ge::isAlmost(axis.lengthSquare(), 1.0f, EPSILON),
                  "Only normalized axes are allowed.");

    real() = cos(angle/T(2));
    imaginery() = sin(angle/T(2))*axis;
}

template<typename T> inline
//------------------------------------------------------------------------------
void BasicQuaternion<T>::makeRotation(const BasicVector<T, 3>& from,
                                      const BasicVector<T, 3>& to)
//------------------------------------------------------------------------------
{
    GE_ASSERT(::Ge::isAlmost(from.lengthSquare(), 1.0f, 0.001f));
    GE_ASSERT(::Ge::isAlmost(to.lengthSquare(), 1.0f, 0.001f));
    if (from.isAlmost(to, 0.001f))
    {
        real() = 1;
        imaginery() = Vector3(0,0,0);
    }
    else if (from.isAlmost(-to, 0.001f))
    {
        makeRotation(from.getRandomNormal(), PI);
    }
    else
    {
        makeRotation(cross(from, to), from*to);
    }
}

template<typename T> inline
//------------------------------------------------------------------------------
bool BasicQuaternion<T>::isAlmost(const BasicQuaternion<T>& rhs, Float tolerance) const
//------------------------------------------------------------------------------
{
    return ::Ge::isAlmost(real(), rhs.real(), tolerance) &&
           imaginery().isAlmost(rhs.imaginery(), tolerance);
}

template<typename T> inline
//------------------------------------------------------------------------------
void BasicQuaternion<T>::slerp(const BasicQuaternion<T>& q0,
                               const BasicQuaternion<T>& q1, Float t)
//------------------------------------------------------------------------------
{
    GE_ASSERT(0<=t && t<=1);
    if (t == 0 || q0.isAlmost(q1, 0.001f))
    {
        *this = q0;
    }
    else if (t == 1)
    {
        *this = q1;
    }
    else
    {
        Float phi = acos(q0.imaginery()*q1.imaginery() + q0.real()*q1.real());
        GE_ASSERT(phi != 0);
        Float sphi = 1/sin(phi);
        *this = (sin(phi*(1-t))*sphi)*q0;
        *this += (sin(phi*t)*sphi)*q1;
    }
}


//------------------------------------------------------------------------------
// Global functions
//------------------------------------------------------------------------------

template<class T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T> operator *(const BasicQuaternion<T>& lhs, const T& rhs)
//------------------------------------------------------------------------------
{
    BasicQuaternion<T> q(lhs);
    q *= rhs;
    return q;
}

template<class T> inline
//------------------------------------------------------------------------------
BasicQuaternion<T> operator *(const T& lhs, const BasicQuaternion<T>& rhs)
//------------------------------------------------------------------------------
{
    return rhs*lhs;
}

} // namespace Ge
