//------------------------------------------------------------------------------
// (c) Apr-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "MatrixTest.h"
#include <Math/Matrix.h>

namespace
{

using namespace Ge;

template<typename T, UInt R, UInt C>
//------------------------------------------------------------------------------
void fill(BasicMatrix<T, R, C>& m)
//------------------------------------------------------------------------------
{
    for (UInt r = 0; r<R; ++r)
    {
        for (UInt c = 0; c<C; ++c)
        {
            m(r,c) = r*C+c;
        }
    }
}

//------------------------------------------------------------------------------
void fill(Matrix3x3& m,
          Float _00, Float _01, Float _02,
          Float _10, Float _11, Float _12,
          Float _20, Float _21, Float _22)
//------------------------------------------------------------------------------
{
    m(0,0) = _00;
    m(0,1) = _01;
    m(0,2) = _02;
    m(1,0) = _10;
    m(1,1) = _11;
    m(1,2) = _12;
    m(2,0) = _20;
    m(2,1) = _21;
    m(2,2) = _22;
}

//------------------------------------------------------------------------------
void makeId(Matrix3x3& m)
//------------------------------------------------------------------------------
{
    m(0,0) = 1; m(0,1) = 0; m(0,2) = 0;
    m(1,0) = 0; m(1,1) = 1; m(1,2) = 0;
    m(2,0) = 0; m(2,1) = 0; m(2,2) = 1;
}

} // namespace

namespace Ge
{

//------------------------------------------------------------------------------
void MatrixTest::run() const
//------------------------------------------------------------------------------
{
    Matrix3x3 m;
    fill(m);
    GE_ASSERT(m(0,0) == 0); GE_ASSERT(m(0,1) == 1); GE_ASSERT(m(0,2) == 2);
    GE_ASSERT(m(1,0) == 3); GE_ASSERT(m(1,1) == 4); GE_ASSERT(m(1,2) == 5);
    GE_ASSERT(m(2,0) == 6); GE_ASSERT(m(2,1) == 7); GE_ASSERT(m(2,2) == 8);

    Matrix3x3 m2(m);
    GE_ASSERT(m == m2);

    m2 = m;
    GE_ASSERT(m == m2);
    m2(0,0) = 20;
    GE_ASSERT(m != m2);

    makeId(m);
    GE_ASSERT(m == Matrix3x3::IDENTITY());

    // Test vector multiplication.
    {
        Matrix3x4 m34;
        fill(m34);
        Vector4 v(9,8,5,1);
        GE_ASSERT(Vector3(21,113,205) == m34*v);
    }
    
    // Test matrix multiplication.
    {
        Matrix3x3 m;
        fill(m);

        Matrix3x3 m2;
        fill(m2, 1,4,6,
                 2,12,7,
                 19,98,1234);
        
        Matrix3x3 result;
        fill(result, 40,208,2475,
                     106,550,6216,
                     172,892,9957);
        
        GE_ASSERT(result == m*m2);    
    }

    // Test 3x4 multiplication
    {
        Matrix3x4 m1;
        Matrix3x4 m2;
        fill(m1);
        fill(m2);
        Matrix4x4 m1_4;
        Matrix4x4 m2_4;
        m1_4 = Matrix4x4::IDENTITY();
        m2_4 = Matrix4x4::IDENTITY();
        m1_4.setSubMatrix(m1);
        m2_4.setSubMatrix(m2);
        Matrix3x4 result1 = m1*m2;
        Matrix4x4 result4x4 = m1_4*m2_4;
        Matrix3x4 result2;
        result4x4.getSubMatrix(result2);
        GE_ASSERT(result1 == result2);
    }

    // Test makeLookAt
    {
        Matrix3x4 lookAt;
        makeLookAt(lookAt, Vector3(0,0,-10), Vector3::ORIGIN(), Vector3(0,1,0));
    }

    // Test invert
    {
        Matrix4x4 view;
        makeView(view, Vector3(1,2,3), Vector3(0,0,0), Vector3(0,1,0));
        Matrix4x4 inverse(view);
        invertAffine(inverse);
        inverse = view*inverse;
        GE_ASSERT(inverse.isAlmost(Matrix4x4::IDENTITY()));
    }
}

} // namespace Ge
