//------------------------------------------------------------------------------
// (c) Mai-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "PrimitivesTest.h"

#include <Math/Sphere.h>
#include <Math/Plane.h>
#include <Math/HalfSpace.h>
#include <Math/Frustum.h>


namespace
{

using namespace Ge;

//------------------------------------------------------------------------------
void testSphere()
//------------------------------------------------------------------------------
{
    Sphere a(Vector3(1,0,0), 0.5f);
    Sphere b(Vector3(-1,0,0), 0.5f);
    
    // Test intersection & containment
    GE_ASSERT(!a.intersects(b));
    GE_ASSERT(!a.contains(Vector3(0,1,0)));
    
    b.setRadius(2.0f);
    GE_ASSERT(a.intersects(b));
    GE_ASSERT(a.contains(Vector3(0.75f,0,0)));

    Sphere c(Vector3(0.5f,0,0), 2);
    Sphere d(Vector3(0,0,0), 1);
    GE_ASSERT(c.contains(d));
    
    // Test sphere merging
    Sphere e(d);
    e.merge(c);
    GE_ASSERT(e.contains(c));
    GE_ASSERT(e.contains(d));

    e = c;
    e.merge(d);
    GE_ASSERT(e.contains(c));
    GE_ASSERT(e.contains(d));

    d.setCenter(Vector3(3,6,8));
    e = c;
    e.merge(d);
    GE_ASSERT(e.contains(c));
    GE_ASSERT(e.contains(d));

    // Test vertex merging
    Vector3 vertices[50];
    UInt i;
    for (i = 0; i<50; i++)
    {
        vertices[i][0] = rand();
        vertices[i][1] = rand();
        vertices[i][2] = rand();
    }

    Sphere hull(Vector3(), 0);
    hull.merge(vertices[0], 50);
    for (i = 0; i<50; i++)
    {
        GE_ASSERT(hull.contains(vertices[i]));
    }
}

//------------------------------------------------------------------------------
void testPlane()
//------------------------------------------------------------------------------
{
    // Test distance()
    {
        Plane p(Vector3(3,1,0), 2);
        GE_ASSERT(isAlmost(p.getDistance(Vector3(0,0,0)), 2.0f));
    }

    // Test sphere intersection
    {
        Plane p(Vector3(3,1,0), 2);
        Sphere s(Vector3(0,0,0), 1.9f);
        GE_ASSERT(!p.intersects(s));
        s.setRadius(2.1f);
        GE_ASSERT(p.intersects(s));
    }

    // Test c'tor with plane point
    {
        Plane p2(Vector3(3,1,0), Vector3(0,3,0));
        Sphere s(Vector3(0,0,0), 1.9f);
        s.setRadius(0.9f);
        GE_ASSERT(!p2.intersects(s));
        s.setRadius(1.2f);
        GE_ASSERT(p2.intersects(s));
    }

    // Test transform
    {
        Plane p3(Vector3(0,0,1), 2);
        Matrix4x4 mat;
        makeYRotation(mat, -PI/4);
        setColumn(mat, 3, -1.0f,0.0f,1.0f,1.0f);
        Plane p3Tr(p3.transform(mat));
        GE_ASSERT(isAlmost(p3Tr.getOffset(), 2.0f+::Ge::sqrt(2)));
        GE_ASSERT((p3Tr.getNormal() -
            Vector3(-::Ge::sqrt(2)/2.0f,0,::Ge::sqrt(2)/2.0f)).isAlmost(Vector3::ORIGIN(), EPSILON));
    }
}

//------------------------------------------------------------------------------
void testHalfSpace()
//------------------------------------------------------------------------------
{
    // Sphere intersection
    HalfSpace hs(Plane(Vector3(2,1,0), 3));
    Sphere s(Vector3(0,0,0), 2.5f);
    GE_ASSERT(hs.intersects(s) == Intersect::DISJOINT);
    s.setRadius(3.001f);
    GE_ASSERT(hs.intersects(s) == Intersect::INTERSECTS);
    s.setCenter(Vector3(4,2,0));
    GE_ASSERT(hs.intersects(s) == Intersect::INTERSECTS);
    s.setCenter(Vector3(8,4,0));
    GE_ASSERT(hs.intersects(s) == Intersect::CONTAINS);

    // Vector containment
    Vector3 v(0,0,0);
    GE_ASSERT(!hs.contains(v));
    ::Ge::set(v, 5.0f, 1.0f, 0.0f);
    GE_ASSERT(hs.contains(v));
}

//------------------------------------------------------------------------------
void testFrustum()
//------------------------------------------------------------------------------
{
    // Test normal frustum/sphere intersection
    Frustum f1(1, 5, PI_HALF, PI_HALF);
    // Sphere in front of near plane
    Sphere s1(Vector3(0,0,0), 0.99f);
    GE_ASSERT(f1.intersects(s1) == Intersect::DISJOINT);
    // Sphere big enough to intersect near plane
    s1.setRadius(1.01f);
    GE_ASSERT(f1.intersects(s1) == Intersect::INTERSECTS);
    // Sphere behind far plane
    s1.setCenter(Vector3(0,0,6));
    GE_ASSERT(f1.intersects(s1) == Intersect::INTERSECTS);
    // Sphere small enough to be disjoint from far plane
    s1.setRadius(0.99f);
    GE_ASSERT(f1.intersects(s1) == Intersect::DISJOINT);
    // Move sphere into front part of frustum
    s1.setCenter(Vector3(0,0,2));
    GE_ASSERT(f1.intersects(s1) == Intersect::CONTAINS);
    // Make sphere bigger, to not be contained
    s1.setRadius(1.01f);
    GE_ASSERT(f1.intersects(s1) == Intersect::INTERSECTS);
    // Make sphere containing frustum
    s1.setCenter(Vector3(0,0,0));
    s1.setRadius(20);
    GE_ASSERT(f1.intersects(s1) == Intersect::INTERSECTS);

    // Test transformed frustum (from (1,1,1), look at origin, Y-axis is up)
    Matrix4x4 view;
    makeView(view, Vector3(1,1,1), Vector3(0,0,0), Vector3(0,1,0));
    Matrix4x4 lookAt;
//    lookAt.makeBillboard(view);
    makeBillboard(lookAt, view);
    setColumn(lookAt, 3, 1.0f,1.0f,1.0f,1.0f);
    Plane test(Vector3(0,0,1), 1);
    test = test.transform(lookAt);
    Frustum f1Tr(f1.transform(lookAt));
    // Sphere in front of near plane
    Sphere s2(Vector3(1,1,1), 0.99f);
    GE_ASSERT(f1Tr.intersects(s2) == Intersect::DISJOINT);
    // Sphere big enough to intersect near plane
    s2.setRadius(1.01f);
    GE_ASSERT(f1Tr.intersects(s2) == Intersect::INTERSECTS);
    // Sphere behind far plane ((-1,-1,-1).normalize() is direction vector
    // of the frustum)
    Vector3 dir(Vector3(-1,-1,-1));
    dir.normalize();
    s2.setCenter(Vector3(1,1,1) + 6.0f*dir);
    GE_ASSERT(f1Tr.intersects(s2) == Intersect::INTERSECTS);
    // Sphere small enough to be disjoint from far plane
    s2.setRadius(0.99f);
    GE_ASSERT(f1Tr.intersects(s2) == Intersect::DISJOINT);
    // Move sphere into front part of frustum
    s2.setCenter(Vector3(1,1,1) + 2.0f*dir);
    GE_ASSERT(f1Tr.intersects(s2) == Intersect::CONTAINS);
    // Make sphere bigger, to not be contained
    s2.setRadius(1.01f);
    GE_ASSERT(f1Tr.intersects(s2) == Intersect::INTERSECTS);
}

} // namespace

namespace Ge
{

//------------------------------------------------------------------------------
void PrimitivesTest::run() const
//------------------------------------------------------------------------------
{
    testSphere();
    testPlane();
    testHalfSpace();
    testFrustum();
}

} // namespace Ge
