#ifndef MATH_VECTOR_H
#define MATH_VECTOR_H
#pragma once

//------------------------------------------------------------------------------
// (c) 03-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include <System/Base.h>
#include <ostream>

namespace Ge
{

// The 2, 3 and 4 dimensional vectors are no longer derived from the general
// vector class because writing e.g. (a+b).cross(c); wasn't possible because
// (a+b) was of the mother vector class type. So you'd have had to write
// temp = a+b; temp.cross(c);
// So basically the problem was that the derived classes used the operators
// of the mother class, but those returned objects/references of the mother
// class that couldn't use the derived classes'  interface.
// The current solution puts dimension specific functionality into external
// functions.
template <typename T, UInt DIM>
class BasicVector
{
public:
    static const BasicVector& ORIGIN();

    //------------------------------------------------------------------------------
    // Constructors
    //------------------------------------------------------------------------------
    BasicVector();

    BasicVector(const T& x, const T& y);

    BasicVector(const T& x, const T& y, const T& z);

    BasicVector(const T& x, const T& y, const T& z, const T& w);

    BasicVector(const BasicVector&);

    //------------------------------------------------------------------------------
    // Accessors
    //------------------------------------------------------------------------------
    const T& operator [](UInt index) const;

    T& operator [](UInt index);

    //------------------------------------------------------------------------------
    // Operators
    //------------------------------------------------------------------------------
    BasicVector& operator =(const BasicVector&);

    BasicVector& operator +=(const BasicVector&);

    BasicVector& operator -=(const BasicVector&);

    BasicVector& operator *=(const T&);

    BasicVector operator -() const;

    bool operator ==(const BasicVector&) const;

    bool operator !=(const BasicVector&) const;


    //------------------------------------------------------------------------------
    // Norms
    //------------------------------------------------------------------------------
    T lengthSquare() const;

    T length() const;

    T maxNorm() const;

    T sumNorm() const;
    

    //------------------------------------------------------------------------------
    // Tools
    //------------------------------------------------------------------------------
    void normalize();

    BasicVector getRandomNormal() const;
    
    T getAngleTo(const BasicVector&) const;

    BasicVector projectOn(const BasicVector&) const;

    bool isAlmost(const BasicVector&, T tolerance = EPSILON) const;
    
    // Use this class for the Pred template parameter if you use a map of BasicVectors
    struct Pred
    {
        // MSVC: can't be put into .inl file
        inline
        bool operator ()(const BasicVector& lhs, const BasicVector& rhs) const
        {
            for (UInt i = 0; i<DIM && lhs[i] == rhs[i]; i++) {}
            return (i<DIM && lhs[i] < rhs[i]) ? true : false;
        }
    };

private:
    // Returns index of first non zero coordinate or -1
    Int getFirstNonZero() const;

    T mData[DIM];
};


//------------------------------------------------------------------------------
// Users should use these types
//------------------------------------------------------------------------------

typedef BasicVector<Float, 2> Vector2;

typedef BasicVector<Float, 3> Vector3;

typedef BasicVector<Float, 4> Vector4;

//------------------------------------------------------------------------------
// Globals
//------------------------------------------------------------------------------

template<typename T, UInt DIM>
T operator *(const BasicVector<T, DIM>&, const BasicVector<T, DIM>&);

template<typename T, UInt DIM>
BasicVector<T, DIM> operator *(const T&, const BasicVector<T, DIM>&);

template<typename T, UInt DIM>
BasicVector<T, DIM> operator *(const BasicVector<T, DIM>&, const T&);

template<typename T, UInt DIM>
BasicVector<T, DIM> operator +(const BasicVector<T, DIM>&, const BasicVector<T, DIM>&);

template<typename T, UInt DIM>
BasicVector<T, DIM> operator -(const BasicVector<T, DIM>&, const BasicVector<T, DIM>&);

template<typename T, UInt DIM>
std::ostream& operator <<(std::ostream&, const BasicVector<T, DIM>&);

//------------------------------------------------------------------------------
// Special dimensions
//------------------------------------------------------------------------------

template<typename T>
BasicVector<T, 3> cross(const BasicVector<T, 3>& lhs, const BasicVector<T, 3>& rhs);

template<typename T>
void set(BasicVector<T, 2>&, const T& x, const T& y);

template<typename T>
void set(BasicVector<T, 3>&, const T& x, const T& y, const T& z);

template<typename T>
void set(BasicVector<T, 4>&, const T& x, const T& y, const T& z, const T& w);

template<typename T>
void set(BasicVector<T, 4>&, const BasicVector<T, 3>&, const T& w);

} // namespace Ge

#include <Math/Vector.inl>
#endif // MATH_VECTOR_H
