//------------------------------------------------------------------------------
// (c) 03-2001 Gottfried Chen
//------------------------------------------------------------------------------

#include <System/Assert.h>
#include <Math/Tools.h>
#include <stdlib.h>


//------------------------------------------------------------------------------
// Members
//------------------------------------------------------------------------------

namespace Ge
{

template<typename T, UInt DIM>
//------------------------------------------------------------------------------
const BasicVector<T, DIM>& BasicVector<T, DIM>::ORIGIN()
//------------------------------------------------------------------------------
{
    // Static variables are guaranteed to be 0.
    static BasicVector<T, DIM> origin;
    return origin;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>::BasicVector()
//------------------------------------------------------------------------------
{
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>::BasicVector(const T& x, const T& y)
//------------------------------------------------------------------------------
{
    GE_CTASSERT(DIM == 2);
    set(*this, x, y);
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>::BasicVector(const T& x, const T& y, const T& z)
//------------------------------------------------------------------------------
{
    GE_CTASSERT(DIM == 3);
    set(*this, x, y, z);
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>::BasicVector(const T& x, const T& y, const T& z, const T& w)
//------------------------------------------------------------------------------
{
    GE_CTASSERT(DIM == 4);
    set(*this, x, y, z, w);
}



template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>::BasicVector(const BasicVector& vec)
//------------------------------------------------------------------------------
{
    memcpy(mData, vec.mData, DIM*sizeof(T));
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>& BasicVector<T, DIM>::operator +=(const BasicVector& rhs)
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<DIM; i++)
        (*this)[i] += rhs[i];
    
    return *this;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>& BasicVector<T, DIM>::operator =(const BasicVector& rhs)
//------------------------------------------------------------------------------
{
    if (&rhs == this)
        return *this;

    memcpy(mData, rhs.mData, DIM*sizeof(T));
    
    return *this;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>& BasicVector<T, DIM>::operator -=(const BasicVector& rhs)
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<DIM; i++)
        (*this)[i] -= rhs[i];

    return *this;
}



template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM>& BasicVector<T, DIM>::operator *=(const T& rhs)
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<DIM; i++)
        (*this)[i] *= rhs;

    return *this;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
bool BasicVector<T, DIM>::operator ==(const BasicVector<T, DIM>& rhs) const
//------------------------------------------------------------------------------
{
    bool ret(true);
    return memcmp(mData, rhs.mData, DIM*sizeof(T)) == 0;
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
bool BasicVector<T, DIM>::operator !=(const BasicVector& rhs) const
//------------------------------------------------------------------------------
{
    return !((*this) == rhs);
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM> BasicVector<T, DIM>::operator -() const
//------------------------------------------------------------------------------
{
    BasicVector ret(*this);
    for (UInt i = 0; i<DIM; i++)
        ret[i] = -ret[i];

    return ret;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T BasicVector<T, DIM>::lengthSquare() const
//------------------------------------------------------------------------------
{
    T ret(((*this)[0])*((*this)[0]));
    for (UInt i = 1; i<DIM; i++)
        ret += (*this)[i]*(*this)[i];

    return ret;
}

template<> inline
//------------------------------------------------------------------------------
Float BasicVector<Float, 3>::lengthSquare() const
//------------------------------------------------------------------------------
{
    return (*this)[0]*(*this)[0] +
           (*this)[1]*(*this)[1] + 
           (*this)[2]*(*this)[2];
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T BasicVector<T, DIM>::length() const
//------------------------------------------------------------------------------
{
    return sqrt(lengthSquare());
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T BasicVector<T, DIM>::maxNorm() const
//------------------------------------------------------------------------------
{
    T ret(fabs((*this)[0]));
    for (UInt i = 1; i<DIM; i++)
        ret = maximum(fabs(ret), fabs((*this)[i]));

    return ret;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T BasicVector<T, DIM>::sumNorm() const
//------------------------------------------------------------------------------
{
    T ret(fabs((*this)[0]));
    for (UInt i = 1; i<DIM; i++)
        ret += fabs((*this)[i]);

    return ret;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
void BasicVector<T, DIM>::normalize()
//------------------------------------------------------------------------------
{
    GE_ASSERT(!isAlmost(ORIGIN()));
    *this *= 1/length();
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
const T& BasicVector<T, DIM>::operator [](UInt index) const
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(index < DIM, "BasicVector index out of range");
    return mData[index];
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T& BasicVector<T, DIM>::operator [](UInt index)
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(index < DIM, "BasicVector index out of range");
    return mData[index];
}


template<typename T> inline
//------------------------------------------------------------------------------
void set(BasicVector<T, 2>& v, const T& x, const T& y)
//------------------------------------------------------------------------------
{
    v[0] = x;
    v[1] = y;
}


template<typename T> inline
//------------------------------------------------------------------------------
void set(BasicVector<T, 3>& v, const T& x, const T& y, const T& z)
//------------------------------------------------------------------------------
{
    v[0] = x;
    v[1] = y;
    v[2] = z;
}

template<typename T> inline
//------------------------------------------------------------------------------
void set(BasicVector<T, 4>& v, const T& x, const T& y, const T& z, const T& w)
//------------------------------------------------------------------------------
{
    v[0] = x;
    v[1] = y;
    v[2] = z;
    v[3] = w;
}

template<typename T>
//------------------------------------------------------------------------------
void set(BasicVector<T, 4>& v4, const BasicVector<T, 3>& v3, const T& w)
//------------------------------------------------------------------------------
{
    v4[0] = v3[0];
    v4[1] = v3[1];
    v4[2] = v3[2];
    v4[3] = w;
}

template<typename T, UInt DIM>
//--------------------------------------------------------------------------------------------------
BasicVector<T, DIM> BasicVector<T, DIM>::getRandomNormal() const
//--------------------------------------------------------------------------------------------------
{
    BasicVector<T, DIM> ret;
    Int index = getFirstNonZero();
    
    if (index != -1)
    {
        UInt i;
        // construct BasicVector v with v*(*this) = 0
        for (i = 0; i<DIM; i++)
        {
            ret[i] = T(rand()%100 - 50);
            if (ret[i] == 0)
                ret[i] = 1;
        }
        
        BasicVector<T, DIM> buf(ret);

        T s1(0);
        for (i = 0; i<index; i++)
            s1 += ret[i]*mData[i];

        T s2(0);
        for (i = index+1; i<DIM; i++)
            s2 += ret[i]*mData[i];

        ret[index] = (-s1 - s2)/mData[index];
        ret.normalize();

        GE_ASSERT(!ret.isAlmost(BasicVector<T, DIM>::ORIGIN()));
    }
    return ret;
}


template<typename T, UInt DIM> inline
//--------------------------------------------------------------------------------------------------
Int BasicVector<T, DIM>::getFirstNonZero() const
//--------------------------------------------------------------------------------------------------
{
    for (Int i = 0; i<DIM; i++)
        if (mData[i] != T(0))
            return i;
    return -1;
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T BasicVector<T, DIM>::getAngleTo(const BasicVector<T, DIM>& vector) const
//------------------------------------------------------------------------------
{
    return acos((*this)*vector/(this->length()*vector.length()));
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM> BasicVector<T, DIM>::projectOn(const BasicVector<T, DIM>& vector) const
//------------------------------------------------------------------------------
{
    if ((*this)*vector == 0)
        return BasicVector<T, DIM>();
    else
        return (((*this)*vector)/vector.lengthSquare())*vector;
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
bool BasicVector<T, DIM>::isAlmost(const BasicVector<T, DIM>& rhs, T tolerance) const
//------------------------------------------------------------------------------
{
    UInt i = 0;
    while(i<DIM && ::Ge::isAlmost(mData[i], rhs[i], tolerance))
    {
        ++i;
    }

    return i == DIM;
}


template<typename T> inline
//------------------------------------------------------------------------------
BasicVector<T, 3> cross(const BasicVector<T, 3>& lhs, const BasicVector<T, 3>& rhs)
//------------------------------------------------------------------------------
{
    return BasicVector<T, 3>(lhs[1]*rhs[2] - lhs[2]*rhs[1],
                             lhs[2]*rhs[0] - lhs[0]*rhs[2],
                             lhs[0]*rhs[1] - lhs[1]*rhs[0]);
}


//------------------------------------------------------------------------------
// Global functions
//------------------------------------------------------------------------------


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
T operator *(const BasicVector<T, DIM>& lhs, const BasicVector<T, DIM>& rhs)
//------------------------------------------------------------------------------
{
    using namespace Ge;
    T ret(lhs[0]*rhs[0]);
    for (UInt i = 1; i<DIM; i++)
        ret += lhs[i]*rhs[i];

    return ret;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM> operator *(const T& lhs, const BasicVector<T, DIM>& rhs)
//------------------------------------------------------------------------------
{
    using namespace Ge;
    BasicVector<T, DIM> ret(rhs);
    return ret *= lhs;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM> operator *(const BasicVector<T, DIM>& lhs, const T& rhs)
//------------------------------------------------------------------------------
{
    return rhs*lhs;
}

template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM> operator +(const BasicVector<T, DIM>& lhs, const BasicVector<T, DIM>& rhs)
//------------------------------------------------------------------------------
{
    BasicVector<T, DIM> ret(lhs);
    return ret += rhs;
}


template<typename T, UInt DIM> inline
//------------------------------------------------------------------------------
BasicVector<T, DIM> operator -(const BasicVector<T, DIM>& lhs, const BasicVector<T, DIM>& rhs)
//------------------------------------------------------------------------------
{
    BasicVector<T, DIM> ret(lhs);
    return ret -= rhs;
}

template<typename T, UInt DIM>
//------------------------------------------------------------------------------
std::ostream& operator <<(std::ostream& os, const BasicVector<T, DIM>& vector)
//------------------------------------------------------------------------------
{
    os<<'(';
    for (UInt i = 0; i<DIM-1; i++)
        os<<vector[i]<<", ";
    return os<<vector[DIM-1]<<')';
}


} // namespace Ge





