#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "FaultFormation.h"
#include <Math/Vector.h>
#include <Math/Filter.h>

namespace Ge
{

//------------------------------------------------------------------------------
FaultFormation::FaultFormation(unsigned int xSize, unsigned int ySize,
                               float startHeight, float endHeight,
                               unsigned int iterations, unsigned int itPerFilter,
                               float filterConstant) :
//------------------------------------------------------------------------------
mXSize(xSize),
mYSize(ySize),
mStartHeight(startHeight),
mHeightDiff(endHeight-startHeight),
mIterations(iterations),
mMinHeight(0),
mMaxHeight(0),
mItPerFilter(itPerFilter),
mConstant(filterConstant),
mDone(false)
{
    GE_ASSERT(startHeight>=0 && endHeight>=0 && startHeight>endHeight);

    mMap = GE_NEW float[xSize*ySize];
    memset(mMap, 0, xSize*ySize*sizeof(float));
}
               

//------------------------------------------------------------------------------
FaultFormation::~FaultFormation()
//------------------------------------------------------------------------------
{
    GE_DELETE_A mMap;
}

//------------------------------------------------------------------------------
void FaultFormation::calculate()
//------------------------------------------------------------------------------
{
    if (!mDone)
    {
        float currentHeight(0);
        // Start and direction of separator plane. <norm> is a normal of <dir>
        Vector2 start, dir, norm;
        // Vector from start to currently passed point.
        Vector2 curDir;
        for (unsigned int i = 0; i<mIterations; ++i)
        {
            currentHeight = mIterations == 1 ?
                            mStartHeight :
                            mStartHeight + (i/(mIterations-1))*mHeightDiff;

            set(start, Float(rand()%mXSize), Float(rand()%mYSize));
            dir = Vector2(rand()%mXSize, rand()%mYSize) - start;
            set(norm, dir[1], -dir[0]);
        
            // Raise points on one side of separator plane
            for (unsigned int y = 0; y<mYSize; ++y)
            {
                for (unsigned int x = 0; x<mXSize; ++x)
                {
                    curDir = Vector2(x, y) - start;
                    // If current point is on "posititive" side of separator
                    // plane, then add <currentHeight>.
                    if (curDir*norm > 0)
                        mMap[y*mXSize + x] += currentHeight;
                }
            }

            if ((i%mItPerFilter == 0) && (mConstant != 0))
            {
                firSquare(mMap, mXSize, mYSize, mConstant, Filter::BIDIRECTIONAL);
            }
        }

        calculateMinMax();
        mDone = true;
    }
}

//------------------------------------------------------------------------------
float FaultFormation::get(unsigned int x, unsigned int y) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(x<mXSize && y<mYSize);
    return mMap[y*mXSize+x];
}

//------------------------------------------------------------------------------
unsigned int FaultFormation::getXSize() const
//------------------------------------------------------------------------------
{
    return mXSize;
}

//------------------------------------------------------------------------------
unsigned int FaultFormation::getYSize() const
//------------------------------------------------------------------------------
{
    return mYSize;
}

//------------------------------------------------------------------------------
float FaultFormation::getMaxHeight() const
//------------------------------------------------------------------------------
{
    return mMaxHeight;
}

//------------------------------------------------------------------------------
float FaultFormation::getMinHeight() const
//------------------------------------------------------------------------------
{
    return mMinHeight;
}

//------------------------------------------------------------------------------
void FaultFormation::calculateMinMax()
//------------------------------------------------------------------------------
{
    mMinHeight = mMaxHeight = mMap[0];
    float current(0);
    unsigned int size(getXSize()*getYSize());
    for (unsigned int i = 1; i<size; ++i)
    {
        current = mMap[i];
        if (current > mMaxHeight)
            mMaxHeight = current;
        else if (current < mMinHeight)
            mMinHeight = current;
    }
}

//------------------------------------------------------------------------------
const float* FaultFormation::getMap() const
//------------------------------------------------------------------------------
{
    return mMap;
}

} // namespace Ge


