#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 02-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "GerstnerWaves.h"
#include <Math/Vector.h>

namespace
{

const Ge::Float G(9.81f);

} // namespace

namespace Ge
{

//------------------------------------------------------------------------------
GerstnerWaves::GerstnerWaves(UInt xSize, UInt ySize, Float xWorld, Float yWorld) :
//------------------------------------------------------------------------------
mXSize(xSize),
mYSize(ySize),
mXWldSz(xWorld/xSize),
mYWldSz(yWorld/ySize)
{
    mVertices = GE_NEW Vector3[mXSize*mYSize];
    mIndices = GE_NEW UInt16[(mXSize-1)*(mYSize-1)*6];
    
    UInt indexPos(0);
    UInt16 cornerVertex[4];
    UInt16* currentSquare;
    for (UInt y = 0; y<mYSize-1; ++y)
    {
        for (UInt x = 0; x<mXSize-1; ++x)
        {
            // Indices of the corners of one square
            cornerVertex[0] = y*mXSize+x;
            cornerVertex[1] = cornerVertex[0]+1;
            cornerVertex[2] = cornerVertex[0]+mXSize+1;
            cornerVertex[3] = cornerVertex[2]-1;
     
            // Triangulate current square
            currentSquare = mIndices+indexPos;
            currentSquare[0] = cornerVertex[0];
            currentSquare[1] = cornerVertex[1];
            currentSquare[2] = cornerVertex[2];

            currentSquare[3] = cornerVertex[0];
            currentSquare[4] = cornerVertex[2];
            currentSquare[5] = cornerVertex[3];

            indexPos += 6;
        }
    }
}

//------------------------------------------------------------------------------
UInt GerstnerWaves::getVertexCount() const
//------------------------------------------------------------------------------
{
    return mXSize*mYSize;
}

//------------------------------------------------------------------------------
Vector3* GerstnerWaves::getVertices()
//------------------------------------------------------------------------------
{
    return mVertices;
}

//------------------------------------------------------------------------------
UInt GerstnerWaves::getIndexCount() const
//------------------------------------------------------------------------------
{
    return (mXSize-1)*(mYSize-1)*6;
}

//------------------------------------------------------------------------------
UInt16* GerstnerWaves::getIndices()
//------------------------------------------------------------------------------
{
    return mIndices;
}

//------------------------------------------------------------------------------
void GerstnerWaves::update(const Timer::Unit& t)
//------------------------------------------------------------------------------
{
    for (UInt y = 0; y<mYSize; ++y)
    {
        UInt yLine = y*mXSize;
        for (UInt x = 0; x<mXSize; ++x)
        {
            Vector2 deltaXZ(0, 0);
            Float deltaY(0);
            // Center the water around (0, 0)
            Vector2 x0((int(x)-int(mXSize)/2)*mXWldSz,
                       (int(mYSize)/2-int(y))*mYWldSz);
            for (UInt i = 0; i<mWaves.size(); ++i)
            {
                Wave& w = mWaves[i];
                Float temp = w.mK*x0 - w.mFrequency*t + w.mPhase;
                deltaXZ += w.mK*(1.0f/w.mLength)*w.mAmplitude*sin(temp);
                deltaY += w.mAmplitude*cos(temp);
            }

            mVertices[yLine+x][0] = x0[0] - deltaXZ[0];
            mVertices[yLine+x][1] = deltaY;
            mVertices[yLine+x][2] = x0[1] - deltaXZ[1];
        }
    }
}

//------------------------------------------------------------------------------
void GerstnerWaves::addWave(const Vector2& k, Float length,
                            Float amplitude, Float phase)
//------------------------------------------------------------------------------
{
    Wave w;
    w.mK = k;
    w.mK.normalize();
    w.mLength = TWO_PI/length;
    w.mK *= w.mLength;
    w.mAmplitude = amplitude;
    w.mFrequency = sqrt(G*w.mLength);
    w.mPhase = phase;
    mWaves.push_back(w);
}

} // namespace Ge