#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "HeightMapConverter.h"
#include "Terrain.h"

#include <Math/Interpolation.h>

namespace Ge
{

//------------------------------------------------------------------------------
void HeightMapConverter::createMesh(const Terrain& map,
    UInt& vertexCount, Vector3*& vertices,
    Vector3** normals, Vector2** uvs,
    UInt* indexCount, UInt16** indices) const
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(map.getMap().getXSize() > 1 && map.getMap().getYSize() > 1, "HeightMap size invalid");
    GE_ASSERT_MSG(map.getMap().getXSize()*map.getMap().getYSize() < 65536, "Too many vertices for 16 bit indices");

    if (map.getMap().getXSize() <= 1 || map.getMap().getYSize() <= 1 ||
        map.getMap().getXSize()*map.getMap().getYSize() >= 65536)
    {
        // Map invalid
        vertexCount = 0;
        vertices = 0;
        if (normals != 0)
            *normals = 0;
        if (uvs != 0)
            *uvs = 0;
        
        indexCount = 0;
        indices = 0;
    }
    else
    {
        // Build the mesh
        if (map.getDeltaXzMap().getXSize() > 0)
            buildVertexArrayWithDeltaXzMap(map, vertexCount, vertices, normals, uvs);
        else
            buildVertexArray(map, vertexCount, vertices, normals, uvs);

        // Triangulate it
        if (indices != 0 && indexCount != 0)
            triangulate(map, *indexCount, *indices);
    }
}
             
//------------------------------------------------------------------------------
void HeightMapConverter::buildVertexArray(const Terrain& map,
    UInt& vertexCount, Vector3*& vertices,
    Vector3** normals, Vector2** uvs) const
//------------------------------------------------------------------------------
{
    UInt width(map.getMap().getXSize());
    UInt height(map.getMap().getYSize());
    vertexCount = width*height;

    // Allocate memory if needed
    if (mAllocateMemory)
    {
        vertices = GE_NEW Vector3[vertexCount];
        if (normals != 0)
            *normals = GE_NEW Vector3[vertexCount];

        if (uvs != 0)
            *uvs = GE_NEW Vector2[vertexCount];
    }

    Float left(-Float(width-1)*map.getXZStep()/2);
    Float x(left);
    Float z(Float(height-1)*map.getXZStep()/2);
    Float stepY = (map.getYRange().getRight()-map.getYRange().getLeft())/255;
    for (Int yi = 0; yi<height; ++yi)
    {
        for (Int xi = 0; xi<width; ++xi)
        {
            // Calculate model coordinate for this vertex
            set(vertices[yi*width+xi], x, lerp(0, 255, map.getYRange().getLeft(), map.getYRange().getRight(), map.getMap().get(xi, yi)), z);

            // Calculate normal for this vertex
            if (normals != 0)
            {
                Float sx = map.getMap().get(xi<width-1 ? xi+1 : xi, yi) - map.getMap().get(xi>0 ? xi-1 : xi, yi);
                if (xi == 0 || xi == width-1)
                    sx *= 2;

                Float sy = map.getMap().get(xi, yi<height-1 ? yi+1 : yi) - map.getMap().get(xi, yi>0 ?  yi-1 : yi);
                if (yi == 0 || yi == height -1)
                    sy *= 2;
                
                set((*normals)[yi*width+xi], -sx*stepY, 2*map.getXZStep(), sy*stepY);
                (*normals)[yi*width+xi].normalize();
            }

            // Calculate texture coordinate for this vertex
            if (uvs != 0)
                set((*uvs)[yi*width+xi], Float(xi)/Float(width-1), Float(yi)/Float(height-1));

            // Next pixel in height map
            x += map.getXZStep();
        }

        // Next line
        x = left;
        z -= map.getXZStep();
    }
}                         

//------------------------------------------------------------------------------
void HeightMapConverter::buildVertexArrayWithDeltaXzMap(
        const Terrain& map, UInt& vertexCount, Vector3*& vertices,
        Vector3** normals, Vector2** uvs) const
//------------------------------------------------------------------------------
{
    UInt width(map.getMap().getXSize());
    UInt height(map.getMap().getYSize());
    vertexCount = width*height;

    // Allocate memory if needed
    if (mAllocateMemory)
    {
        vertices = GE_NEW Vector3[vertexCount];
        if (normals != 0)
            *normals = GE_NEW Vector3[vertexCount];

        if (uvs != 0)
            *uvs = GE_NEW Vector2[vertexCount];
    }

    Float left(-Float(width-1)*map.getXZStep()/2);
    Float x(left);
    Float z(Float(height-1)*map.getXZStep()/2);
    Float stepY = (map.getYRange().getRight()-map.getYRange().getLeft())/255;
    std::pair<Float, Float> delta;
    for (UInt yi = 0; yi<height; ++yi)
    {
        for (UInt xi = 0; xi<width; ++xi)
        {
            delta = map.getDeltaXzMap().get(yi*width+xi);
            // Calculate model coordinate for this vertex
            set(vertices[yi*width+xi], x+delta.first,
                lerp(0, 255, map.getYRange().getLeft(), map.getYRange().getRight(), map.getMap().get(xi, yi)),
                z-delta.second);

            // Calculate texture coordinate for this vertex
            if (uvs != 0)
            {
                set((*uvs)[yi*width+xi], Float(xi)/Float(width-1), Float(yi)/Float(height-1));
            }

            // Next pixel in height map
            x += map.getXZStep();
        }

        // Next line
        x = left;
        z -= map.getXZStep();
    }

    // Do normal calculation. Need all vertex data before, because the delta
    // map moves the grid points around.
    Vector3 neighbours[4];
    if (normals != 0)
    {
        for (UInt yi = 0; yi<height; ++yi)
        {
            for (UInt xi = 0; xi<width; ++xi)
            {
                neighbours[0] = xi == width-1 ?
                                Vector3::ORIGIN() :
                                vertices[yi*width + xi+1] - vertices[yi*width+xi];
                neighbours[1] = yi == height-1 ?
                                Vector3::ORIGIN() :
                                vertices[(yi+1)*width + xi] - vertices[yi*width+xi];
                neighbours[2] = xi == 0 ?
                                Vector3::ORIGIN() :
                                vertices[yi*width + xi-1] - vertices[yi*width+xi];
                neighbours[3] = yi == 0 ?
                                Vector3::ORIGIN() :
                                vertices[(yi-1)*width + xi] - vertices[yi*width+xi];

                (*normals)[yi*width+xi] =
                    cross(neighbours[0], neighbours[1]) +
                    cross(neighbours[1], neighbours[2]) +
                    cross(neighbours[2], neighbours[3]) +
                    cross(neighbours[3], neighbours[0]);

                (*normals)[yi*width+xi].normalize();
            }
        }
    }
}

//------------------------------------------------------------------------------
void HeightMapConverter::triangulate(const Terrain& map,
    UInt& indexCount, UInt16*& indices) const
//------------------------------------------------------------------------------
{
    UInt width(map.getMap().getXSize());
    UInt height(map.getMap().getYSize());

    // Number of squares, two triangles (= 6 indices) per square.
    indexCount = (width-1)*(height-1)*6;
    if (mAllocateMemory)
        indices = GE_NEW UInt16[indexCount];
    
    // Square is clockwise 01
    //                     32
    unsigned char cornerY[4];
    UInt16 cornerVertex[4];

    // Height differences
    unsigned char topBottom(0);
    unsigned char bottomTop(0);

    // Current index position
    UInt indexPos(0);
    for (UInt i = 0; i<height-1; ++i)
    {
        for (UInt j = 0; j<width-1; ++j)
        {
            // Indices of the corners of one square
            cornerVertex[0] = i*width+j;
            cornerVertex[1] = cornerVertex[0]+1;
            cornerVertex[2] = cornerVertex[0]+width+1;
            cornerVertex[3] = cornerVertex[2]-1;

            // Height values of the square
            if (map.getRule() == MIN_HEIGHT_VARIATION ||
                map.getRule() == MAX_HEIGHT_VARIATION)
            {
                cornerY[0] = map.getMap().get(j, i);
                cornerY[1] = map.getMap().get(j+1, i);
                cornerY[2] = map.getMap().get(j+1, i+1);
                cornerY[3] = map.getMap().get(j, i+1);

                topBottom = abs(Int(cornerY[0]) - Int(cornerY[2]));
                bottomTop = abs(Int(cornerY[1]) - Int(cornerY[3]));
            }

            // Triangulate current square
            switch(map.getRule())
            {
            case MIN_HEIGHT_VARIATION:
                {    
                    if (topBottom < bottomTop)
                        triangulateTopBottom(&(indices[indexPos]), cornerVertex);
                    else
                        triangulateBottomTop(&(indices[indexPos]), cornerVertex);
                }
                break;

            case MAX_HEIGHT_VARIATION:
                {    
                    if (topBottom > bottomTop)
                        triangulateTopBottom(&(indices[indexPos]), cornerVertex);
                    else
                        triangulateBottomTop(&(indices[indexPos]), cornerVertex);
                }
                break;
                
            case TOP_TO_BOTTOM:
                triangulateTopBottom(&(indices[indexPos]), cornerVertex);
                break;

            case BOTTOM_TO_TOP:
                triangulateBottomTop(&(indices[indexPos]), cornerVertex);
                break;

            default:
                GE_ASSERT_MSG(false, "Unknown triangulation rule.");
            }
            
            indexPos += 6;
        }
    }
}

//------------------------------------------------------------------------------
void HeightMapConverter::triangulateTopBottom(UInt16* currentSquare,
                                              UInt16* corners) const
//------------------------------------------------------------------------------
{
    currentSquare[0] = corners[0];
    currentSquare[1] = corners[2];
    currentSquare[2] = corners[3];
 
    currentSquare[3] = corners[0];
    currentSquare[4] = corners[1];
    currentSquare[5] = corners[2];
}

//------------------------------------------------------------------------------
void HeightMapConverter::triangulateBottomTop(UInt16* currentSquare,
                                              UInt16* corners) const
//------------------------------------------------------------------------------
{
    currentSquare[0] = corners[0];
    currentSquare[1] = corners[1];
    currentSquare[2] = corners[3];
 
    currentSquare[3] = corners[1];
    currentSquare[4] = corners[2];
    currentSquare[5] = corners[3];    
}

//------------------------------------------------------------------------------
HeightMapConverter::HeightMapConverter(bool allocateMemory) :
//------------------------------------------------------------------------------
mAllocateMemory(allocateMemory)
{
    
}

} // namespace Ge

