#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "MidpointDisplacement.h"
#include <Math/Random.h>
#include <Math/Tools.h>
#include <Math/Filter.h>
#include <System/Assert.h>
#include <System/Memory.h>

#include <string.h>

namespace Ge
{

//------------------------------------------------------------------------------
MidpointDisplacement::MidpointDisplacement(unsigned int sizePower, float displacement,
                                           float roughness,
                                           float filterConstant) :
//------------------------------------------------------------------------------
mLevels(sizePower),
mDisplacement(displacement),
mDone(false),
mConstant(filterConstant)
{
    // Well, 2^12 + 1 should be enough.
    GE_ASSERT_MSG(sizePower < 12, "Map too big.");
    mSize = pow(2, sizePower)+1;
    mMultiplier = pow(2, -roughness);
    mMap = GE_NEW float[mSize*mSize];
    memset(mMap, 0, mSize*mSize*sizeof(float));
}
               
//------------------------------------------------------------------------------
MidpointDisplacement::~MidpointDisplacement()
//------------------------------------------------------------------------------
{
    GE_DELETE_A mMap;
}

//------------------------------------------------------------------------------
void MidpointDisplacement::calculate()
//------------------------------------------------------------------------------
{
    if (!mDone)
    {
        for (unsigned int level = 0; level<mLevels; ++level)
        {
            diamond(level, mDisplacement/2.0f);
            square(level, mDisplacement/2.0f);

            mDisplacement *= mMultiplier;
        }
        
        if (mConstant != 0)
        {
            firSquare(mMap, mSize, mSize, mConstant, Filter::BIDIRECTIONAL);
        }

        calculateMinMax();
        mDone = true;
    }
}

//------------------------------------------------------------------------------
void MidpointDisplacement::diamond(unsigned int level, float displacement)
//------------------------------------------------------------------------------
{
    unsigned int step(pow(2, mLevels-level));
    unsigned int substep(step/2);
    float value(0);
    for (unsigned int y = 0; y<mSize-1; y+=step)
    {
        for (unsigned int x = 0; x<mSize-1; x+=step)
        {
            // Calc E (see diagram in "Game Programming Gems")
            value = (get(x, y) + get(x+step, y) +
                     get(x, y+step) + get(x+step, y+step))/4.0f +
                     frand(-displacement, displacement);

            set(x+substep, y+substep, value);
        }
    }
}

//------------------------------------------------------------------------------
void MidpointDisplacement::square(unsigned int level, float displacement)
//------------------------------------------------------------------------------
{
    unsigned int step(pow(2, mLevels-level));
    unsigned int substep(step/2);
    float value(0);
    for (unsigned int y = 0; y<mSize-1; y+=step)
    {
        for (unsigned int x = 0; x<mSize-1; x+=step)
        {
            // Calc F
            value = (get(x, y) + get(x+step, y) +
                get(x+substep, y+substep) + get(x+substep, y>0 ? y-substep : mSize-1-substep))/4.0f +
                frand(-displacement, displacement);
            set(x+substep, y, value);

            // Calc G
            value = (get(x>0 ? x-substep : mSize-1-substep, y+substep) + get(x+substep, y+substep) +
                get(x, y+step) + get(x, y))/4.0f +
                frand(-displacement, displacement);
            set(x, y+substep, value);

            // Special case right border (terrain wraps)
            if (x == mSize-1-step)
            {
                // Calc H
                value = (get(x+substep, y+substep) + get(substep, y+substep) +
                    get(x+step, y+step) + get(x+step, y))/4.0f +
                    frand(-displacement, displacement);
                set(x+step, y+substep, value);
            }

            // Special case bottom border (terrain wraps)
            if (y == mSize-1-step)
            {
                // Calc I
                value = (get(x, y+step) + get(x+step, y+step) +
                    get(x+substep, substep) + get(x+substep, y+substep))/4.0f +
                    frand(-displacement, displacement);
                set(x+substep, y+step, value);
            }
        }
    }
}


//------------------------------------------------------------------------------
unsigned int MidpointDisplacement::getSize() const
//------------------------------------------------------------------------------
{
    return mSize;
}

//------------------------------------------------------------------------------
float MidpointDisplacement::get(unsigned int x, unsigned int y) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(x<mSize && y<mSize);
    return mMap[y*mSize+x];
}

//------------------------------------------------------------------------------
void MidpointDisplacement::set(unsigned int x, unsigned int y, float value)
//------------------------------------------------------------------------------
{
    GE_ASSERT(x<mSize && y<mSize);
    mMap[y*mSize+x] = value;
}

//------------------------------------------------------------------------------
float MidpointDisplacement::getMaxHeight() const
//------------------------------------------------------------------------------
{
    return mMaxHeight;
}

//------------------------------------------------------------------------------
float MidpointDisplacement::getMinHeight() const
//------------------------------------------------------------------------------
{
    return mMinHeight;
}

//------------------------------------------------------------------------------
void MidpointDisplacement::calculateMinMax()
//------------------------------------------------------------------------------
{
    mMinHeight = mMaxHeight = mMap[0];
    float current(0);
    unsigned int size(getSize()*getSize());
    for (unsigned int i = 1; i<size; ++i)
    {
        current = mMap[i];
        if (current > mMaxHeight)
            mMaxHeight = current;
        else if (current < mMinHeight)
            mMinHeight = current;
    }
}

//------------------------------------------------------------------------------
const float* MidpointDisplacement::getMap() const
//------------------------------------------------------------------------------
{
    return mMap;
}

} // namespace Ge


