#include "Precompiled.h"

#include "Newave.h"

#include <System/Base.h>

#include <Math/Tools.h>

#include <string.h>

namespace Ge
{

// Dampening term of the water surface (must be in [0,1], where 0 is
// instant dampening and 1 is none.
const Float Newave::DAMPENING = 0.99f;

//------------------------------------------------------------------------------
Newave::Newave(UInt size, Float speed) :
//------------------------------------------------------------------------------
mSpeed(speed),
mSize(size)
{
    mForce = GE_NEW Float[size*size];
    mPosition = GE_NEW Float[size*size];
    mVelocity = GE_NEW Float[size*size];
}

//------------------------------------------------------------------------------
Newave::~Newave()
//------------------------------------------------------------------------------
{
    if (mForce != 0)
    {
        GE_DELETE_A mForce;
    }

    if (mPosition != 0)
    {
        GE_DELETE_A mPosition;
    }

    if (mVelocity != 0)
    {
        GE_DELETE_A mVelocity;
    }

    mSize = 0;
}

//------------------------------------------------------------------------------
void Newave::update()
//------------------------------------------------------------------------------
{
    memset(mForce, 0, sizeof(Float)*mSize*mSize);

    Float d(0);
    Array2D<Float> force(mForce, mSize, mSize);
    Array2D<Float> position(mPosition, mSize, mSize);
    for (UInt y=2; y<mSize-2; y++)
    {
        for (UInt x=2; x<mSize-2; x++) 
        {
            d = position(y, x)-position(y, x-1);
            force(y, x) -= d;
            force(y, x-1) += d;

            d = position(y, x)-position(y-1, x);
            force(y, x) -= d;
            force(y-1, x) += d;

            d = (position(y, x)-position(y, x+1)); 
            force(y, x) -= d;
            force(y, x+1) += d;

            d = (position(y, x)-position(y+1, x)); 
            force(y, x) -= d;
            force(y+1, x) += d;

            d = (position(y, x)-position(y+1, x+1))*INV_SQRT_TWO; 
            force(y, x) -= d;
            force(y+1, x+1) += d;

            d = (position(y, x)-position(y-1, x-1))*INV_SQRT_TWO; 
            force(y, x) -= d;
            force(y-1, x-1) += d;

            d = (position(y, x)-position(y+1, x-1))*INV_SQRT_TWO; 
            force(y, x) -= d;
            force(y+1, x-1) += d;

            d = (position(y, x)-position(y-1, x+1))*INV_SQRT_TWO; 
            force(y, x) -= d;
            force(y-1, x+1) += d;
        }
    }

    UInt index;
    for(y=1; y<mSize-1; y++)
    {
        for(UInt x=1; x<mSize-1; x++)
        {
            index = y*mSize + x;
            mVelocity[index] += mForce[index]*mSpeed;
            mPosition[index] = (mPosition[index] + mVelocity[index])*DAMPENING;
        }
    }
}

//------------------------------------------------------------------------------
void Newave::reset(Mode mode)
//------------------------------------------------------------------------------
{
    memset(mForce, 0, sizeof(Float)*mSize*mSize);
    memset(mVelocity, 0, sizeof(Float)*mSize*mSize);
    Array2D<Float> position(mPosition, mSize, mSize);
    for(UInt y=0; y<mSize; y++)
    {
        for(UInt x=0; x<mSize; x++)
        {
            if (y==0 || x==0 || y==mSize-1 || x==mSize-1)
                position(y, x)=0.0;
            else
            {
                switch(mode)
                {
                case FLAT:
                    position(y, x) = 0.0;
                    break;

                case SPIKE:
                    position(y, x)= (y==x && y == mSize/2) ? mSize*1.5 : 0.0;
                    break;

                case HOLE:
                    position(y, x)= (!((y > mSize/3 && x > mSize/3)&&(y < mSize*2/3 && x < mSize*2/3))) ? mSize/4 : 0.0;
                    break;

                case DIAGONALWALL:
                    position(y, x)= (((mSize-y)-x<3) && ((mSize-y)-x>0)) ? mSize/6 : 0.0;
                    break;

                case SIDEWALL:
                    position(y, x)= (y==1) ? mSize/4 : 0.0;
                    break;

                case DIAGONALBLOCK:
                    position(y, x)= ((mSize-y)-x<3) ? mSize/6 : 0.0;
                    break;

                case MIDDLEBLOCK:
                    position(y, x)= ((y > mSize/3 && x > mSize/3)&&(y < mSize*2/3 && x < mSize*2/3)) ? mSize/4 : 0.0;
                    break;

                case CORNERBLOCK:
                    position(y, x)= ((y > mSize*3/4 && x > mSize*3/4)) ? mSize/4 : 0.0;
                    break;

                case HILL:
                    position(y, x)= 
                        (sin(PI * ((Float)y/(Float)mSize)) +
                         sin(PI * ((Float)x/(Float)mSize)))* mSize/6.0;
                    break;        

                case HILLFOUR:
                    position(y, x)= 
                        (sin(PI*2 * ((Float)y/(Float)mSize)) +
                         sin(PI*2 * ((Float)x/(Float)mSize)))* mSize/6.0;
                    break;        
                }
            }
        }
    }
}


//------------------------------------------------------------------------------
void Newave::drop(UInt x, UInt y)
//------------------------------------------------------------------------------
{
    GE_ASSERT(x>=1 && y>=1);
    GE_ASSERT(x<=mSize-2 && y<=mSize-2);

    mPosition[y*mSize+x] = mSize*0.2f;
}

//------------------------------------------------------------------------------
UInt Newave::getSize() const
//------------------------------------------------------------------------------
{
    return mSize;
}

//------------------------------------------------------------------------------
const Float* Newave::getMap() const
//------------------------------------------------------------------------------
{
    return mPosition;
}

} // namespace Ge