#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "ParticleDeposition.h"
#include <System/Base.h>
#include <Math/Tools.h>
#include <Math/Random.h>
#include <Math/Filter.h>

namespace
{
    // For a point x, y going through all x+xDelta, y+yDelta gives all
    // neighbouring points.
    int xDelta[] = {  1, 1, 1, 0, -1, -1, -1, 0 };
    int yDelta[] = { -1, 0, 1, 1,  1,  0, -1, 1 };
} // namespace


namespace Ge
{

//------------------------------------------------------------------------------
ParticleDeposition::ParticleDeposition(unsigned int xSize, unsigned int ySize,
    float filterConstant, unsigned int dropsPerStep, unsigned int mountains,
    unsigned int minParticlesPerMountain, unsigned int maxParticlesPerMountain,
    float invertCalderaPercentage) :
//------------------------------------------------------------------------------
mXSize(xSize),
mYSize(ySize),
mConstant(filterConstant),
mDrops(dropsPerStep),
mMountains(mountains),
mMinPPM(minParticlesPerMountain),
mMaxPPM(maxParticlesPerMountain),
mDone(false),
mPercentage(invertCalderaPercentage)
{
    GE_ASSERT(xSize > 1 && ySize > 1);
    GE_ASSERT(mPercentage>0.5f && mPercentage<=1.0f);
    mMap = GE_NEW unsigned char[mXSize*mYSize];
    memset(mMap, 0, mXSize*mYSize);
}
               
//------------------------------------------------------------------------------
ParticleDeposition::~ParticleDeposition()
//------------------------------------------------------------------------------
{
    GE_DELETE_A mMap;
}

//------------------------------------------------------------------------------
void ParticleDeposition::calculate()
//------------------------------------------------------------------------------
{
    if (!mDone)
    {
        unsigned int particleCount;
        unsigned int dpx, dpy; // dp = drop point
        Particle particle(*this);
        for (unsigned int i = 0; i<mMountains; ++i)
        {
            // Number of particles for this mountain.
            particleCount = rand(mMinPPM, mMaxPPM);

            // Initial drop point for this mountain
            dpx = rand(0, mXSize-1);
            dpy = rand(0, mYSize-1);
            for (unsigned int p = 0; p<particleCount; ++p)
            {
                particle.drop(dpx, dpy);
                if (p%mDrops == 0)
                {
                    unsigned int stepDirection(rand()%8);
                    dpx = (dpx+xDelta[stepDirection])%getXSize();
                    dpy = (dpy+yDelta[stepDirection])%getYSize();
                }
            }
        }

//        // Invert caldera
//        unsigned int mx, my;
//        unsigned char max;
//        findPeak(mx, my, max);
//        invert(mx, my, mPercentage*float(max));

        normalize();

        // Filter
        if (mConstant != 0)
            firSquare(mMap, getXSize(), getYSize(), mConstant, Filter::BIDIRECTIONAL);

        mDone = true;
    }
}

//------------------------------------------------------------------------------
void ParticleDeposition::normalize()
//------------------------------------------------------------------------------
{
    unsigned int x, y;
    unsigned char max;
    findPeak(x, y, max);

    // Normalize
    unsigned int size(getXSize()*getYSize());
    for (unsigned int j = 0; j<size; ++j)
    {
        mMap[j] = unsigned char(255.0f*(float(mMap[j])/float(max)));
    }
}

//------------------------------------------------------------------------------
void ParticleDeposition::invert(unsigned int x, unsigned int y,
                                unsigned char limit)
//------------------------------------------------------------------------------
{
    unsigned char currentHeight(get(x, y));
    mMap[y*mXSize + x] = unsigned char(2*int(limit)-int(currentHeight));
    for (unsigned int i = 0; i<8; ++i)
    {
        if (get((x+xDelta[i])%getXSize(), (y+yDelta[i])%getYSize()) > limit)
        {
            invert((x+xDelta[i])%getXSize(), (y+yDelta[i])%getYSize(), limit);
        }
    }
}

//------------------------------------------------------------------------------
unsigned int ParticleDeposition::getXSize() const
//------------------------------------------------------------------------------
{
    return mXSize;
}

//------------------------------------------------------------------------------
unsigned int ParticleDeposition::getYSize() const
//------------------------------------------------------------------------------
{
    return mYSize;
}

//------------------------------------------------------------------------------
unsigned char ParticleDeposition::get(unsigned int x, unsigned int y) const
//------------------------------------------------------------------------------
{
    GE_ASSERT(x < mXSize && y < mYSize);
    return mMap[y*mXSize + x];
}  

//------------------------------------------------------------------------------
void ParticleDeposition::raise(unsigned int x, unsigned int y)
//------------------------------------------------------------------------------
{
    GE_ASSERT(x < mXSize && y < mYSize);
    mMap[y*mXSize + x] += mMap[y*mXSize + x]<255 ? 1 : 0;
}

//------------------------------------------------------------------------------
void ParticleDeposition::findPeak(unsigned int& x, unsigned int& y,
                                  unsigned char& height) const
//------------------------------------------------------------------------------
{
    height = 0;
    unsigned int maxPos(0);
    unsigned int size(getXSize()*getYSize());
    // Find max
    for (unsigned int i = 0; i<size; ++i)
    {
        if (mMap[i] > height)
        {
            height = mMap[i];
            maxPos = i;
        }
    }

    x = maxPos%mXSize;
    y = maxPos/mXSize;
}

//------------------------------------------------------------------------------
const unsigned char* ParticleDeposition::getMap() const
//------------------------------------------------------------------------------
{
    return mMap;
}


//------------------------------------------------------------------------------
// Particle members
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
void ParticleDeposition::Particle::drop(int x, int y)
//------------------------------------------------------------------------------
{
    // Search for lower neighbour.
    unsigned int searchStart(rand());
    unsigned int currentSearchPoint;
    for (unsigned int i = 0; i<8; ++i)
    {
        currentSearchPoint = (searchStart+i)%8;
        if (mPd.get((x+xDelta[currentSearchPoint])%mPd.getXSize(),
                    (y+yDelta[currentSearchPoint])%mPd.getYSize()) <
            mPd.get(x, y))
        {
            drop((x+xDelta[currentSearchPoint])%mPd.getXSize(),
                 (y+yDelta[currentSearchPoint])%mPd.getYSize());
            return;
        }
    }

    // No neighbour found. Raise point.
    mPd.raise(x, y);
}

//------------------------------------------------------------------------------
ParticleDeposition::Particle::Particle(ParticleDeposition& pd) :
//------------------------------------------------------------------------------
mPd(pd)
{
}

} // namespace Ge


