#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "StatisticalWater.h"
#include <System/Trace.h>
#include <Math/Random.h>


//------------------------------------------------------------------------------
// Helpers
//------------------------------------------------------------------------------

namespace
{

using namespace Ge;

const Float G = 9.81f;

inline
//------------------------------------------------------------------------------
Float phillips(Float a, const Vector2& k, const Vector2& wind)
//------------------------------------------------------------------------------
{
    Float k2 = k.lengthSquare();
    if (k2 == 0)
        return 0;

    Float v2 = wind.lengthSquare();
    Float l = v2/G;

    Float ret = a*
                (exp(-1/(k2*square(l))) / square(k2))*
                (square(k*wind) / (k2*v2));

    GE_ASSERT_MSG(ret>=0, "We take the sqrt of this value later, so it must be positive");
    return ret;
}

} // namespace







//------------------------------------------------------------------------------
// Members
//------------------------------------------------------------------------------

namespace Ge
{

//------------------------------------------------------------------------------
StatisticalWater::StatisticalWater(UInt xPower, UInt yPower,
                                   Float a, const Vector2& wind,
                                   Float xWorld, Float yWorld) :
//------------------------------------------------------------------------------
mA(a),
mWind(wind),
m2PiByXWorld(TWO_PI/xWorld),
m2PiByYWorld(TWO_PI/yWorld),
mXPower(xPower),
mYPower(yPower),
mFft(xPower, yPower)
{
    mXSize = 1<<xPower;
    mYSize = 1<<yPower;
    
    mH0 = GE_NEW Complex[mYSize*mXSize];
    mh = GE_NEW Complex[mYSize*mXSize];

    calculateH0();
}

//------------------------------------------------------------------------------
StatisticalWater::~StatisticalWater()
//------------------------------------------------------------------------------
{
    GE_DELETE_A mH0;
    GE_DELETE_A mh;
}


//------------------------------------------------------------------------------
void StatisticalWater::calculateH0()
//------------------------------------------------------------------------------
{    
    UInt yLine;
    for (UInt y = 0; y<mYSize; ++y)
    {
        yLine = y*mXSize;
        for (UInt x = 0; x<mXSize; ++x)
        {
            Complex zeta(gaussian(), gaussian());
            
            mH0[yLine+x] = INV_SQRT_TWO*zeta*
                sqrt(phillips(mA, Vector2(indexToWorldX(x), indexToWorldY(y)), mWind));    
        }
    }
}

//------------------------------------------------------------------------------
void StatisticalWater::update(const Timer::Unit& t)
//------------------------------------------------------------------------------
{
    Float kl;
    Float wkt; // x(k)*t
    Complex ep, em; // exp(i*w(k)*t), exp(-i*w(k)*t)
    UInt yLineMirr; // yLine mirrored at mYSize/2
    UInt kNegIndex; // Array index of vector -k

    Complex result;
    UInt yLine;
    Vector2 k;

    // We need only to loop over the half map, because the rest of
    // the values gets calculated by taking the conjugate.
    UInt yHalf = mYSize/2 + 1;
    for (UInt y = 0; y<yHalf; ++y)
    {
        yLine = y*mXSize;
        // Mirror the y line index for calculation of -k
        // The line below evalutes yLineMirr = y == 0 ? 0 : (mYSize-y)*mXSize;
        // by wrapping the heightmap, since it is periodic.
        yLineMirr = mod(mYSize-y, mYSize)*mXSize;
        for (UInt x = 0; x<mXSize; ++x)
        {
            k = Vector2(indexToWorldX(x), indexToWorldY(y));
            kl = k.length();
      
            wkt = sqrt(kl*G)*t;
           
            ep = Complex(cos(wkt), sin(wkt));
            em = std::conj(ep);

            // yLineMirr+mXSize-x is the index of -k
            kNegIndex = yLineMirr + mod(mXSize-x, mXSize);
  
            // This is h~(K, t) from the Tessendorf paper.
            result = mH0[yLine+x]*ep +
                     std::conj(mH0[kNegIndex])*em;

            // Store it for later transformation
            store(yLine+x, k, kl, result);
            // h~(-K) = conj(h~(K))
            if (y != yHalf-1)
            {
                store(kNegIndex, -k, kl, std::conj(result));
            }
        }
    }

    // Do the FFT to the time domain.
    transformToTime();
}


//------------------------------------------------------------------------------
void StatisticalWater::store(UInt index, const Vector2&, Float, const Complex& result)
//------------------------------------------------------------------------------
{
    mh[index] = result;
}

//------------------------------------------------------------------------------
void StatisticalWater::transformToTime()
//------------------------------------------------------------------------------
{
    getFft().calculate(mh);
    
    UInt yLine(0);
    for (UInt y = 0; y<mYSize; ++y)
    {
        yLine = y*mXSize;
        for (UInt x = 0; x<mXSize; ++x)
        {
            // The neg1Pow term results from shifting the sum index from [-N/2, N)
            // to [0, N)
            mh[yLine+x] *= Float(neg1Pow(x+y));
        }
    }
}

//------------------------------------------------------------------------------
const Complex* StatisticalWater::getMap() const
//------------------------------------------------------------------------------
{
    return mh;
}

//------------------------------------------------------------------------------
UInt StatisticalWater::getXSize() const
//------------------------------------------------------------------------------
{
    return mXSize;
}

//------------------------------------------------------------------------------
UInt StatisticalWater::getYSize() const
//------------------------------------------------------------------------------
{
    return mYSize;
}

} // namespace Ge