//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef STATISTICAL_WATER_H
#define STATISTICAL_WATER_H

#include <System/Types.h>
#include <Math/Vector.h>
#include <System/Timer.h>
#include <Math/Fft.h>


namespace Ge
{

// Heightmap:
// x is right
// y is up
class StatisticalWater
{
public:
    // Parameters:
    // - <x>, <y> are the sizes in world units (i.e. meters).
    // - 2^<rowPower>, 2^<columnPower> are the number of rows, columns in
    //   the array used to represent the water.
    // - <a> is a numerical constant used for the Phillips spectrum.
    // - <wind> is a vector in world units describing the wind direction
    //   and speed.
    //
    // Reasonable values are:
    // (6, 6, 0.0008f, Vector2(30,10), 100.0f, 100.0f);
    StatisticalWater(UInt xPower, UInt yPower,
                     Float a, const Vector2& wind,
                     Float xWorld, Float yWorld);
    ~StatisticalWater();

    // Time is absolute (i.e. not relative to last frame)
    void update(const Timer::Unit&);

    const Complex* getMap() const;
    UInt getXSize() const;
    UInt getYSize() const;

protected:
    UInt getXPower() const;
    UInt getYPower() const;

    // The tick() function calls store() in a loop over the whole grid.
    // tick() only calculates the h~(K, t) value from the Tessendorf paper.
    // store() stores the value into the frequency domain map for later
    // transformation. ChoppyWaves additionally stores values for the choppy
    // delta map.
    // <index> is the current index in mh.
    // <result> is h~(K, t)
    // <k> is K
    // <kLen> is <k>.length()
    virtual
    void store(UInt index, const Vector2& k, Float kLen, const Complex& result);

    // After calculating the heightfield in the frequency domain. Transform
    // it back to the time domain.
    virtual
    void transformToTime();

    const Fft2D& getFft() const;

private:
    void calculateH0();
    // Transforms an array index into a world coordinate. (The array has a
    // [mRows][mColumns] index range. But the grid is centered at 0).
    Float indexToWorldX(UInt index);
    Float indexToWorldY(UInt index);

    Complex* mH0;
    Complex* mh;

    Float mA;
    Vector2 mWind;
    Float m2PiByXWorld;
    Float m2PiByYWorld;
    UInt mXPower;
    UInt mYPower;
    UInt mXSize;
    UInt mYSize;
    Fft2D mFft;
};

inline
//------------------------------------------------------------------------------
UInt StatisticalWater::getXPower() const
//------------------------------------------------------------------------------
{
    return mXPower;
}

inline
//------------------------------------------------------------------------------
UInt StatisticalWater::getYPower() const
//------------------------------------------------------------------------------
{
    return mYPower;
}

inline
//------------------------------------------------------------------------------
Float StatisticalWater::indexToWorldX(UInt index)
//------------------------------------------------------------------------------
{
    return (Float(index) - mXSize/2)*(m2PiByXWorld);
}

inline
//------------------------------------------------------------------------------
Float StatisticalWater::indexToWorldY(UInt index)
//------------------------------------------------------------------------------
{
    return (Float(index) - mYSize/2)*(m2PiByYWorld);
}

inline
//------------------------------------------------------------------------------
const Fft2D& StatisticalWater::getFft() const
//------------------------------------------------------------------------------
{
    return mFft;
}

} // namespace Ge

#endif // STATISTICAL_WATER_H
