#include "Precompiled.h"
#include "Terrain.h"
#include "HeightMapConverter.h"
#include "TextureBlender.h"

namespace Ge
{

//------------------------------------------------------------------------------
// Ground members
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
Terrain::Ground::Ground(const char* name,
                               const Interval& height, float heightRelease,
                               const Interval& slope, float slopeRelease) :
//------------------------------------------------------------------------------
mHeight(height),
mHeightRelease(heightRelease),
mSlope(slope),
mSlopeRelease(slopeRelease)
{
    mImage = GE_NEW Image(name);
    mImage->convertTo(Image::RGB_888);
}

//------------------------------------------------------------------------------
void Terrain::Ground::getHeightParameters(Interval& height,
                                                 float& heightRelease) const
//------------------------------------------------------------------------------
{
    height = mHeight;
    heightRelease = mHeightRelease;
}

//------------------------------------------------------------------------------
void Terrain::Ground::getSlopeParameters(Interval& slope,
                                                float& slopeRelease) const
//------------------------------------------------------------------------------
{
    slope = mSlope;
    slopeRelease = mSlopeRelease;
}

//------------------------------------------------------------------------------
const UInt8* Terrain::Ground::get(unsigned int x, unsigned int y) const
//------------------------------------------------------------------------------
{
    y = y%mImage->getHeight();
    x = x%mImage->getHeight();
    return mImage->getData() + (x+y*mImage->getWidth())*(3);
}

//------------------------------------------------------------------------------
// Terrain members
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
Terrain::Terrain(HeightMap hm) :
//------------------------------------------------------------------------------
mHeightMap(hm),
mDeltaMap(0, 0, 0, 0),
mRule(DEFAULT),
mXZStep(1)
{
    mTextureRect.mLeft = 0;
    mTextureRect.mTop = 0;
    mTextureRect.mRight = hm.getXSize();
    mTextureRect.mBottom = hm.getYSize();
    mYRange.set(0, 10);
}

//------------------------------------------------------------------------------
Terrain::Terrain(HeightMap hm, DeltaXzMap dm) :
//------------------------------------------------------------------------------
mHeightMap(hm),
mDeltaMap(dm),
mRule(DEFAULT),
mXZStep(1)
{
    mTextureRect.mLeft = 0;
    mTextureRect.mTop = 0;
    mTextureRect.mRight = hm.getXSize();
    mTextureRect.mBottom = hm.getYSize();
    mYRange.set(0, 10);
}

//------------------------------------------------------------------------------
void Terrain::createMesh(bool allocateMemory,
    unsigned int& vertexCount, Vector3*& vertices,
    Vector3** normals, Vector2** uvs,
    unsigned int* indexCount, UInt16** indices) const
//------------------------------------------------------------------------------
{
    HeightMapConverter hm(allocateMemory);
    hm.createMesh(
        *this, vertexCount,
        vertices, normals, uvs,
        indexCount, indices);
}

//------------------------------------------------------------------------------
void Terrain::set(Triangulation rule)
//------------------------------------------------------------------------------
{
    mRule = rule;
}
    
//------------------------------------------------------------------------------
void Terrain::setYRange(const Interval& range)
//------------------------------------------------------------------------------
{
    mYRange = range;
}

//------------------------------------------------------------------------------
void Terrain::setXZStep(float step)
//------------------------------------------------------------------------------
{
    mXZStep = step;
}

//------------------------------------------------------------------------------
const Interval& Terrain::getYRange() const
//------------------------------------------------------------------------------
{
    return mYRange;
}

//------------------------------------------------------------------------------
float Terrain::getXZStep() const
//------------------------------------------------------------------------------
{
    return mXZStep;
}

//------------------------------------------------------------------------------
Terrain::Triangulation Terrain::getRule() const
//------------------------------------------------------------------------------
{
    return mRule;
}

//------------------------------------------------------------------------------
void Terrain::add(Ground& g)
//------------------------------------------------------------------------------
{
    mGrounds.push_back(g);
}
    
//------------------------------------------------------------------------------
const Terrain::Grounds& Terrain::getGrounds() const
//------------------------------------------------------------------------------
{
    return mGrounds;
}

//------------------------------------------------------------------------------
void Terrain::setTextureRect(const Rect<unsigned int>& rect)
//------------------------------------------------------------------------------
{
    mTextureRect = rect;
}

//------------------------------------------------------------------------------
const Rect<unsigned int>& Terrain::getTextureRect() const
//------------------------------------------------------------------------------
{
    return mTextureRect;
}

//------------------------------------------------------------------------------
void Terrain::createTexture(Image& im) const
//------------------------------------------------------------------------------
{
    TextureBlender tb;
    tb.createTexture(*this, im);
}

//------------------------------------------------------------------------------
const HeightMap& Terrain::getMap() const
//------------------------------------------------------------------------------
{
    return mHeightMap;
}

//------------------------------------------------------------------------------
const DeltaXzMap& Terrain::getDeltaXzMap() const
//------------------------------------------------------------------------------
{
    return mDeltaMap;
}


} // namespace Ge