//------------------------------------------------------------------------------
// (c) 01-2002 Gottfried Chen
//------------------------------------------------------------------------------

#ifndef TERRAIN_H
#define TERRAIN_H

#include <System/Types.h>
#include <System/Pointer.h>
#include <Image/Image.h>
#include <Math/Interval.h>
#include <Math/Vector.h>

#include <vector>

#include "HeightMap.h"
#include "DeltaXzMap.h"

namespace Ge
{


// A Terrain contains all the information for one terrain. This is
// - a heightmap
// - the y-range and x,z steps for mesh creation
// - the base images for texture generation
// A Terrain uses the HeightMapConverter for mesh creation and the
// TextureBlender for texture creation.
class Terrain
{
public:
    // Several triangulation rules
    enum Triangulation
    {
        MIN_HEIGHT_VARIATION,
        MAX_HEIGHT_VARIATION,
        TOP_TO_BOTTOM,
        BOTTOM_TO_TOP,
        DEFAULT = MIN_HEIGHT_VARIATION
    };

    // A ground is a texture with application rules
    class Ground
    {
    public:
        Ground(const char* name,
               const Interval& height, float heightRelease,
               const Interval& slope, float slopeRelease);

    private:
        friend class TextureBlender;

        void getHeightParameters(Interval& height,
                                 float& heightRelease) const;

        // Attention: The slope is not the slope in the final mesh, because the
        // actual slope there depends on the y-scaling factor you provide in the
        // HeightMapConverter. Here we assums a scale of 1 in x,y,z.
        void getSlopeParameters(Interval& slope,
                                float& slopeRelease) const;
        
        // Get pixel at x, y. The texture wraps if the x, y boundaries are exceeded.
        const UInt8* get(unsigned int x, unsigned int y) const;
        
        // Filename of the texture
        Pointer<Image> mImage;
        
        Interval mHeight;
        float mHeightRelease;

        // Slope is in [0, PI/2]
        Interval mSlope;
        float mSlopeRelease;
    };

    typedef std::vector<Ground> Grounds;

    //------------------------------------------------------------------------------
    // Map
    //------------------------------------------------------------------------------
    explicit
    Terrain(HeightMap);
    Terrain(HeightMap, DeltaXzMap);

    const HeightMap& getMap() const;
    const DeltaXzMap& getDeltaXzMap() const;

    //------------------------------------------------------------------------------
    // Mesh
    //------------------------------------------------------------------------------
    // Memory for vertex data is allocates if <allocateMemory> is true.
    // The user takes ownership of the memory allocated for vertex data and <indices>
    // arrays. <indices>, <normal> and <uvs> are calculated if != 0.
    void createMesh(bool allocateMemory,
                 unsigned int& vertexCount, Vector3*& vertices,
                 Vector3** normals, Vector2** uvs,
                 unsigned int* indexCount, UInt16** indices) const;

    void set(Triangulation);
    Triangulation getRule() const;

    // [0, 255] is scaled to this range. Default is [0, 10]
    void setYRange(const Interval&);
    const Interval& getYRange() const;

    // A step of one in x/y direction on the height map is mapped onto
    // the step size in x/-z direction. (-z because we use a left hand system).
    // Default = 1.
    void setXZStep(float);
    float getXZStep() const;


    //------------------------------------------------------------------------------
    // Texture
    //------------------------------------------------------------------------------
    // Add a ground to the collection.
    void add(Ground&);
    const Grounds& getGrounds() const;
    // Only this rectangle of the height map will be taken into account for
    // generating the texture.
    void setTextureRect(const Rect<unsigned int>& range);
    const Rect<unsigned int>& getTextureRect() const;
    // The texture for the given height map/range is created.
    void createTexture(Image&) const;

private:
    Grounds mGrounds;
    Rect<unsigned int> mTextureRect;

    Triangulation mRule;
    Interval mYRange;
    float mXZStep;

    HeightMap mHeightMap;
    DeltaXzMap mDeltaMap;
};

} // namespace Ge

#endif // TERRAIN_H