#include "Precompiled.h"
#include "TextureBlender.h"

#include <Math/Tools.h>
#include <Math/Vector.h>
#include <Math/Interpolation.h>

#include <Engine/Color.h>
#include <System/ByteOrder.h>
#include <System/Timer.h>

#include "math.h"

namespace Ge
{




//------------------------------------------------------------------------------
// TextureBlender members
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
TextureBlender::TextureBlender() :
//------------------------------------------------------------------------------
mHeightPerTexel(0),
mSlopePerTexel(0)
{
}

//------------------------------------------------------------------------------
void TextureBlender::createTexture(const Terrain& terrain, Image& im) const
//------------------------------------------------------------------------------
{
    GE_ASSERT_MSG(im.getBitDepth() == 32, "Only 32 bit terrain textures supported");

    mHeightPerTexel = GE_NEW float[im.getHeight()*im.getWidth()];
    calculateHeightPerTexelMap(terrain, im);

    mSlopePerTexel = GE_NEW float[im.getHeight()*im.getWidth()];
    calculateSlopePerTexelMap(terrain, im);
    
    // For each texel evaluate all grounds and calculate the final color.
    Vector4 rgba(0,0,0,0);
    float factorSum(0);
    unsigned int yline;
    for (unsigned int y = 0; y<im.getHeight(); ++y)
    {
        yline = y*im.getWidth();
        for (unsigned int x = 0; x<im.getWidth(); ++x)
        {
            // Reset values for this texel.
            set(rgba, 0.0f,0.0f,0.0f,0.0f);
            factorSum = 0;

            for (Terrain::Grounds::const_iterator g = terrain.getGrounds().begin();
                 g!=terrain.getGrounds().end(); ++g)
            {
                float factor = getHeightFactorForTexel(g, yline+x);

                if (factor != 0)
                    factor *= getSlopeFactorForTexel(g, yline+x);
             
                // Add to destination color
                if (factor != 0)
                {
                    const UInt8* pixel = g->get(x, y);
                    rgba[0] += factor*float(*pixel)/255.0f;
                    ++pixel;
                    rgba[1] += factor*float(*pixel)/255.0f;
                    ++pixel;
                    rgba[2] += factor*float(*pixel)/255.0f;
                    factorSum += factor;
                }

            }
            
            // Normalize the destination color
            if (factorSum != 0)
            {
                rgba *= 1.0f/factorSum;
            }

            rgba[3] = 1;
            switch (im.getFormat())
            {
            case Image::RGBA_8888:
                *(reinterpret_cast<UInt32*>(im.getData()) + yline + x) =
                    Color(Color::RGBA, rgba).get(Color::RGBA);
                break;

            case Image::BGRA_8888:
                *(reinterpret_cast<UInt32*>(im.getData()) + yline + x) =
                    Color(Color::RGBA, rgba).get(Color::BGRA);
                break;

            default:
                GE_ASSERT_MSG(false, "Unknown picture format");
            }

            theByteOrder.to(*(reinterpret_cast<UInt32*>(im.getData()) + yline + x), ByteOrder::BIG_ENDIAN);
        }
    }

    GE_DELETE_A mHeightPerTexel;
    GE_DELETE_A mSlopePerTexel;
}

//------------------------------------------------------------------------------
float TextureBlender::getSlopeFactorForTexel(Terrain::Grounds::const_iterator g,
                                             unsigned int offset) const
//------------------------------------------------------------------------------
{
    float factor;

    Interval s(0,0);
    float srel;
    g->getSlopeParameters(s, srel);

    float sdelta = s.getDistance(mSlopePerTexel[offset]);
    if (sdelta == 0)
        factor = 1;    
    else if (sdelta >= srel)
        factor = 0;
    else
        factor = (srel - sdelta)/srel;

    return factor;
}

//------------------------------------------------------------------------------
float TextureBlender::getHeightFactorForTexel(Terrain::Grounds::const_iterator g,
                                              unsigned int offset) const
//------------------------------------------------------------------------------
{
    float factor;
 
    Interval h(0,0);
    float hrel;
    g->getHeightParameters(h, hrel);
    
    float distance = h.getDistance(mHeightPerTexel[offset]);
    if (distance == 0)
        factor = 1;
    else if (distance >= hrel)
        factor = 0;
    else
        factor = (hrel - distance)/hrel;

    return factor;
}


//------------------------------------------------------------------------------
void TextureBlender::calculateHeightPerTexelMap(const Terrain& terrain, const Image& im) const
//------------------------------------------------------------------------------
{
    unsigned int yline;
    for (unsigned int y = 0; y<im.getHeight(); ++y)
    {
        yline = y*im.getWidth();
        for (unsigned int x = 0; x<im.getWidth(); ++x)
        {
            mHeightPerTexel[yline+x] = calculateHeightForTexel(terrain, im, x, y);
        }
    }
}

//------------------------------------------------------------------------------
float TextureBlender::calculateHeightForTexel(const Terrain& terrain, const Image& im,
                                              unsigned int x, unsigned int y) const
//------------------------------------------------------------------------------
{
    float realX = lerp(0, im.getWidth(), terrain.getTextureRect().mLeft, terrain.getTextureRect().mRight, x);
    float realY = lerp(0, im.getHeight(), terrain.getTextureRect().mTop, terrain.getTextureRect().mBottom, y);
    unsigned int realXInt(realX);
    unsigned int realYInt(realY);

    if (realXInt+1 == terrain.getMap().getXSize())
    {
        --realXInt;
        realX -= 1;
    }

    if (realYInt+1 == terrain.getMap().getYSize())
    {
        --realYInt;
        realY -= 1;
    }

    return lerp(realXInt, realYInt, realXInt+1, realYInt+1,
        terrain.getMap().get(realXInt, realYInt), terrain.getMap().get(realXInt+1, realYInt),
        terrain.getMap().get(realXInt+1, realYInt+1), terrain.getMap().get(realXInt, realYInt+1),
        realX, realY);
    
    /* Weighted sum instead of lerp (should be the same actually)
    float ret = terrain.getMap().get(realXInt, realYInt)*(1-(realX-realXInt))*(1-(realY-realYInt))+
           terrain.getMap().get(realXInt+1, realYInt)*(realX-realXInt)*(1-(realY-realYInt))+
           terrain.getMap().get(realXInt+1, realYInt+1)*(realX-realXInt)*(realY-realYInt)+
           terrain.getMap().get(realXInt, realYInt+1)*(1-(realX-realXInt))*(realY-realYInt);

    GE_ASSERT(ret>=0 && ret<=255);
    return ret;
    */
}




//------------------------------------------------------------------------------
void TextureBlender::calculateSlopePerTexelMap(const Terrain& terrain, const Image& im) const
//------------------------------------------------------------------------------
{
    Timer timer;

    // Need this for atan() below
    float xStep = lerp(0, im.getWidth(), terrain.getTextureRect().mLeft, terrain.getTextureRect().mRight, 1);
    float yStep = lerp(0, im.getHeight(), terrain.getTextureRect().mTop, terrain.getTextureRect().mBottom, 1);
    float heightScale = lerp(0, 255, terrain.getYRange().getLeft(), terrain.getYRange().getRight(), 1);

//    // Calculate per texel slope map (i.e. like normal map, but just remember the
//    // slope). Approximate slope by taking the max of the slopes in x, y direction.
//    unsigned int yline;
//    float samples[2];
//    for (unsigned int y = 1; y<im.getHeight()-1; ++y)
//    {
//        yline = y*im.getWidth();
//        for (unsigned int x = 1; x<im.getWidth()-1; ++x)
//        {
//            // Calculate slope in x 
//            samples[0] = heightScale*fabs(mHeightPerTexel[yline+x+1]-mHeightPerTexel[yline+x-1])/(2*xStep);
//            // Calculate slope in y
//            samples[1] = heightScale*fabs(mHeightPerTexel[yline+x+im.getWidth()]-mHeightPerTexel[yline+x-im.getWidth()])/(2*yStep);
//            
//            mSlopePerTexel[yline+x] = atan(maximum(samples[0], samples[1]));
//            GE_ASSERT(0<=mSlopePerTexel[yline+x] &&
//                      mSlopePerTexel[yline+x]<=PI_HALF);
//            
//
//            
//        }
//    }
//
    
    // Calculate per texel normal. This is slower but more accurate.
    unsigned int yline;
    Vector3 normal;
    for (unsigned int y = 0; y<im.getHeight(); ++y)
    {
        yline = y*im.getWidth();
        for (unsigned int x = 0; x<im.getWidth(); ++x)
        {
            float sx = (mHeightPerTexel[yline + (x<im.getWidth()-1 ? x+1 : x)] -
                        mHeightPerTexel[yline + (x>0 ? x-1 : x)])*heightScale;
            if (x == 0 || x == im.getWidth()-1)
                sx *= 2;

            float sy = (mHeightPerTexel[(y<im.getHeight()-1 ? yline+im.getWidth() : yline) + x] -
                        mHeightPerTexel[(y>0 ?  yline-im.getWidth() : yline) + x])*heightScale;
            if (y == 0 || y == im.getHeight()-1)
                sy *= 2;

            set(normal, -sx, 2*xStep, -xStep*sy/yStep);
            normal.normalize();   
            mSlopePerTexel[yline+x] = acos(fabs(normal[1]));
        }
    }

    unsigned int time = timer.get();
}

} // namespace Ge