#include "Precompiled.h"
#include "Water2D.h"
#include <System/Base.h>
#include <string.h>

namespace Ge
{

// Depth of displacement used by drop()
const Float Water2D::DROP_DEPTH = 3.0f;

// Dampening term of the water surface (must be in [0,1], where 0 is
// instant dampening and 1 is none.
const Float Water2D::DAMPENING = 0.99f;

// MIN_DELTA is the distance of two grid points (i.e. <h> in Gomez' article
// in Game Programming Gems).
const Float Water2D::MIN_DELTA = 0.2f;

//------------------------------------------------------------------------------
Water2D::Water2D(UInt xSize, UInt ySize, Float speed) :
//------------------------------------------------------------------------------
mXSize(xSize),
mYSize(ySize),
mCurrent(0)
{
    UInt size = xSize*ySize;

    mArray[0] = GE_NEW Float[size];
    mArray[1] = GE_NEW Float[size];

    memset(mArray[0], 0, size*sizeof(Float));
    memset(mArray[1], 0, size*sizeof(Float));

    // <speed> is <c> in Gomez' article.
    mSpeedFactor = speed*speed*MIN_DELTA*MIN_DELTA;
}

//------------------------------------------------------------------------------
Water2D::~Water2D()
//------------------------------------------------------------------------------
{
    for (UInt i = 0; i<2; ++i)
    {
        if (mArray[i] != 0)
        {
            GE_DELETE_A mArray[i];
        }
    }
}

//------------------------------------------------------------------------------
void Water2D::drop(UInt x, UInt y)
//------------------------------------------------------------------------------
{
    GE_ASSERT(x>=1 && y>=1);
    GE_ASSERT(x<=mXSize-2 && y<=mYSize-2);
    
    // Well, that should be approximately a circle.
    access(mCurrent, x, y) = DROP_DEPTH;
    
    if (x>=2)
    {
        access(mCurrent, x-1, y) = DROP_DEPTH*0.5f;
    }
    
    if (x<=mXSize-3)
    {
        access(mCurrent, x+1, y) = DROP_DEPTH*0.5f;
    }
    
    if (y>=2)
    {
        access(mCurrent, x, y-1) = DROP_DEPTH*0.5f;
    }

    if (y<=mYSize-3)
    {
        access(mCurrent, x, y+1) = DROP_DEPTH*0.5f;
    }
}


//------------------------------------------------------------------------------
void Water2D::update()
//------------------------------------------------------------------------------
{
    const Float A = mSpeedFactor;

    // See Gomez' article why this condition must be fulfilled.
    GE_ASSERT_MSG(A <= 0.5f, "Water becomes unstable.");
    
    const Float B = 2 - 4*A;

    // Swap buffers
    UInt previous = mCurrent;
    mCurrent = mCurrent == 0 ? 1 : 0;

    // Recalculate surface
    for (UInt i = 1; i<mXSize-1; ++i)
    {
        for (UInt j = 1; j<mYSize-1; ++j)
        {
            access(mCurrent, i, j) = (access(previous, i+1, j) + access(previous, i-1, j) +
                                      access(previous, i, j-1) + access(previous, i, j+1))*A +
                                      B*access(previous, i, j) -
                                      access(mCurrent, i, j);

            access(mCurrent, i, j) *= DAMPENING;
        }
    }
}

//------------------------------------------------------------------------------
const Float* Water2D::getMap() const
//------------------------------------------------------------------------------
{
    return mArray[mCurrent];
}

} // namespace Ge