#ifndef WATER_2D_H
#define WATER_2D_H

#include <System/Types.h>

namespace Ge
{

// This implements the 2D water used in some old demos. It's faster than the Newave water,
// but doesn't look as good. (For an explanation of the algorithm see Game Programming Gems).
class Water2D
{
public:
    static const Float DROP_DEPTH;
    static const Float DAMPENING;
    static const Float MIN_DELTA;

    // - <xSize>, <ySize> are the size of the water pool
    // - <speed> is the speed of wave propagation
    Water2D(UInt xSize, UInt ySize, Float speed = 1.0f);

    ~Water2D();

    // Release a drop of water into the simulation.
    // - <x> must be in [1, <xSize>-2]
    // - <y> must be in [1, <ySize>-2]
    void drop(UInt x, UInt y);

    // I tried to make update() time dependent. But this didn't work out because
    // this water method relies on fixed time intervals. So introducing time as
    // a parameter made the solution unstable. The only method that might work is
    // assuming a fixed frame time t. If a frame takes longer than t then just
    // call update as many times as needed to catch up (e.g. if a frame took
    // 3*t then call update 3 times). This won't work if the water itself is the
    // bottleneck though because then the update call multiplier will increase
    // to infinity.
    // The reason why time dependence didn't work is because this is an explicit
    // partial differential equation solver that relies on constant time slices.
    // (see Gomez' article in "Game Programming Gems").
    void update();

    const Float* getMap() const;
    UInt getXSize() const { return mXSize; }
    UInt getYSize() const { return mYSize; }

private:
    Float& access(UInt index, UInt x, UInt y);

    Float* mArray[2];
    UInt mCurrent;
    UInt mXSize;
    UInt mYSize;
    Float mSpeedFactor;
};

inline
//------------------------------------------------------------------------------
Float& Water2D::access(UInt index, UInt x, UInt y)
//------------------------------------------------------------------------------
{
    return mArray[index][x + y*mXSize];
}

} // namespace Ge

#endif // WATER_2D_H

