#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Sep-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "BumpContext.h"

#include <Application/Settings.h>

#include <Engine/Quad.h>
#include <Engine/Mesh.h>
#include <Engine/Link.h>
#include <Engine/Renderer.h>
#include <Engine/TexturePool.h>
#include <Engine/MorphFrame.h>
#include <Engine/BumpMaterial.h>
#include <Engine/RenderCache.h>

#include <Image/Image.h>

#include <System/FilePath.h>
#include <System/FileSystem.h>


namespace
{

using namespace Ge;

const UInt flipVertexCount = 9;
const UInt flipIndexCount = 8*3;

Vector3 flipVertices[] =
{
    Vector3(-2,2,0),
    Vector3(0,2,0),
    Vector3(2,2,0),

    Vector3(-2,0,0),
    Vector3(0,0,0),
    Vector3(2,0,0),

    Vector3(-2,-2,0),
    Vector3(0,-2,0),
    Vector3(2,-2,0)
};

Vector2 flipUvs[] =
{
    // Mirrored
    Vector2(0,0),
    Vector2(1,0),
    Vector2(0,0),

    Vector2(0,1),
    Vector2(1,1),
    Vector2(0,1),

    Vector2(0,0),
    Vector2(1,0),
    Vector2(0,0)

//    // Stretched
//    Vector2(0,0),
//    Vector2(0.5f,0),
//    Vector2(1,0),
//    Vector2(1,1),
//    Vector2(0.5f,1),
//    Vector2(0,1)
};

Vector3 flipNormals[] =
{
    Vector3(0,0,-1),
    Vector3(0,0,-1),
    Vector3(0,0,-1),

    Vector3(0,0,-1),
    Vector3(0,0,-1),
    Vector3(0,0,-1),

    Vector3(0,0,-1),
    Vector3(0,0,-1),
    Vector3(0,0,-1)
};

UInt16 flipIndices[] =
{
    0,4,3,
    0,1,4,
    1,5,4,
    1,2,5,
    3,7,6,
    3,4,7,
    4,8,7,
    4,5,8
};

//------------------------------------------------------------------------------
Mesh* createFlipMesh()
//------------------------------------------------------------------------------
{
    Mesh* ret = 0;
    
    Vector3* vertices = GE_NEW Vector3[flipVertexCount];
    memcpy(vertices, flipVertices, sizeof(Vector3)*flipVertexCount);
    
    Vector3* normals = GE_NEW Vector3[flipVertexCount];
    memcpy(normals, flipNormals, sizeof(Vector3)*flipVertexCount);
    
    Vector2* uvs0 = GE_NEW Vector2[flipVertexCount];
    memcpy(uvs0, flipUvs, sizeof(Vector2)*flipVertexCount);
    Vector2* uvs1 = GE_NEW Vector2[flipVertexCount];
    memcpy(uvs1, flipUvs, sizeof(Vector2)*flipVertexCount);
    
    UInt16* indices = GE_NEW UInt16[flipIndexCount];
    memcpy(indices, flipIndices, sizeof(UInt16)*flipIndexCount);

    ret = GE_NEW Mesh(1, flipVertexCount, vertices, normals, uvs0, 0,
                      flipIndexCount, indices);
    
    // Dynamic because colors are updated every frame.
    ret->setDynamic(true);

    ret->setUvs(1, Mesh::UV_ARRAY, uvs1);

    return ret;
}

} // namespace


namespace Ge
{

//------------------------------------------------------------------------------
BumpContext::BumpContext() :
//------------------------------------------------------------------------------
Context("BumpContext"),
mCamera(*Locator::create(Matrix3x4::IDENTITY())),
mCameraController(mCamera),
mLink(0),
mLinkClone(0),
mLinkFlip(0)
{
}

//------------------------------------------------------------------------------
void BumpContext::init()
//------------------------------------------------------------------------------
{
    // Position camera & light
    Vector3 cameraPos(0, 1, -5);
    mLight = cameraPos;
    Matrix3x4 lookAt;
    makeLookAt(lookAt, cameraPos, Vector3(0,0,0), Vector3(0,1,0));
    mCamera.getLocator().get() = lookAt;
    mCamera.getLocator().update();
    mCameraController.setDistance(cameraPos.length());
    mCameraController.recalculate();
    mWorld.setCamera(&mCamera);

    // Create normal map
    FilePath path(theFileSystem().getApplicationPath());
    path += theSettings.getVariable("data").asString();
    path += "/textures/debugHeight.tga";
    Image height(path.get().c_str());
    path.setFile("debugnormal.tga");
    Image* normal = height.createNormalMap(30);
    normal->save(Image::TGA, path.get().c_str());
    GE_DELETE normal;

    // Create a quad
    Link* quad = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()),
                             *Quad::create(2, 2));
    quad->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(-1.2f,-1,0));
    quad->getLocator()->update();
    quad->calculateBound();
    BumpMaterial* quadMaterial = GE_NEW BumpMaterial;
    quadMaterial->set(0, *(theTexturePool().get(path.get().c_str())));
    path.setFile("h_bitterman.tga");
    quadMaterial->set(1, *(theTexturePool().get(path.get().c_str())));
    quad->set(*quadMaterial);
    quadMaterial->setLink(*quad);
    
    Vector2* uv1 = GE_NEW Vector2[4];
    uv1[0] = Vector2(0,1);
    uv1[1] = Vector2(0,0);
    uv1[2] = Vector2(1,1);
    uv1[3] = Vector2(1,0);
    // Needs to be dynamic, because colors get updated every frame.
    quad->getMesh()->setDynamic(true);
    quad->getMesh()->setUvs(1, Mesh::UV_ARRAY, uv1);
    quad->getMesh()->calculateTangents();

    mLink = quad;

    // Create a clone of the quad
    mLinkClone = quad->clone();
    mLinkClone->set(*Locator::create(Matrix3x4::IDENTITY()));
    makeZRotation(mLinkClone->getLocator()->get(), PI);
    mLinkClone->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(1.2f,-1,0));
    mLinkClone->getLocator()->update();
    mLinkClone->calculateBound();
    ((BumpMaterial*)mLinkClone->getMaterial())->setLink(*mLinkClone);

    // Create a quad with texture mirrored in the middle
    mLinkFlip = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()),
                            *createFlipMesh());
    mLinkFlip->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(0,2.5f,0));
    mLinkFlip->getLocator()->update();
    mLinkFlip->calculateBound();
    mLinkFlip->set(*quadMaterial->clone());
    mLinkFlip->getMesh()->calculateTangents();
    ((BumpMaterial*)mLinkFlip->getMaterial())->setLink(*mLinkFlip);
    GE_TRACE("Before: "<<flipVertexCount<<", After: "<<
             mLinkFlip->getMesh()->getVertexCount()<<std::endl);

    mWorld.add(*mLink);
    mWorld.add(*mLinkClone);
    mWorld.add(*mLinkFlip);
}

//------------------------------------------------------------------------------
void BumpContext::update()
//------------------------------------------------------------------------------
{
    // Rotate camera and light
    mCameraController.update();
    mCamera.getLocator().get().getColumn(MatrixBase::TRANSLATION, mLight);

    mLink->setLight(0, Vector4(mLight[0], mLight[1], mLight[2], 1));
    mLinkClone->setLight(0, Vector4(mLight[0], mLight[1], mLight[2], 1));
    mLinkFlip->setLight(0, Vector4(mLight[0], mLight[1], mLight[2], 1));
}

//------------------------------------------------------------------------------
void BumpContext::render()
//------------------------------------------------------------------------------
{
    theRenderer().render(mWorld);
}

//------------------------------------------------------------------------------
void BumpContext::exit()
//------------------------------------------------------------------------------
{
}

} // namespace Ge
