#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) 04-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "ChoppyWavesContext.h"

#include <Engine/Mesh.h>
#include <Engine/Link.h>
#include <Engine/RenderCache.h>
#include <Engine/TexturePool.h>

#include <Application/Metronome.h>
#include <Application/Settings.h>

#include <System/FilePath.h>
#include <System/FileSystem.h>

#include <Terrain/HeightMap.h>
#include <Terrain/Terrain.h>

namespace
{
    const Ge::Float CHOPPINESS_DELTA = 0.01f;
}


namespace Ge
{

//------------------------------------------------------------------------------
ChoppyWavesContext::ChoppyWavesContext() :
//------------------------------------------------------------------------------
WaterContextBase("ChoppyWavesContext", 8, 0.3f),
mChoppyWaves(-0.15f, 6, 6, 0.0008f, Vector2(30,10), 100.0f, 100.0f),
mDeltaMap(mChoppyWaves.getXSize(), mChoppyWaves.getYSize(),
          mChoppyWaves.getDeltaMapX(), mChoppyWaves.getDeltaMapZ()),
mMesh(0),
mVertices(0),
mNormals(0),
mDecChop(Keyboard::A),
mIncChop(Keyboard::D),
mRepDecChop(mDecChop),
mRepIncChop(mIncChop)
{
}

//------------------------------------------------------------------------------
void ChoppyWavesContext::init()
//------------------------------------------------------------------------------
{
    WaterContextBase::init();

    //------------------------------------------------------------------------------
    // Init water
    //------------------------------------------------------------------------------
    mChoppyWaves.update(theMetronome().getAbsolute()/1000.0f);

    //------------------------------------------------------------------------------
    // Generate mesh
    //------------------------------------------------------------------------------

    HeightMap hm(mChoppyWaves);
    Terrain terrain(hm, mDeltaMap);

    UInt vertexCount = 0;
    UInt indexCount = 0;
    terrain.createMesh(true,
                       vertexCount, mVertices,
                       &mNormals, 0,
                       &indexCount, &mIndices);

    //------------------------------------------------------------------------------
    // Add to scene
    //------------------------------------------------------------------------------
    mMesh = GE_NEW Mesh(1, vertexCount, mVertices, mNormals, 0, 0,
                        indexCount, mIndices);
    mMesh->setDynamic(true);
    mMesh->setUvs(0, Mesh::UV_SPHERE);

    Link* l = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()), *mMesh);

//    // Need this so that bounding sphere gets calculated
//    g->updateGeometry();

    // Set texture
    FilePath path(theFileSystem().getApplicationPath());
    path += theSettings.getVariable("data").asString();
    path += "/Textures/ReflectOcean.tga";
    Material* material = GE_NEW Material;
    material->set(0, *(theTexturePool().get(path.get().c_str())));
    l->set(*material);

    // ToDo: Need borders to loop, for tiling the water
//    Link* copy = l->clone();
//    Locator* copyLocator = Locator::create(Matrix3x4::IDENTITY());
//    copyLocator->get().setColumn(MatrixBase::TRANSLATION, Vector3(63,0,0));
//    copyLocator->update();
//    copy->set(*copyLocator);
//
//    Link* copy2 = l->clone();
//    copyLocator = Locator::create(Matrix3x4::IDENTITY());
//    copyLocator->get().setColumn(MatrixBase::TRANSLATION, Vector3(0,0,63));
//    copyLocator->update();
//    copy2->set(*copyLocator);
//
//    Link* copy3 = l->clone();
//    copyLocator = Locator::create(Matrix3x4::IDENTITY());
//    copyLocator->get().setColumn(MatrixBase::TRANSLATION, Vector3(63,0,63));
//    copyLocator->update();
//    copy3->set(*copyLocator);

    // Add geometry to world
    l->calculateBound();
    getWorld().add(*l);
//    getWorld().add(*copy);
//    getWorld().add(*copy2);
//    getWorld().add(*copy3);

    // Add light
    getWorld().add(getLight());
}
    
//------------------------------------------------------------------------------
void ChoppyWavesContext::update()
//------------------------------------------------------------------------------
{
    WaterContextBase::update();

    updateChoppiness();

    mChoppyWaves.update(theMetronome().getAbsolute()/1000.0f);

    HeightMap hm(mChoppyWaves);
    Terrain terrain(hm, mDeltaMap);

    terrain.set(Terrain::BOTTOM_TO_TOP);
    terrain.setYRange(Interval(-getWaterHeight(), getWaterHeight()));

    UInt vertexCount = 0;
    terrain.createMesh(false,
                       vertexCount, mVertices,
                       &mNormals, 0,
                       0, 0);

    if (mMesh->cache() != 0)
    {
        mMesh->cache()->update();
    }
}

//------------------------------------------------------------------------------
void ChoppyWavesContext::updateChoppiness()
//------------------------------------------------------------------------------
{
    mRepIncChop.update();
    mRepDecChop.update();
    Float lambda = mChoppyWaves.getLambda();
    if (mRepDecChop.isButtonDown())
    {
        lambda += CHOPPINESS_DELTA;
        lambda = minimum(0.0f, lambda);
    }
    else if (mRepIncChop.isButtonDown())
    {
        lambda -= CHOPPINESS_DELTA;
    }
    mChoppyWaves.setLambda(lambda);
}

//------------------------------------------------------------------------------
void ChoppyWavesContext::exit()
//------------------------------------------------------------------------------
{
    // Destructor of mWorld should clean everything up.
}

} // namespace Ge



    
