#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jun-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "LwsImportContext.h"

#include <Application/ScaledMetronome.h>
#include <Application/Settings.h>

#include <Engine/ModelPool.h>
#include <Engine/LwsImporter.h>
#include <Engine/CoordinateAxes.h>
#include <Engine/Lines.h>
#include <Engine/XzGrid.h>
#include <Engine/Renderer.h>

#include <System/Log.h>
#include <System/FileSystem.h>


namespace Ge
{

//------------------------------------------------------------------------------
LwsImportContext::LwsImportContext() :
//------------------------------------------------------------------------------
Context("LwsImportContext"),
mCamera(*Locator::create(Matrix3x4::IDENTITY())),
mCameraController(mCamera),
mModel(0),
mOpenFile(Keyboard::O),
mToggleAnim(Keyboard::SPACE),
mDecAnimSpeed(Keyboard::S),
mIncAnimSpeed(Keyboard::W),
mRepDecAnimSpeed(mDecAnimSpeed),
mRepIncAnimSpeed(mIncAnimSpeed),
mAnimation("NoName"),
mAnimSpeed(1),
mAnimOn(true),
mWorld(0)
{
}

//------------------------------------------------------------------------------
void LwsImportContext::init()
//------------------------------------------------------------------------------
{
    mLight = GE_NEW Light;

    FilePath path(theFileSystem().getApplicationPath());
    path += theSettings.getVariable("data").asString();
    path += "animations/sailjump/sails_jump.lws";
    openModel(path.get().c_str());
}

//------------------------------------------------------------------------------
void LwsImportContext::update()
//------------------------------------------------------------------------------
{
    // Check for open file
    mOpenFile.update();
    if (mOpenFile.isButtonDown())
    {
        openModel();
    }

    // Check animation keys
    mToggleAnim.update();
    if (mToggleAnim.isButtonDown())
    {
        toggleAnim();
    }

    mRepDecAnimSpeed.update();
    if (mRepDecAnimSpeed.isButtonDown())
    {
        changeAnimSpeed(1.0f/1.5f);
    }

    mRepIncAnimSpeed.update();
    if (mRepIncAnimSpeed.isButtonDown())
    {
        changeAnimSpeed(1.5f);
    }

    // Rotate camera and light
    mCameraController.update();
    Vector3 cameraDir;
    mCamera.getLocator().get().getColumn(MatrixBase::AT, cameraDir);
    Vector4 lightPos;
    set(lightPos, -cameraDir, 0.0f);
    mLight->setPosition(lightPos);

    // Update animation
    mAnimation.update(theScaledMetronome().getAbsolute());

    // After the local transformations in the Locators are updates,
    // propagate the changes to the global transformations
    if (mModel != 0)
    {
        mModel->getLocator()->update();
        mModel->calculateBound();
    }
}

//------------------------------------------------------------------------------
void LwsImportContext::render()
//------------------------------------------------------------------------------
{
    if (mWorld != 0)
    {
        theRenderer().render(*mWorld);
    }
}

//------------------------------------------------------------------------------
void LwsImportContext::exit()
//------------------------------------------------------------------------------
{
    if (mWorld != 0)
    {
        GE_DELETE mWorld;
    }
}

//------------------------------------------------------------------------------
void LwsImportContext::openModel(const char* file)
//------------------------------------------------------------------------------
{
    std::string path;
    
    bool ok = true;
    if (file == 0)
    {
        const char* filter = "Lightwave scenes (*.lws)\0*.lws\0\0";
        ok = openLoadDialog(path, filter);
    }
    else
    {
        path = file;
    }

    Cluster* model = 0;
    if (ok)
    {
        theModelPool().setFlipUvs(true);
        LwsImporter lws(path.c_str(), theModelPool());
        model = lws.importModel();
        ok = model != 0;
        if (ok)
        {
            mAnimation = lws.importAnimation();
        }
    }

    if (ok)
    {
        if (mWorld != 0)
        {
            GE_DELETE mWorld;
        }

        mWorld = GE_NEW World;
        mWorld->setCamera(&mCamera);
        mWorld->add(*mLight);

        mModel = model;
        mAnimation.attachTo(*mModel->getLocator());

        // Need this so that bounding sphere gets calculated
        mModel->getLocator()->update();
        mModel->calculateBound();
        Float boundSize(1.5*mModel->getBound().getRadius());

        // Position camera
        Vector3 cameraPos(0, boundSize/2, -boundSize);
        Matrix3x4 lookAt;
        makeLookAt(lookAt, cameraPos, Vector3(0,0,0), Vector3(0,1,0));
        mCamera.getLocator().get() = lookAt;
        mCamera.getLocator().update();
        mCameraController.setDistance(cameraPos.length());
        mCameraController.recalculate();
    
        // Create scene
//        Node* root = GE_NEW Node("Root");
//        Geometry* g = GE_NEW Geometry(*CoordinateAxes::create(2*boundSize));
//        g->setName("CoordinateAxes");
//        root->addChild(*g);
//        g = GE_NEW Geometry(*XzGrid::create(minimum(UInt(boundSize), UInt(20)), true));
//        g->setName("Grid");
//        root->addChild(*g);


        mWorld->add(*mModel);

//#ifndef NDEBUG
//        // Print scene tree for debugging
//        root->printSubTree(GE_LOG_DEBUG("Scene tree:"<<std::endl));
//#endif
    
        mAnimation.reset(theScaledMetronome().getAbsolute());
    }
}

//------------------------------------------------------------------------------
void LwsImportContext::toggleAnim()
//------------------------------------------------------------------------------
{
    theScaledMetronome().setRatio(mAnimOn ? 0 : mAnimSpeed);
    mAnimOn = !mAnimOn;
}

//------------------------------------------------------------------------------
void LwsImportContext::changeAnimSpeed(Float factor)
//------------------------------------------------------------------------------
{
    mAnimSpeed *= factor;
    theScaledMetronome().setRatio(mAnimSpeed);
}


} // namespace Ge
