#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Mai-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "NewaveContext.h"

#include "TestFramework.h"

#include <System/FilePath.h>
#include <System/FileSystem.h>

#include <Engine/Mesh.h>
#include <Engine/Link.h>
#include <Engine/RenderCache.h>
#include <Engine/TexturePool.h>

#include <Application/Metronome.h>
#include <Application/Settings.h>

#include <Terrain/HeightMap.h>
#include <Terrain/Terrain.h>

namespace Ge
{

//------------------------------------------------------------------------------
NewaveContext::NewaveContext() :
//------------------------------------------------------------------------------
WaterContextBase("NewaveContext", 8, 2),
mNewave(60, 0.01f),
mMesh(0),
mVertices(0),
mNormals(0),
mDrop(Keyboard::SPACE)
{
}

//------------------------------------------------------------------------------
void NewaveContext::init()
//------------------------------------------------------------------------------
{
    WaterContextBase::init();

    //------------------------------------------------------------------------------
    // Init pick
    //------------------------------------------------------------------------------
    mPick.setCamera(getCamera());
    mPick.setWindow(theFramework.getWindow());

    //------------------------------------------------------------------------------
    // Init water
    //------------------------------------------------------------------------------
    mNewave.reset(Newave::CORNERBLOCK);
    mNewave.update();

    //------------------------------------------------------------------------------
    // Generate mesh
    //------------------------------------------------------------------------------

    HeightMap hm(mNewave);
    Terrain terrain(hm);

    UInt vertexCount = 0;
    UInt indexCount = 0;
    terrain.createMesh(true,
                       vertexCount, mVertices,
                       &mNormals, 0,
                       &indexCount, &mIndices);

    //------------------------------------------------------------------------------
    // Add to scene
    //------------------------------------------------------------------------------
    mMesh = GE_NEW Mesh(1, vertexCount, mVertices, mNormals, 0, 0,
                        indexCount, mIndices);
    mMesh->setDynamic(true);
    mMesh->setUvs(0, Mesh::UV_SPHERE);

    Link* l = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()), *mMesh);

//    // Need this so that bounding sphere gets calculated
//    g->updateGeometry();
    
    // Set texture
    FilePath path(theFileSystem().getApplicationPath());
    path += theSettings.getVariable("data").asString();
    path += "/Textures/Reflect.tga";
    Material* material = GE_NEW Material;
    material->set(0, *(theTexturePool().get(path.get().c_str())));
    l->set(*material);

    // Add geometry to world
    l->calculateBound();
    getWorld().add(*l);
    
    // Add light
    getWorld().add(getLight());
}
    
//------------------------------------------------------------------------------
void NewaveContext::update()
//------------------------------------------------------------------------------
{
    WaterContextBase::update();

    updateDrop();

    // Fixed update frequency of 100hz, clamped down to avoid escalation
    UInt updateCount = minimum(Int(theMetronome().getDelta()/10), 4);
    for (UInt i = 0; i<updateCount; ++i)
    {
        mNewave.update();
    }

    HeightMap hm(mNewave);
    Terrain terrain(hm);

    terrain.set(Terrain::BOTTOM_TO_TOP);
    terrain.setYRange(Interval(-getWaterHeight(), getWaterHeight()));

    UInt vertexCount = 0;
    terrain.createMesh(false,
                       vertexCount, mVertices,
                       &mNormals, 0,
                       0, 0);

    if (mMesh->cache() != 0)
    {
        mMesh->cache()->update();
    }
}

//------------------------------------------------------------------------------
void NewaveContext::updateDrop()
//------------------------------------------------------------------------------
{
    mDrop.update();

    if (mDrop.isPressed())
    {
        mPick.update();

        Plane xz(Vector3(0,1,0), 0);
        Vector3 point;

        Vector3 camPos;
        getCamera().getLocator().get().getColumn(MatrixBase::TRANSLATION, camPos);
        if (intersects(xz, mPick.getMouseRay(), &point) == Intersect::INTERSECTS &&
            camPos[1] > 0)
        {
            point[2] = -point[2];
            point += Vector3(mNewave.getSize()/2.0f, 0, mNewave.getSize()/2.0f);
            
            clamp(point[0], 0.0f, mNewave.getSize()-1.0f);            
            clamp(point[2], 0.0f, mNewave.getSize()-1.0f);

            Vector2 delta;
            theMouse().getDelta(delta);
            if ((mDrop.isButtonDown() || delta != Vector2::ORIGIN()) &&
                1<=point[0] && point[0]<=mNewave.getSize()-2 &&
                1<=point[2] && point[2]<=mNewave.getSize()-2)
            {
                mNewave.drop(point[0], point[2]);
            }
        }
    }
}

//------------------------------------------------------------------------------
void NewaveContext::exit()
//------------------------------------------------------------------------------
{
    // Destructor of mWorld should clean everything up.
}

} // namespace Ge

