#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jul-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "ShadowVolumeContext.h"

#include <Engine/Mesh.h>
#include <Engine/Box.h>
#include <Engine/Renderer.h>
#include <Engine/RenderCache.h>
#include <Engine/Locator.h>
#include <Engine/Link.h>
#include <Engine/ModelPool.h>
#include <Engine/TexturePool.h>
#include <Engine/VideoCard.h>

#include <Application/Metronome.h>
#include <Application/Settings.h>

#include <System/FilePath.h>
#include <System/FileSystem.h>

namespace
{
    using namespace Ge;

    //------------------------------------------------------------------------------
    void initLightParticles(Emitter& e)
    //------------------------------------------------------------------------------
    {
        FilePath path(theFileSystem().getApplicationPath());
        path += theSettings.getVariable("data").asString();
        path += "/Textures/fire.tga";
        Material* material = GE_NEW Material;
        material->set(0, *(theTexturePool().get(path.get().c_str())));
        e.set(*material);
        e.setSpeed(Vector3(0,0.5f,0));
        e.setSpread(degToRad(10));
        e.set(Color(1,1,0,1), Color(1,0,0,0));
        e.setSize(1.0f, 0.2f);
    }

    Pointer<Link> linkBox1;
    Pointer<Link> linkBox2;
    Pointer<Cluster> xMesh;
}


namespace Ge
{

//------------------------------------------------------------------------------
ShadowVolumeContext::ShadowVolumeContext() :
//------------------------------------------------------------------------------
Context("ShadowVolumeContext"),
mCamera(*Locator::create(Matrix3x4::IDENTITY())),
mCameraController(mCamera),
mLightParticles(30, 0.5f)
{
}

//------------------------------------------------------------------------------
void ShadowVolumeContext::init()
//------------------------------------------------------------------------------
{
    if (!theVideoCard().supports(VideoCard::DS_24_8))
    {
        return;
    }

    mCamera.getLocator().get().setColumn(MatrixBase::TRANSLATION, Vector3(0,0,-5));
    mCamera.getLocator().update();
    mCameraController.setDistance(5);
    mCameraController.recalculate();
    mCamera.setNearPlane(0.3f);
    mCamera.setFarPlane(100000);

    mWorld.setCamera(&mCamera);
    mLight = GE_NEW Light;
    mLight->setRange(60);
    
    initLightParticles(mLightParticles);

    // Create room
    Mesh* mesh = Box::create(10,5,10);
    mesh->invert();
    Locator* locRoom = Locator::create(Matrix3x4::IDENTITY());
    Link* linkRoom = GE_NEW Link(*locRoom, *mesh);
    
    // Create some boxes
    mesh = Box::create(0.5, 0.5, 0.5);
    mesh->calculateNormals(true, true);
    mesh->calculateFacePlanes();
    mesh->calculateConnectivity();

    Locator* locBox1 = GE_NEW Locator;
    locBox1->get() = Matrix3x4::IDENTITY();
    locBox1->get().setColumn(MatrixBase::TRANSLATION, Vector3(-2,-1,0));
    locBox1->update();
    linkBox1 = GE_NEW Link(*locBox1, *mesh);
    linkBox1->setShadowCaster(true);

    Locator* locBox2 = GE_NEW Locator;
    locBox2->get() = Matrix3x4::IDENTITY();
    locBox2->get().setColumn(MatrixBase::TRANSLATION, Vector3(2,1,0));
    locBox2->update();
    linkBox2 = linkBox1->clone();
    linkBox2->set(*locBox2);
    
//    // Load .obj
//    FilePath oPath(theFileSystem().getApplicationPath());
//    oPath += theSettings.getVariable("data").asString();
//    oPath += "animations/testrobot/test.obj";
//    Cluster* oMesh = theModelPool().get(oPath.get().c_str());
//    oMesh->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(-3,-2,0));
//    oMesh->getLocator()->update();
//    //oMesh->setShadowCaster(true);

    // Load .x
    FilePath xPath(theFileSystem().getApplicationPath());
    xPath += theSettings.getVariable("data").asString();
    xPath += "xfiles/tiger.x";
    xMesh = theModelPool().get(xPath.get().c_str());
    xMesh->setShadowCaster(true);

    // Calculate bounding volumes
    linkBox1->calculateBound();
//    oMesh->calculateBound();
    xMesh->calculateBound();
    linkBox2->calculateBound();
    linkRoom->calculateBound();

    mWorld.add(*linkBox1);
//    mWorld.add(*oMesh);
    mWorld.add(*xMesh);
    mWorld.add(*linkBox2);
    mWorld.add(*linkRoom);

    // Add light
    mWorld.add(*mLight);
}

//------------------------------------------------------------------------------
void ShadowVolumeContext::update()
//------------------------------------------------------------------------------
{
    if (!theVideoCard().supports(VideoCard::DS_24_8))
    {
        return;
    }

    mCameraController.update();
        
    Vector4 lightPos;
    lightPos[0] = 3*cos(theMetronome().getAbsolute()/2000.0f);
    lightPos[1] = 1.5f;
    lightPos[2] = 3*sin(theMetronome().getAbsolute()/2000.0f);
    lightPos[3] = 1.0f;
    mLight->setPosition(lightPos);
    mLightParticles.setPosition(Vector3(lightPos[0],lightPos[1],lightPos[2]));
    mLightParticles.update();

    linkBox1->setLight(0, lightPos);
    linkBox2->setLight(0, lightPos);
    xMesh->setLight(0, lightPos);
}

//------------------------------------------------------------------------------
void ShadowVolumeContext::render()
//------------------------------------------------------------------------------
{
    if (!theVideoCard().supports(VideoCard::DS_24_8))
    {
        theRenderer().print(5, 80, "No 8 bit stencil supported");
        return;
    }

    // Set projection to infinity
//    Matrix4x4 projection;
//    makeProjection(projection, mCamera.getNearPlane(),
//                               mCamera.getFieldOfView(),
//                               mCamera.getAspectRatio());
//
//    theRenderer().setProjectionMatrix(projection);

    theRenderer().render(mWorld);
//    theRenderer().renderShadow(); -> obsolete, I'm not rendering a shadow quad anymore
    theRenderer().render(mLightParticles);
}

//------------------------------------------------------------------------------
void ShadowVolumeContext::exit()
//------------------------------------------------------------------------------
{
}

} // namespace Ge
