#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jun-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "TerrainContext.h"

#include <System/Log.h>
#include <System/FilePath.h>
#include <System/FileSystem.h>

#include <Engine/Mesh.h>
#include <Engine/Link.h>
#include <Engine/TexturePool.h>
#include <Engine/Renderer.h>

#include <Terrain/MidpointDisplacement.h>
#include <Terrain/FaultFormation.h>
#include <Terrain/ParticleDeposition.h>
#include <Terrain/HeightMap.h>
#include <Terrain/Terrain.h>

#include <Image/Image.h>

#include <Application/Settings.h>

namespace
{

using namespace Ge;

inline
//------------------------------------------------------------------------------
FilePath& path()
//------------------------------------------------------------------------------
{
    static FilePath ret(FilePath(theFileSystem().getApplicationPath().c_str()) +
                        (theSettings.getVariable("data").asString() +
                        "/Textures/").c_str());
    return ret;
}

//------------------------------------------------------------------------------
void addBaseTextures(Terrain& terrain)
//------------------------------------------------------------------------------
{
    terrain.add(Terrain::Ground((path()+"lumps.jpg").get().c_str(),
                                Interval(0, 30), 10, Interval(0, degToRad(60)), 0.1f));
    terrain.add(Terrain::Ground((path()+"grass6.tga").get().c_str(),
                                Interval(30, 120), 10, Interval(0, degToRad(60)), 0.1f));
    terrain.add(Terrain::Ground((path()+"rock086.tga").get().c_str(),
                                Interval(120, 200), 10, Interval(0, degToRad(60)), 0.1f));
    terrain.add(Terrain::Ground((path()+"snow3256.jpg").get().c_str(),
                                Interval(200, 255), 10, Interval(0, degToRad(60)), 0.1f));
    terrain.add(Terrain::Ground((path()+"rock.bmp").get().c_str(),
                                Interval(0, 255), 10, Interval(degToRad(60), PI_HALF), 0.1f));

//    // Test slope
//    terrain.add(Terrain::Ground((path()+"grass6.tga").get().c_str(),
//                                Interval(0, 255), 20, Interval(0, degToRad(45)), 0.0f));
//    terrain.add(Terrain::Ground((path()+"rock086.tga").get().c_str(),
//                                Interval(0, 255), 20, Interval(degToRad(45), degToRad(90)), 0.0f));
    
//    // Test height
//    terrain.add(TextureBlender::Ground((path()+"grass6.tga").get().c_str(),
//                                       Interval(0, 120), 0, Interval(0, PI_HALF), 0.0f));
//    terrain.add(TextureBlender::Ground((path()+"rock086.tga").get().c_str(),
//                                       Interval(120, 255), 0, Interval(0, PI_HALF), 0.0f));
}

template<class ALGORITHM>
//------------------------------------------------------------------------------
void generateTerrain(ALGORITHM& height, Mesh*& mesh, Image& texture)
//------------------------------------------------------------------------------
{
    HeightMap hm(height);
    Terrain terrain(hm);

    terrain.setYRange(Interval(0, 30));
    addBaseTextures(terrain);    

    //------------------------------------------------------------------------------
    // Generate mesh
    //------------------------------------------------------------------------------

    Vector3* vertices = 0;
    Vector3* normals = 0;
    Vector2* uvs = 0;
    UInt vertexCount(0);
    
    UInt16* indices = 0;
    UInt indexCount(0);

    terrain.createMesh(true,
                       vertexCount, vertices,
                       &normals, &uvs,
                       &indexCount, &indices);
        
    //------------------------------------------------------------------------------
    // Generate texture
    //------------------------------------------------------------------------------   
   
    terrain.createTexture(texture);

#ifndef NDEBUG
    // Save generated texture for debugging
    texture.save(Image::BMP, "TerrainTexture.bmp");
#endif

    //------------------------------------------------------------------------------
    // Add to scene
    //------------------------------------------------------------------------------
    mesh = GE_NEW Mesh(1, vertexCount, vertices, normals, uvs, 0,
                       indexCount, indices);
}

} // namespace

namespace Ge
{

//------------------------------------------------------------------------------
TerrainContext::TerrainContext() :
//------------------------------------------------------------------------------
Context("TerrainContext"),
mCamera(*Locator::create(Matrix3x4::IDENTITY())),
mCameraController(mCamera),
mChangeAlgorithm(Keyboard::SPACE),
mCurrentAlgorithm(FAULT_FORMATION),
mWorld(0)
{
}

//------------------------------------------------------------------------------
void TerrainContext::init()
//------------------------------------------------------------------------------
{
    mLight = GE_NEW Light;
    generateWorld();
}

//------------------------------------------------------------------------------
void TerrainContext::generateWorld()
//------------------------------------------------------------------------------
{
    if (mWorld)
    {
        GE_DELETE mWorld;
        mWorld = 0;
    }

    mWorld = GE_NEW World;
    mWorld->setCamera(&mCamera);
    mWorld->add(*mLight);

    Mesh* mesh = 0;
    Image texture(256, 256, Image::RGBA_8888);
    
    // Reomve old texture from pool.
    theTexturePool().clear("Terrain");

    //------------------------------------------------------------------------------
    // Generate height map
    //------------------------------------------------------------------------------

    switch (mCurrentAlgorithm)
    {
    case IMAGE_MAP:
        {
            Image height((path()+"HeightMap.tga").get().c_str());
            generateTerrain(height, mesh, texture);
        }
        break;

    case FAULT_FORMATION:
        {
            FaultFormation height(100, 100, 1, 0.1f, 64, 1, 0.2f);
            height.calculate();
            generateTerrain(height, mesh, texture);
        }
        break;

    case MIDPOINT_DISPLACEMENT:
        {
            MidpointDisplacement height(6, 32, 1.0f, 0.4f);
            height.calculate();
            generateTerrain(height, mesh, texture);
        }
        break;

    case PARTICLE_DEPOSITION:
        {
            ParticleDeposition height(100, 100, 0.5f, 30, 8, 2000, 6000, 0.51f);    
            height.calculate();
            generateTerrain(height, mesh, texture);
        }
        break;

    default:
        GE_ASSERT_MSG(false, "Unknown algorithm.");
    }

    Link* l = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()), *mesh);

//    // Need this so that bounding sphere gets calculated
    l->calculateBound();
//    g->updateGeometry();
    Float boundSize(1.5*l->getBound().getRadius());

    // Repostion camera so that the whole terrain is seen
    Matrix3x4 lookAt;
    Vector3 cameraPos(0,boundSize/2,-boundSize);
    makeLookAt(lookAt, cameraPos, Vector3(0,0,0), Vector3(0,1,0));
    mCamera.getLocator().get() = lookAt;
    mCamera.getLocator().update();
    mCameraController.setDistance(cameraPos.length());
    mCameraController.recalculate();

    Material* material = GE_NEW Material;
    material->set(0, *(theTexturePool().get(texture, "Terrain")));
    l->set(*material);

    mWorld->add(*l);

    // Add coordinate axes
//    g = GE_NEW Geometry(*CoordinateAxes::create(2*boundSize));
//    g->setName("CoordinateAxes");
//    root->addChild(*g);
    
   
//#ifndef NDEBUG
//    // Print scene tree for debugging
//    root->printSubTree(GE_LOG_DEBUG("Scene tree:"<<std::endl));
//#endif
}

//------------------------------------------------------------------------------
void TerrainContext::update()
//------------------------------------------------------------------------------
{
    // Rotate camera and light
    mCameraController.update();
    Vector3 cameraDir;
    mCamera.getLocator().get().getColumn(MatrixBase::AT, cameraDir);
    Vector4 lightPos;
    set(lightPos, -cameraDir, 0.0f);
    mLight->setPosition(lightPos);

    // Check for algorithm switch
    mChangeAlgorithm.update();
    if (mChangeAlgorithm.isButtonDown())
    {
        mCurrentAlgorithm = Algorithm(UInt(mCurrentAlgorithm)+1);
        if (mCurrentAlgorithm == ALGORITHM_COUNT)
        {
            mCurrentAlgorithm = ALGORITHM_START;
        }

        generateWorld();
    }
}

//------------------------------------------------------------------------------
void TerrainContext::render()
//------------------------------------------------------------------------------
{
    if (mWorld)
    {
        theRenderer().render(*mWorld);
    }
}

//------------------------------------------------------------------------------
void TerrainContext::exit()
//------------------------------------------------------------------------------
{
    if (mWorld)
    {
        GE_DELETE mWorld;
    }
}

} // namespace Ge
