#include "Precompiled.h"
//------------------------------------------------------------------------------
// (c) Jul-2002 Gottfried Chen
//------------------------------------------------------------------------------

#include "TestContext.h"

#include <Application/Settings.h>

#include <Engine/ModelPool.h>
#include <Engine/Quad.h>
#include <Engine/SphereMesh.h>
#include <Engine/Mesh.h>
#include <Engine/Link.h>
#include <Engine/Renderer.h>
#include <Engine/TexturePool.h>

#include <System/FilePath.h>
#include <System/FileSystem.h>

namespace Ge
{

//------------------------------------------------------------------------------
TestContext::TestContext() :
//------------------------------------------------------------------------------
Context("TestContext"),
mCamera(*Locator::create(Matrix3x4::IDENTITY())),
mCameraController(mCamera)
{
}

//------------------------------------------------------------------------------
void TestContext::init()
//------------------------------------------------------------------------------
{
    // Position camera
    Vector3 cameraPos(0, 10, -10);
    Matrix3x4 lookAt;
    makeLookAt(lookAt, cameraPos, Vector3(0,0,0), Vector3(0,1,0));
    mCamera.getLocator().get() = lookAt;
    mCamera.getLocator().update();
    mCameraController.setDistance(cameraPos.length());
    mCameraController.recalculate();
    mWorld.setCamera(&mCamera);

    // Create light
    mLight = GE_NEW Light;

    // Load .obj cube with more than one material
    FilePath oPath(theFileSystem().getApplicationPath());
    oPath += theSettings.getVariable("data").asString();
    oPath += "animations/testrobot/cube.obj";
    Cluster* oMesh = theModelPool().get(oPath.get().c_str());
    oMesh->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(-3,-2,0));
    oMesh->getLocator()->update();
    oMesh->calculateBound();

    // Clone cube
    Cluster* oMeshClone = oMesh->clone();
    oMeshClone->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(-3,2,0));
    oMeshClone->getLocator()->update();
    oMeshClone->calculateBound();

    // Load .obj robot with single material
    oPath.set((theFileSystem().getApplicationPath() +
               theSettings.getVariable("data").asString() +
               "animations/testrobot/test.obj").c_str());
    Cluster* robot = theModelPool().get(oPath.get().c_str());
    robot->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(-5,-2,0));
    robot->getLocator()->update();
    robot->calculateBound();

    // Load .x tiger with recalculated normals
    theModelPool().setRecalculateNormals(true);
    FilePath xPath(theFileSystem().getApplicationPath());
    xPath += theSettings.getVariable("data").asString();
    xPath += "xfiles/tiger.x";
    Cluster* xMesh = theModelPool().get(xPath.get().c_str());
    xMesh->calculateBound();
    theModelPool().setRecalculateNormals(false);

//    // Load .x spaceship
//    theModelPool().setRecalculateNormals(true);
//    xPath.setFilename("
//    FilePath xPath(theFileSystem().getApplicationPath());
//    xPath += theSettings.getVariable("data").asString();
//    xPath += "tigertest/tiger.x";
//    Cluster* xMesh = theModelPool().get(xPath.get().c_str());
//    xMesh->calculateBound();
//    theModelPool().setRecalculateNormals(false);

    // Create a quad
    Link* quad = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()),
                             *Quad::create(2, 2));
    quad->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(2,0,0));
    quad->getLocator()->update();
    quad->calculateBound();
    FilePath path(theFileSystem().getApplicationPath());
    path += theSettings.getVariable("data").asString();
    path += "/animations/boat/boat.tga";
    Material* quadMaterial = GE_NEW Material;
    quadMaterial->set(0, *(theTexturePool().get(path.get().c_str())));
    quad->set(*quadMaterial);

    // Create a sphere
    Link* sphere = GE_NEW Link(*Locator::create(Matrix3x4::IDENTITY()),
                               *SphereMesh::create(1.5f));
    sphere->getLocator()->get().setColumn(MatrixBase::TRANSLATION, Vector3(2,3,0));
    sphere->getLocator()->update();
    sphere->calculateBound();
    
    mWorld.add(*mLight);
    mWorld.add(*oMesh);
    mWorld.add(*oMeshClone);
    mWorld.add(*robot);
    mWorld.add(*xMesh);
    mWorld.add(*quad);
    mWorld.add(*sphere);
}

//------------------------------------------------------------------------------
void TestContext::update()
//------------------------------------------------------------------------------
{
    // Rotate camera and light
    mCameraController.update();
    Vector3 cameraDir;
    mCamera.getLocator().get().getColumn(MatrixBase::AT, cameraDir);
    Vector4 lightPos;
    set(lightPos, -cameraDir, 0.0f);
    mLight->setPosition(lightPos);
}

//------------------------------------------------------------------------------
void TestContext::render()
//------------------------------------------------------------------------------
{
    theRenderer().render(mWorld);
}

//------------------------------------------------------------------------------
void TestContext::exit()
//------------------------------------------------------------------------------
{
}

} // namespace Ge
